#ifndef _bss_h
#define _bss_h
#ifndef lint
static char *bss_h_sccsid = "@(#) bss.h 2.6 98/03/19\n";
#endif  /* lint */

/******************************************************************************

  bss.h  SW March 92

  Okapi BSS header containing externs etc

******************************************************************************/
#include "defines.h"
#include "parms.h"

#ifndef NGROUPS_MAX
#define NGROUPS_MAX 32
#endif

#define LOG_STEM "bss_syslog"

#define SELECTION_TREE_CROSSOVER 5 /* Probably 5 - 8 is about right */

#define MAXIXREAD 1000000	/* Inc from 4096 May 93, from 200000 Oct 96 */
#define MAXSETRECS 16384	/* Inc from 2048 Apr 95 */

/* Feb 96: for real-valued weights. See bss_combine and i0_parse+.y */

#ifdef FLOAT
#define MAXWEIGHT (float)1e20	/* Weight must be < this and > next */
#else
#define MAXWEIGHT (double)1e20	
#endif
#define MINWEIGHT -(MAXWEIGHT)
#define WEIGHT_NOT_ASSIGNED  MAXWEIGHT * 2
#define Legalweight(w)  (w<MAXWEIGHT && w>MINWEIGHT)

/* OP definitions for combine(). Stored in Setrec.op */
/* NB if these are changed change translate_opcode() */

#define OP_COPY -4		/* For limit only, or simply copy */
#define OP_TOPN -3		/* Pseudo-op */
#define OP_NOOP -2		/* Unrecognised op */
#define OP_LKUP -1		/* Dummy. For Setrec, not combine() */
#define OP_BM1 0		/* Best match, weights & thresholds supplied */
#define OP_BM11 1		/* Coeff of BM9 applied to F4 */
#define OP_BM15 2
#define OP_BM1100 3		/* Universal BM11 op with k1 and k2 */
#define OP_BM1500 4		/* Universal BM15 op with k1 and k2 */
#define OP_BM25 5		/* New "power" version of bm11 etc */
#define OP_BM250 6
#define OP_BM251 24		/* New May 97 */
#define OP_BM2500 7
#define OP_BM26 23
#define OP_BM2600 25
#define OP_LR1 8
#define OP_LR2 9
#define OP_LR3 50
#define OP_LR4 51
#define OP_LR5 52

#define OP_AND1 10		/* Normal AND, no weights in output */
#define OP_AND2 11		/* AND as limit */
#define OP_ANDL OP_AND2
#define OP_MARK 21		/* Oct 95 */
#define OP_AND3 OP_MARK
#define OP_OR1 12		

#define OP_NOT1 14		/* Normal A and-not B */
#define OP_NOT2 15		/* NOT as limit */
#define OP_NOT3 16		/* Removes set[1] from set[0] */
#define OP_UNCOMBINE OP_NOT3
#define OP_NOTL OP_NOT2
#define OP_ADJ1 17		/* Ignores stopwords */
#define OP_ADJ2 18		/* Counts stopwords */
#define OP_SAMEF1 19		/* Same field */
#define OP_SAMES1 20		/* Same field and sentence */

#define OP_LIMIT 13		/* March 96: decided to make limit an op */

/* Flags values for lookup and combine etc. Many of these values come
   through from the interface. The top 8 bits of FFL may contain a "mark"
   of up to 7 bits + sign. If op is OP_MARK they are copied to (or removed
   from if negative) the flags field of a bss_Gp (see below) struct. If op
   is not OP_MARK they would be ignored. If a set has marks they are inherited
   by any  derived set. This is determined by the HASMARKS bit of the set's
   pstgtype field. */

#define FFL_NOPOS 1		/* If 1-bit set no positional data output
				   by combine (Nov 94) */
#define FFL_NOSAVE 2		/* Makes no output set if 2-bit set */
#define FFL_FILE 4	
#define FFL_MEM 8		/* Probably, if MEM no file, see doc */
#define FFL_OPEN 16		/* For setup_pstgs(). create if nec and open
				   files as appropriate. E.g. if just going
				   to be used for merge. */
#define FFL_KEEP 32		/* Keep file open on close_pstgs()
				   (default close). */
#define FFL_RECOVER_MEM 64	/* Hint to close_pstgs to recover allocated
				   and/or mapped memory. May write it to a
				   file if file doesn't exist. Not implemented*/
#define FFL_DELETE 128		/* Delete input sets after merge */
#define FFL_DO_STATS 256	/* Does weightsum and weightsumsq if true */
#define FFL_HOLD_MEM 512	/* Holds memory pages for set if possible */

#define FFL_NOTF 1024		/* See also FFL_NOPOS. In applications, NOTF
				   implies NOPOS (done at i0 level). */
#define FFL_NO_SELECTION_TREE 2048
				/* Tells bss_combine() not to use selection
				   tree even if num >= crossover */
/* Need also perhaps way of indicating (and doing) field restriction(s).
   Might need another flags-type word with bit set for any wanted field. */

#define FFL_GETMARK(fl) ((fl)>>24)
#define FFL_DEMARK(fl)  ((fl) &= 0x00ffffff)

/* Set condition flags */

#define OP_FLAGS_BMLIKE 16
#define OP_FLAGS_ANDLIKE 32
#define OP_FLAGS_ADJLIKE 64
#define OP_FLAGS_ORLIKE 128
#define OP_FLAGS_LRLIKE 256
#define OP_FLAGS_NOTLIKE 512
#define OP_FLAGS_LIMITLIKE 1024
#define OP_FLAGS_TOPN 2048

#define WEIGHT_DIST_GRAD 20	/* Number of weight_dist_vals */
				/* Bit temporary */
/******************************************************************************

  posting type flag values

******************************************************************************/

/* NB if these are changed change translate_pstgtype */

#define HASWT 2			/* Posting has a weight field */
#define HASTF 4			/* Posting might have neither positional
				   fields nor any term frequency field
				   (by default there is at least a TF field) */
#define LONGPOS 8		/* Long positional fields */
#define TEXTPOS 16		/* Long text-type pos fields */
#define HASPOS (LONGPOS|TEXTPOS) /* Posting has positional fields */

#define HASCONTRIBS 32

#define TEXTPSTGTYPE 64
#define HASPASSAGES 128
#define HASMARKS 256		/* Oct 95, OP_AND3 (OP_MARK) */

/* <rec> or <addr> field is always 4 bytes, but <recnum> is 3 bytes.
   Weights are 4, LONGPOS and TEXTPOS are 4.
   The posting length or numpos pseudo-fields are "1 1/2"
   (1 or 2, see the functions fput_vshort() etc). */

#define PTYPE12 TEXTPSTGTYPE|TEXTPOS|HASTF
                /* Type 4, 6 or 8 index postings */
#define PTYPE13 TEXTPSTGTYPE|LONGPOS|HASTF
                /* Type 5, 7 or 9 index postings */
#define PTYPE14 TEXTPSTGTYPE|HASTF

/* Defines for set dump commands (Jan 96) */
#define D_ALL 0
#define D_TTF 1
#define D_NP  2

/******************************************************************************

  bss_Gp

  The most general posting structure, used by read_pstg_unpacked()

  March 96: these are unnecessarily big. Should use just pointers for the
  three arrays instead of for only one of them.

******************************************************************************/

typedef struct para_rec {
  short fd ;			/* Field number */
  short dummy ;			/* Just to fill it if not packed */
  u_short sp, fp ;			/* Start & finish paragraph number (0- ) */
  float wt ;
} Para_rec ;

/* #define MAXPARRECS 255   Never used more than 2 (Apr 97) */
#define MAXPARRECS 2

typedef struct bss_gp {
  u_short maxpos ;		/* With maxpara, implies size */
  u_short numpos ;		/* If ! HASPOS then this field holds tf
				   (the number of pos fields there were.
				   Hence working out how to read a
				   posting depends on knowledge of
				   the HASPOS bits.
				   (Jan 96) */
  short numpars ;
  short coord ;			/* No of nonzero tfs (might as well be short
				   altho only 0 - 255) */
  int rec ;			/* low 24 bits only unless database LARGE (Feb 1998) */
  float wt ;			/* Jan 96 */
  u_char flags ;			/* New Oct 95: to hold rel info etc */
  u_char dummyflags ;
  Para_rec pars[MAXPARRECS] ;
  /* Next 2 fields only for op=OP_SPEC.., getting log regression data etc */
  struct str_tf {
    u_short str ;			/* Stream no (0 - 255) */
    u_short tf ;			/* Term frequency for this stream */
  } t[256] ;
  U_pos *p ;
} bss_Gp ;

/* Coordination level is number of nonzero tfs */

/* Packs down for file storage into

[<numprs>][<coord>][<numpos>]<rec>[<wt>][<flags>][<prs>][<trmno><tf>]][<pos>]
     1         1       1|2     3    4       2       10       2     2     4

*/

/******************************************************************************

  posting_flag values for bss_Gp

******************************************************************************/

#define PSTG_REL 1		/* 1-bit of flags field of bss_Gp */


/******************************************************************************

  Setrec struct, used for all BSS sets

  Rearranged and modified many times, last Jan 96

******************************************************************************/

typedef struct bss_setrec {
  BOOL used ;			/* True if in use */
  int num ;			/* The record's index in the global array of
				   setrecs. Also acts as the
				   variable part of the postings filenames.
				   File names are bss_pf<0|1>_<num>.<pid> */
  int ctime ;			/* Creation time in seconds */
  int atime ;			/* Initially as ctime. Updated on leaving
				   setup_pstgs, open_pstgs, close_pstgs. */
#define S_ATOMIC 1
#define S_MERGE 2
#define S_LIMIT 4
#define S_REMOVE 8
#define S_RLIMIT 16
#define S_TOPN 32
#define S_IRN 64
  int source_type ;		/* See defines just above. Atomic status
				   field removed March 96 */
  char *search_term ;		/* malloc'd storage for retrieved term if
				   successful single term search */
  int search_type ;		/* 0: exact match, -2: glt, -1: gngt,
				   1: lnlt, 2: lgt. (If source_type=0) */
  int search_result ;		/* As search_type, & 255 for complete failure.
				   Check this (May 93) */
#define MIXED_ATTRIBUTE -1
#define NO_ATTRIBUTE -2
  int attr_num ;		/* Seq number (0..) of the attribute used in
				   lookup (source_type 0). Maintained
				   through combine ops as far as poss. If no
				   single attribute applies this is given as
				   -1 (read "mixed") (New Dec 94) */
  /* Next few fields only used for term lookup sets (source type 0) */
  int ixnum ;			/* If source_type is 0, ixnum is that
				   of the index used, ix_chunk and ix_offset
				   the chunk and offset of the ITR which was
				   retrieved. (These enable retrieval of
				   next and previous terms.) */
  INDEX *ix ;			/* Database's index struct. NULL if set
				   result of merge from more than one index. */
  int ix_chunk ;
  int ix_offset ;
  u_int limits ;		/* If source_type 0, the limits used in
				   lookup. Jan 96: should be used for any
				   type */
  /* Next two new in 2.05 Nov 96, not used yet */
  struct db *db ;		/* Set's database */
  char *db_name ;		/* Might be used if set kept when db closed */

  BOOL adj_no_good ;		/* FALSE iff adj can be done on this set
				   (i.e. is output from pure OR and/or ADJ) */
  int op ;			/* op which produced this set */
  int n_streams ;		/* No of input streams, 1 if this is a
				   single term lookup, otherwise supposed to
				   be cumulative (but check sometime). */
  int pstgtype ;		/* See HASWT, HASPOS etc, above. This is
				   vital as packing/unpacking postings
				   depends on it. */

  double aw, gw, mpw ;		/* Weight thresholds in merge */
  double maxweight ;		/* Highest weight achieved */
  double minweight ;		/* Not seriously used (Jan 96) */
  int nmpw, ngw, naw ;		/* Result (max poss weight, "good" weight,
				   total). If no weights all have same value.
				   Note naw is actual total no. of pstgs, not
				   necessarily the same as fsize because some
				   pstgs may not have been read. */
  int nmaxweight ;		/* Number with the highest weight attained */
  int ttf ;			/* Actual total term frequency (merge as well
				   as atomic).*/
  int max_tf ;			/* Atomic only: guide to the size of the pstg
				   field. Index types > 7 only. */
  int weight_dist_grad ;	/* Scale for weight_dist[]. If 0 nothing
				   recorded, else in steps of mpw/<this> */
  int weight_dist[WEIGHT_DIST_GRAD + 1] ; /* To store numbers of postings in
					     0-n%, n-2n%
					     and so on, of max poss wt */
  double weight_floors[WEIGHT_DIST_GRAD + 1] ;
  /* Postings counted in weight_dist[j] have weight >= floor[j] and
     < floor[j + 1] (floor[WDG] considered infinite). Jan 96: shouldn't be
       needed, cheaply calculated on the fly */
  /* Two fields added Nov 94 for occasional statistical purposes (SER) */
  double weightsum ;		/* Just that */
  double weightsumsq ;		/* Sum of squares of weights */

  int totalp ;			/* Number of pstgs to be considered. If atomic,
				   and limit != 0, it's the number of
				   "unlimited" postings for the term. If from a
				   merge, it's the sum of the numbers of
				   postings in all the input streams, even if
				   this is greater than big N for the limit.
				   Used in estimating how much memory to
				   allocate for the set's posting storage.
				   (Rev Feb 96)*/
  int total_tf ;		/* Term frequency analogue of totalp */
  /* Postings are stored in file(s) and/or in memory. The memory may be
     a mapping of the index's postings file (if atomic) */

  /* FILES */

  /* There may be two files. File 0 is a raw and complete file of postings
     as produced by a lookup or combine. File 1 is a (possibly partial) file
     of pointers to the postings in File 0 in descending weight order.
     Each record in file 1 is the 4-byte offset of a record in file 0.
     File 1 (recfile) may be kept open between function calls, so a copy of
     the file pointer is kept, in fact both file pointers are kept if file is
     open). Note that in general file 1 is incomplete (one rarely need to
     output all the postings in weight order, and if outputting sequentially
     file 1 is not needed */

#define SET_FILE 1
#define SET_FILE_EXISTS 2
#define SET_FOPEN_READ 4
#define SET_FOPEN_WRITE 8
#define SET_FOPEN (SET_FOPEN_READ|SET_FOPEN_WRITE)
#define SET_FILE0_PF 16		/* Can read pstgs direct from index.
				   Always true for atomic set with limit 0
				   and not NOPOS */
#define SET_MEM 256
#define SET_MEM_MALLOC 512
#define SET_MEM_MAPPED 1024
#define SET_MEM_EXISTS (SET_MEM_MALLOC|SET_MEM_MAPPED)
  
#define SET_PF1_RECLEN sizeof(int) /* These are integer offsets */

  int output_flags[2] ;		/* Replaces file0_pf and gives other info. */
#define SET_NATORDER 1		/* Not sure these'll be used */
#define SET_WTORDER 2
  /* Next three for atomic (lookup) sets only */
  int pf_vol ;			/* Volume number in pstgs file (Apr 95),
				   really a byte at most (0-) */
#ifdef LARGEFILES
  off_t pf_offset ;		/* Offset in own pf */
#else
  long pf_offset ;		/* Offset in own pf */
#endif
  FILE *global_pf ;		/* Index should have postings file open.
				   This is its pointer. (Feb 96). */
  int global_pf_length ;	/* Length of postings in Db's pf before
				   taking account of limits or nopos
				   (atomic only) */
#if LARGEFILES
  off_t global_pf_offset ;	/* Offset in Db's pstgs file. */
#else
  long global_pf_offset ;	/* Offset in Db's pstgs file. */
#endif
  FILE *pfile[2] ;		/* File pointer file i, or NULL if the file is
				   closed or doesn't exist. */
  /* Feb 96: the '0' elements in the following 2 aren't used any more */
  int np[2] ;			/* Number of recs in file or memory <= naw */
  int p_length[2] ;		/* Actual length of packed postings for this
				   set. Added Nov 95 for memory mapping.
				   If file exists it is equally the file size.
				   Corresponds to pfile[] etc. */
  char *pbase[2] ;		/* New June 95, not properly implemented yet.
				   Memory address for base of postings, or
				   NULL if not in memory. pbase[1] should be
				   cast to int * because the memory contains
				   integer offsets (((int *)pbase[1])[n]
				   is the offset of posting n in pfile[0] or
				   from pbase[0] */
  char *pptr[2] ;		/* Pointer to next pstg at or above pbase. */
  int next_pstg[2] ;		/* Posting no. to which file or pptr points */
  char *membase[2] ;		/* If mmapped, membase in general lower
				   than pbase as has to be mult of pagesize */
#if LARGEFILES
  long long memsize[2] ;	/* Actual amount of memory mapped or malloced */
#else
  int memsize[2] ;		/* Actual amount of memory mapped or malloced */
#endif LARGEFILES
  bss_Gp *pstg ;		/* Current structured posting when needed */
  int raw_plength ;		/* (Raw) length of posting in the pstg field
				   (i.e. length when packed) */
} Setrec ;

extern Setrec *Setrecs[MAXSETRECS + 1] ;
extern int Num_Setrecs ;
extern int Current_record_setnum ;

/******************************************************************************

  Macro for returning whether recnum satisfies limit
  Should for speed check Db.lims exists outside this
******************************************************************************/

#define Sats_limit(rec,limit) \
 ((Db.lims && (((Db.lims[(rec) - 1] & (limit)) == (limit)) || ((rec == MAXINT)))))

#define Getlimit(rec) \
     (Db.lims == NULL) ? 0 : Db.lims[(rec) - 1]

extern off_t  Current_rec ;
extern int Current_reclen, Current_recnum ;
extern char *BSS, *Okapi ;
extern char *version, *vdate ;
extern char *Info_buf ;

extern int Pid ;		/* bss_globs.c */
extern int Pagesize ;
extern char Username[] ;
extern int Uid ;
extern int Gid ;
extern int N_groups ;
extern int Groups[] ;
extern int Rec_db ;		/* bss_globs (recording open and close db) */
extern int Maxfiles ;
extern int Largefiles ;
extern int Num_open_files ;

/******************************************************************************

  Highlight on/off tables: type 3 indexes  SW Jan 93

  Set up by bss_highlight_rec()and passed to hl_next_sentence()
  Prob more or less obsolete (July 95)

******************************************************************************/

struct highlight_table {
  int size ;			/* Number of records */
  int ons[256] ;
  int offs[256] ;
} ;

/******************************************************************************

  hl_rec struct for storing highlighting info for SHOW commands

  July 95 SW

******************************************************************************/

struct hl_rec {
  char code ;
  char oset[8] ;
  char l[8] ;
} ;


/******************************************************************************

  From extract.h May 1993

******************************************************************************/

/*****************************************************************************

  extract.h  SW June 1992

  Header file for bss_extract and indexing progs

*****************************************************************************/

#ifdef EXTRACT

/* What is this? */
#define MAX_IXTLEN 127

/* Output types & places */

/* Places */

#define EX_FILE 1
#define EX_STRINGS 2
#define EX_GEN 4
#define EX_BUF 8

/* General types */
/* IX0 - IX3 probably obsolete Jan 96 */
/* #define EX_IX0 0x100		/* OBSOLETE term rec beast */
#define EX_IX4 0x1000		/* term recnum pos_text */
#define EX_IX5 0x2000		/* term recnum pos6 */

#define EX_QE  0x10000		/* term, f. by space for 5 2-byte fields (?).
				   Or should it be fitted directly into
				   a sorted? STRINGS struct */
#define EX_HL  0x20000		/* don't know what to do with this yet.
				   Should it only go in if it's a qterm?
				   See temp_output_term in rfsubs.c.*/
#define EX_PARSE 0x40000		/*  */
#define EX_SUPERPARSE 0x80000

#define IX 0x100000
#define BLURBY 0x200000		/* Prevents output of I_terms from TEXT
				   fields */
#define PRECOUNT 0x400000	/* Else null-terminated */

#define SOURCE 0x800000
#define TEXT 0x1000000
#define EX_DIAG 0x40000000

/* #define EX_OP_IX0 (EX_IX0|IX|PRECOUNT|EX_STRINGS) OBSOLETE */
#define EX_OP_IX1 (EX_IX1|IX|PRECOUNT|EX_STRINGS)
#define EX_OP_IX2 (EX_IX2|IX|PRECOUNT|EX_STRINGS)
#define EX_OP_IX3 (EX_IX3|IX|PRECOUNT|EX_STRINGS)
#define EX_OP_IX4 (EX_IX4|IX|PRECOUNT|EX_STRINGS)
#define EX_OP_IX5 (EX_IX5|IX|PRECOUNT|EX_STRINGS)

#define EX_OP_QE  (EX_QE|PRECOUNT|EX_STRINGS)
#define EX_SUPER (EX_QE|EX_STRINGS|SOURCE)
#define EX_SUPER_NOSOURCE (EX_QE|EX_STRINGS)
#define EX_USER_PARSE (EX_PARSE|EX_BUF)
#define EX_USER_SUPERPARSE (EX_SUPERPARSE|EX_BUF|SOURCE)
#define EX_USER_SUPERPARSE_TEXT (EX_SUPERPARSE|EX_BUF|SOURCE|TEXT)
#define EX_USER_SUPERPARSE_NOSOURCE (EX_SUPERPARSE|EX_BUF)
				/* For parsing user terms. Just outputs string
				   of strings to buf. Also intermediate in
				   qe extraction. */


struct outrec {
  int type ;			/* Output type: how to output the term */
  char *term ;			
  int tl ;			
  char *source_term ;
  int stl ;
  char *output ;		/* This may be cast to struct of some kind or
				   just a character array */
  u_int rec ;			/* Record's address if any, or recnum */
  u_short limits ;
  char sgnum ;			/* Search group number, ? only used in QE */
  char gsclass ;		/* Added Aug 92 for QE
				   Probably only blank or 'I' of interest */
  int fdnum ;			/* Positional info */
  int senum ;
  int wdnum ;			/* Also token number */
  int nt ;			/* Number of tokens in a term */
  int sw ;			/* Number of preceding stopwords */
  U_pos p ;
  short source_offset ;		/* Source info for highlighting termlist */
  short source_len ;
} ;

#endif EXTRACT
#include "bss_protos.h"

#endif ndef _bss_h
