#ifndef _defines_h
#define _defines_h
#ifndef lint
static char *defines_sccsid = "%Z% %M% %I% %E%\n";
#endif  /* lint */

/******************************************************************************

  defines.h  SW Apr 92

  Low-level Okapi defines etc for both interactive and BSS systems

  Mostly taken from stdeqs.h

******************************************************************************/

#if LARGEFILES
#define _LARGEFILE_SOURCE 1
#define _FILE_OFFSET_BITS 64
#endif LARGEFILES

#include <unistd.h>

#ifndef _SSIZE_T		/* Not in BSD */
#define	_SSIZE_T
typedef int	ssize_t;	/* used by functions which return a */
				/* count of bytes or an error indication */
#endif
#include <string.h>
#include <values.h>
#include <stdlib.h>
#include <stdio.h> 
#include <math.h>
#include <sys/time.h>
#include <ctype.h>
#include "site.h"

#define FALSE 0
#define TRUE 1
  
#define TIMER_LOAD_VAL 1000000	/* Start value for timers, doesn't matter so
				   long as it's pretty large */

#define MEM_BREAK_1 4		/* Pagesize * this. If less, malloc is used,
				   a zero memory mapping */

#define MAXLINE1 16384		/* Max line length for various readline
				   calls */
#define NAME_MAX 32		/* Max length of various names: stem
				   functions, gsls, attribute name strings */

#define BLOCKSIZE 512
#define KSIZE (2 * BLOCKSIZE)

#define MAXNUMFLDS 30		/* Not sure why this isn't 31 or 32 */
#define MAXNUMFIELDS MAXNUMFLDS	/* I always forget it */
#define MAXNUMFIELDS_TEXT 4	/* For "text" databases */
#define MAXIX 16		/* Indexes per database, inc from 3 Jan 93 */

#define MAXGSLS MAXIX		/* (June 97), was 3 * MAXIX since Apr 92,
				   before that 3 */
#define MAXRECLEN (127 * BLOCKSIZE) /* Rather arbitrary, unchecked, shld be
				      generous. Doubled to 64 * BLOCKSIZE
				      6 Apr 92, inc to 80 7 Apr 92 - should be
				      trec only. Inc again 9 Apr 92. This is
				      about abs max cos max fd len is 65535 */
#define MAXLIMS 16		/* Max number of limit bits for a database */
#define LIMIT_MASK_SIZE 2	/* Bytes */

#define GOODREADSIZE (16 * BLOCKSIZE)
#define CHUNKSIZE (16 * BLOCKSIZE)
#define SCHUNKSIZE CHUNKSIZE

#define MAXEXPSTGLEN 12 + 256 * 4 /* Max extended index pstg length (Nov 91)
				     (allows 256 positional fields) */
#define MAXTERMSPERFIELD 2048	/* Max index terms per field (for
				   source recording during extraction).
				   Inc from 1024 to 2048 exp Apr 91
				   Inc from 128 to 1024 Oct 89. */

#define MAXMERGESTREAMS 512	/* But you can't have that many unless in mem */

#define MAXINDEXINGPERREC MAXRECLEN
                                /* Length of storage for index terms extracted
                                   from one record */
				/* DOESN'T SEEM TO BE USED (Oct 89) */
#define RFTERMRECLEN 16 /* Guess at mean length of Rftermrec */
#define MAXRFTERMLISTSIZE 128 * BLOCKSIZE /* 9 Apr 92, for TREC: was  16K */
#define MAXRFNUMTERMS MAXRFTERMLISTSIZE / RFTERMRECLEN

#define TAB 0x9

/* Nov 97: don't know that 1st two used, rest are new except TEXT */
#define DB_CAT_TYPE 1		/* These new Apr 94, for speed */
#define DB_AI_TYPE 2
/* New Dec 97: default is not text, not free format, not external */
#define DB_TEXT_TYPE 4		/* Can be long, and textpos type */
#define DB_EXTERNAL_TYPE 16	/* Nov 97: Db made of raw data */
#define DB_FREE_FORMAT_TYPE 32	/* Fields not specified */
#define DB_LARGE_TYPE 64	/* Bibfile vols may be large,
				   bibdir records are 5-byte */

/* #define DB_INDIRECT_TYPE 8	/* Obsolete (< 1997), all dbs are like this
                                   Has indirect record addressing,
				   separate .lims file etc */

#define DB_OPEN_NOSEARCH 0
#define DB_OPEN_SEARCH 1
#define DB_OPEN_BM 2
#define DB_OPEN_LIMIT 4
#define DB_OPEN_SHOW 8		
#define DB_OPEN_PARAS 16	/* If true reads .par into Db.paraptrs */

#define FD_UNCHECKED 0
#define FD_CANT_OPEN -1

#define WEIGHT_DIST_GRAD 20	/* Was a local define somewhere (Nov 95) */

/*****************************************************************************/

#define RECORD_MARK 0x1d	/* Added Apr 94, replaces old HASH */
#define FIELD_MARK 0x1e

/*****************************************************************************/

typedef int BOOL ;
typedef struct kls{
   char * p ; /* Address of start */
   u_short l ; /* Length (0..65535) */
} KLS ;  /* Known-length string */
typedef int PFI();		/* Fn returning int */
typedef void (*PFV)() ;
typedef char *(*PFPC)() ;	/* Fn returning char * */
typedef struct {
  int (*func)() ;		/* Pointer to function */
  char name[NAME_MAX] ;		/* Function name */
} IFN ;

/*****************************************************************************

  From indexing.h May 93

******************************************************************************/

/* Believe this one is old - use not known - not used in BSS */

/* Obsolete #define MAXTERMLEN 80 /* Absolute maximum index term length */

/* Applies to terms in type 0 indexes */
#define MAXITRLEN 128

/* This one applies to terms in indexes types 1, 2 and 3 */
#define MAXEXITERMLEN 256 /* Max extended index term len (Nov 1991) */

/* #define MAXIXMERGERUNS MAXFILES - 9 /* i.e. 55 if 64 */ 

/*****************************************************************************/


#define GSLTOKLEN 5 /* Length of "see" gsl record's token */

/*****************************************************************************

  Some function-like macros

 *****************************************************************************/

#define cvlu(ch) (((ch) >= 'a' && (ch) <= 'z') ? (ch) - 'a' + 'A' : (ch))
#define cvul(ch) (((ch) >= 'A' && (ch) <= 'Z') ? (ch) - 'A' + 'a' : (ch))

/* Macros for reading bytes as unsigned values */
#define cv1(p)  ((u_int) *((u_char *)p))

/* Replaced by function Feb 1990
/* Restored Oct 96 */
/* Modified Nov 96 */
#define cv2hl(p) ((256 * (*(u_char *)(p))) + (*(unsigned char *)((p) + 1)))

/* Replaced by function Feb 1990 */
/* Restored Nov 96 */
#define cv3hl(p) ((65536 * *(unsigned char *)(p)) + cv2hl((p) + 1))

#define cv4hl(p) ((65536 * cv2hl(p)) + cv2hl((p) + 2))


#define cv2lh(p) ((*(unsigned char *)(p)) + (256 * *(unsigned char *)((unsigned char *)(p) + 1)))
#define cv3lh(p) (cv2lh(p) + (65536 * *(unsigned char *)((unsigned char *)(p) + 2)))
#define cv4lh(p) (cv2lh(p) + (65536 * cv2lh((unsigned char *)(p) + 2)))

#if HILO
#define cv2(p) cv2hl(p)
#define cv3(p) cv3hl(p)
#define cv4(p) cv4hl(p)
#else
#define cv2(p) cv2lh(p)
#define cv3(p) cv3lh(p)
#define cv4(p) cv4lh(p)
#endif

/* and for writing 2-byte and 4-byte values into char or BYTE arrays */
/* These macros cast the destination into pointer to unsigned char.
   It is hoped they avoid the perils of sign extension by casting the source
   to unsigned int */
#ifdef HILO
#define w2(p, w) (*(u_char *)(p) = ((unsigned)(w)/256), *((u_char *)(p) +1) = ((unsigned)(w)%256) )
#define w4(p, w) (w2(p, ((unsigned)(w)/65536)), w2((u_char *)(p)+2, ((unsigned)(w)%65536)) )
#else
#define w2(p, w) ( * ((u_char *)(p)+1)=(unsigned)(w)/256, *(u_char *)(p)=(unsigned)(w)%256 )
#endif

/* Macro for rounding non-negative real to nearest integer */
#define round(x) (int)((x) + 0.5)

/* Log function for weights etc */
#define lg(x) (double)(log((double)(x))/log(2) * 10.0)
#define alg(x) (int)(pow((double)2.0, (double)((double)x/(double)10.0)) + (double)0.5)
#define rlg(x) round(lg(x))

#define iceil(i,j) (((i)%(j))?(((i)/(j))+1):((i)/(j)))

#ifdef MMPUT

#define mget_vshort(p) \
((p)==NULL)?(-1):(((*(p))&0x80)?(((*(p))&0x7f)<<7)+*((p)+1):*(p))

#define mput_vshort(u,p) \
((u)>16383)?0:(((u)>127)?(*(p)=(((u)>>7)|0x80), *((p)+1)=(u)%128, 2):\
(*(p)=(u), 1))

#endif


/* Length of a precount string */
#define len(s) cv1(s)

/* STRINGS is a slight generalisation of struct GSL */

typedef struct strings {
  int maxnum, 			/* Max no. of strings */
      maxsize,		        /* Max no. of bytes string storage */
      n,			/* Current no. of strings */
      used ;			/* No. of bytes of storage used */
  char **p ;			/* Array of maxnum pointers  */
  char *data ;			/* Base of array of Maxsize bytes */
} STRINGS ;

/* GEN_STORAGE is a struct for storing items which contain a fixed component
   (fixed) and a variable component (variable). An instantiation may only be
   used for one size of item (i.e. fixsize) because individual sizes are not
   stored (this would be the next level of generalization). This struct is a
   generalization of STRINGS (q.v.)

   The first component of the item must be a pointer to the item's variable
   field(s). The struct should be aligned for any type, so that its component
   items can be simply copied back if required. */

typedef struct {
  int maxnum, 			/* Max no. of items */
      maxvar,   	        /* Max no. of bytes variable storage */
      n,			/* Current no. of items */
      used ;			/* No. of bytes of variable storage used */
  int fixsize ;			/* Sizeof item */
  char *fixed ;			/* Base of array of maxnum items of size
				   fixsize*/
  char *variable ;      	/* Base of array of maxvar bytes */
} GEN_STORAGE ;

/* INDEX_TERM is a struct for holding an index term as it is extracted into
   temporary storage. (Added May 91: not finalised) */

struct index_term {
  char *term ;
  unsigned recaddr ;		/* Record's disk address or similar */
  int fbyte ;			/* Only 6 or 8 bits used in short indexes */
  u_char field ;			/* Not used in short indexes */
  u_char sentence ;		/* Not used in short indexes */
  u_char wordnum ;		/* Not used in short indexes */
  u_char gsclass ;		/* May need more info than this, e.g.? */
  short oset ;			/* Offset from start of field of the source
				   of this term (chars) */
  short slen ;			/* Length of source */
} ;

/******************************************************************************

  gsl struct

  Moved here from parms.h Apr 92

******************************************************************************/

typedef struct gsl {
  int num_links ;		/* Number of index structs holding this GSL.
				   (was a BOOL 'live') */
  char name[16] ;		/* Name (e.g. gsl.lisa), gsl source files
				   are always on Control_dir */
  int (* stem_fn) (char *, int, char *) ; /* The stemming function to be used */
  char func_name[16] ;		/* .. and its name */
  int num ;			/* Number of records stored */
  char ** sptrs ;		/* Base of address array */
  char * data ;			/* Base of data */
} GSL ;

/******************************************************************************

  token struct for recording position and length of tokens in text

  Moved from stdeqs.h Dec 92

******************************************************************************/

#define LONGPOS_MAX_F MAXNUMFIELDS
#define LONGPOS_MAX_S 2047
#define LONGPOS_MAX_T 511
#define LONGPOS_MAX_NT 15
#define LONGPOS_MAX_SW 7

#define TEXTPOS_MAX_F MAXNUMFIELDS_TEXT - 1
#define TEXTPOS_MAX_S 32766	/* Note top value (32767) is reserved */
#define TEXTPOS_MAX_T 255
#define TEXTPOS_MAX_NT 15
#define TEXTPOS_MAX_SW 7

#define MAXTEXTPOS 16383
#define MAXLONGPOS 255		/* Greatest number of pos fields in a posting*/

typedef struct {		/* Positional fields for "text" dbs -- HILO
				   processors, type 4, 6 & 8 indexes */
  unsigned f : 2 ;              /* Reluctantly set max fds to 4 to get enough
				   sentences. I suppose this could be max fds
				   contributing to this index, with a mapping
				   of numbers if necessary (May 94) */
  unsigned s : 15 ;		
  unsigned t : 8 ;
  unsigned nt : 4 ;
  unsigned sw : 3 ;
} Pos_text ;

typedef Pos_text Pos_text_hl ;

typedef struct {		/* Positional fields for "text" dbs -- non-HILO
				   processors */
  unsigned sw : 3 ;
  unsigned nt : 4 ;
  unsigned t : 8 ;
  unsigned s : 15 ;		
  unsigned f : 2 ;  
} Pos_text_lh ;

typedef struct {		/* Positional fields for type 5, 7 & 9 indexes*/
  unsigned f : 5 ;
  unsigned s : 11 ;
  unsigned t : 9 ;
  unsigned nt : 4 ;
  unsigned sw : 3 ;
} Pos6 ;

typedef Pos6 Pos6_hl ;

typedef Pos6_hl Longpos_hl ;

typedef struct {
  unsigned sw : 3 ;
  unsigned nt : 4 ;
  unsigned t : 9 ;
  unsigned s : 11 ;
  unsigned f : 5 ;
} Pos6_lh ;

typedef Pos6_lh Longpos_lh ;

typedef struct {
  unsigned s : 16 ;
  unsigned t : 9 ;
  unsigned nt : 4 ;
  unsigned sw : 3 ;
} Pos_no_f_hl ;

typedef struct {
  unsigned sw : 3 ;
  unsigned nt : 4 ;
  unsigned t : 9 ;
  unsigned s : 16 ;
} Pos_no_f_lh ;

typedef union u_pos {
  unsigned pos ;
  Pos_text_hl pthl ;
  Pos_text_lh ptlh ;
  Longpos_hl  lphl ;
  Longpos_lh  lplh ;
  Pos_no_f_hl pnfhl ;
  Pos_no_f_lh pnflh ;
} U_pos ;

/*****************************************************************************

Externs (mostly in Globs.c)

*****************************************************************************/

extern FILE *stdnothing ;
extern int stdzero ;
extern char D_name[] ;		/* Current database name (e.g. -D on command
				   line). Was Sourcefile = Bibliographic file
				   pathname. Altered June 90 */
extern char *Db_names ;		/* Names of available databases for
				   read_db_avail() */
extern int Db_num ;		/* Number of avail db's (ret by
				   read_db_avail() ) */
extern int Maxfiles, Maxmergestreams ;
extern struct db Db ;		/* In globs.c */
extern struct db *Current_db ;	/* This & next perhaps only in bss_globs */
extern char *Current_db_name ;

/* Next two were in parms.h (Apr 92) */
extern struct gsl Gsls[MAXGSLS] ; /* Global gsls. Gsls[0] is by default the
				     one to be used for processing user
				     keyword-type searches (i.e. no field
				     specified) */

extern char Site[] ;		/* Site name: determines printmessage,
				   welcome message etc. */

extern int LWD ; /* Length to compare for Cpwd (utils) */

/* Tables of punctuation and indexing marks */

extern char striptab[] ;  /* Indexing marks stripped for record display */
extern char indtab[] ;    /* Used by phase1() to change mode */
extern char apunct[] ;    /* Used by phase2() */
extern char mpunct[] ;    /* Used by phase2() */

extern char *Control_dir, *Temp_dir, *Bib_dir, *Okapi, *BSS, *Localbib_dir ;

/* Macro replacing isequal() */

#define isequal(s1,s2,l) (! memcmp(s1,s2,l))

#endif _defines_h

