/*
**  indexing.c
**
**  Command line program to:
**
**  1) enter database index parameters
**  2) create Okapi database
**  3) generate indexes
*/

#define G_EXTERN extern

#include "inx-defs.h"

char *read_env(char *);
void set_env(char *, char *);
char *input_string();
void malloc_error(char *, char *, char *, int);
void malloc_success(char *, char *, char *, int);

int read_parameter_files(struct db_param_info *db_info,
													struct search_groups *index_info) {
	int field_count;
	int index_count;
	int field_no;
	int length;

	int return_code = TRUE;

	char field[32];
	char parameter[SMALL_BUFFER];
	char temp[SMALL_BUFFER];

	char db_name[SMALL_BUFFER];
	char bss_parmpath[SMALL_BUFFER];
	char pathname_file[SMALL_BUFFER];
	char parameter_file[SMALL_BUFFER];
	char search_groups_file[SMALL_BUFFER];
	char field_types_file[SMALL_BUFFER];

	char char_buffer;

	FILE *ptf;
	FILE *prf;
	FILE *sgf;
	FILE *ftf;


	printf("READING PARAMETER FILES.\n");
	printf("========================\n\n");

	printf(" BSS_PARMPATH: ");
	sprintf(bss_parmpath, "%s", input_string());

	printf("DATABASE NAME: ");
	sprintf(db_name, "%s", input_string());

	sprintf(pathname_file, "%s/%s.pathnames", bss_parmpath, db_name);
	sprintf(parameter_file, "%s/%s", bss_parmpath, db_name);
	sprintf(search_groups_file, "%s/%s.search_groups", bss_parmpath, db_name);
	sprintf(field_types_file, "%s/%s.field_types", bss_parmpath, db_name);

	if ((ptf = fopen(pathname_file, "r")) == NULL) {
		fprintf(stderr, "Can't open [%s] for writing\n", pathname_file);
		return_code = FALSE;
	}
	else {
		while (!feof(ptf)) {
			if (fgets(temp, SMALL_BUFFER, ptf) == NULL) {
				break;
			}
			else {
				temp[strlen(temp) - 1] = '\0';
				
				if (strlen(temp) > 0) {
					char_buffer = temp;
					length = strchr(char_buffer, ':') - char_buffer;
					memcpy(field, char_buffer, length);
					field[length] = '\0';
					char_buffer += length + 1;
					sprintf(parameter, "%s", char_buffer);

					if (strcmp(field, "installation_root") == 0) {
						sprintf(db_info->installation_root, "%s", parameter);
					}

					if (strcmp(field, "db_source") == 0) {
						sprintf(db_info->db_source, "%s", parameter);
					}

					if (strcmp(field, "bss_parmpath") == 0) {
						sprintf(db_info->bss_parmpath, "%s", parameter);
					}

					if (strcmp(field, "exchange_format_filename") == 0) {
						sprintf(db_info->exchange_format_filename, "%s", parameter);
					}
				}
			}
		}
	}

	fclose(ptf);

	if ((prf = fopen(parameter_file, "r")) == NULL) {
		fprintf(stderr, "Can't open [%s] for writing\n", parameter_file);
		return_code = FALSE;
	}
	else {
		while (!feof(prf)) {
			if (fgets(temp, SMALL_BUFFER, prf) == NULL) {
				break;
			}
			else {
				temp[strlen(temp) - 1] = '\0';
				
				if (strlen(temp) > 0) {
					char_buffer = temp;
					length = strchr(char_buffer, ':') - char_buffer;
					memcpy(field, char_buffer, length);
					field[length] = '\0';
					char_buffer += length + 1;
					sprintf(parameter, "%s", char_buffer);
					
					if (strcmp(field, "installation_root") == 0) {
						sprintf(db_info->installation_root, "%s", parameter);
					}

					if (strcmp(field, "db_source") == 0) {
						sprintf(db_info->db_source, "%s", parameter);
					}

					if (strcmp(field, "bss_parmpath") == 0) {
						sprintf(db_info->bss_parmpath, "%s", parameter);
					}

					if (strcmp(field, "exchange_format_filename") == 0) {
						sprintf(db_info->exchange_format_filename, "%s", parameter);
					}

					if (strcmp(field, "name") == 0) {
						sprintf(db_info->name, "%s", parameter);
					}

					if (strcmp(field, "lastbibvol") == 0) {
						db_info->lastbibvol = atoi(parameter);
					}

					if (strcmp(field, "bib_basename") == 0) {
						sprintf(db_info->bib_basename, "%s", parameter);
					}

					if (strcmp(field, "bib_dir") == 0) {
						sprintf(db_info->bib_dir, "%s", parameter);
					}

					if (strcmp(field, "bibsize") == 0) {
						db_info->bibsize = atoi(parameter);
					}

					if (strcmp(field, "real_bibsize") == 0) {
						db_info->real_bibsize = atoi(parameter);
					}

					if (strcmp(field, "display_name") == 0) {
						sprintf(db_info->display_name, "%s", parameter);
					}

					if (strcmp(field, "explanation") == 0) {
						sprintf(db_info->explanation, "%s", parameter);
					}

					if (strcmp(field, "nr") == 0) {
						db_info->nr = atoi(parameter);
					}

					if (strcmp(field, "nf") == 0) {
						db_info->nf = atoi(parameter);
						field_count = 0;
					}

					if (strcmp(field, "f_abbrev") == 0) {
						sprintf(db_info->f_abbrev[field_count], "%s", parameter);
						field_count++;
					}

					if (strcmp(field, "rec_mult") == 0) {
						db_info->rec_mult = atoi(parameter);
					}

					if (strcmp(field, "fixed") == 0) {
						db_info->fixed = atoi(parameter);
					}

					if (strcmp(field, "db_type") == 0) {
						sprintf(db_info->db_type, "%s", parameter);
					}

					if (strcmp(field, "has_lims") == 0) {
						db_info->has_lims = atoi(parameter);
					}

					if (strcmp(field, "maxreclen") == 0) {
						db_info->maxreclen = atoi(parameter);
					}

					if (strcmp(field, "ni") == 0) {
						db_info->ni = atoi(parameter);
						index_count = -1;
					}

					if (strcmp(field, "last_ixvol") == 0) {
						/*
						**  next index field
						*/

						index_count++;
						db_info->last_ixvol[index_count] = atoi(parameter);
					}

					if (strcmp(field, "ix_stem") == 0) {
						sprintf(db_info->ix_stem[index_count], "%s", parameter);
					}

					if (strcmp(field, "ix_volsize") == 0) {
						db_info->ix_volsize[index_count] = atoi(parameter);
					}

					if (strcmp(field, "ix_type") == 0) {
						db_info->ix_type[index_count] = atoi(parameter);
					}
				}
			}
		}
	}

	fclose(prf);

	if ((ftf = fopen(field_types_file, "r")) == NULL) {
		fprintf(stderr, "Can't open [%s] for writing\n", field_types_file);
		return_code = FALSE;
	}
	else {
		while (!feof(ftf)) {
			if (fgets(temp, SMALL_BUFFER, ftf) == NULL) {
				break;
			}
			else {
				temp[strlen(temp) - 1] = '\0';

				if (strlen(temp) > 0) {
					char_buffer = temp;
					sscanf(char_buffer, "%d %s", &field_no, parameter);

					if (field_no > 0 && field_no < db_info->nf + 1) {
						sprintf(db_info->field_name[field_no - 1], "%s", parameter);
					}
				}
			}
		}
	}

	fclose(ftf);

	index_count = 0;

	if ((sgf = fopen(search_groups_file, "r")) == NULL) {
		fprintf(stderr, "Can't open [%s] for writing\n", search_groups_file);
		return_code = FALSE;
	}
	else {
		while (!feof(sgf)) {
			if (fgets(temp, SMALL_BUFFER, prf) == NULL) {
				break;
			}
			else {
				temp[strlen(temp) - 1] = '\0';

				if (strlen(temp) > 0) {
					char_buffer = temp;
					sscanf(char_buffer, "%s %d %d %s %s %s %s",
								 (index_info + index_count)->index_name,
								 (index_info + index_count)->v1,
								 (index_info + index_count)->v2,
								 (index_info + index_count)->keyword_type,
								 (index_info + index_count)->stemming_function,
								 (index_info + index_count)->gsl_fname,
								 (index_info + index_count)->index_fields);

					index_count++;
				}
			}
		}
	}

	fclose(sgf);

	return return_code;
}


int write_parameter_files(struct db_param_info *db_info,
													struct search_groups *index_info) {
	int nf;
	int ni;

	int return_code = TRUE;

	char pathname_file[SMALL_BUFFER];
	char parameter_file[SMALL_BUFFER];
	char search_groups_file[SMALL_BUFFER];
	char field_types_file[SMALL_BUFFER];

	FILE *ptf;
	FILE *prf;
	FILE *sgf;
	FILE *ftf;

	sprintf(pathname_file, "%s/%s.pathnames",
					db_info->bss_parmpath, db_info->name);

	sprintf(parameter_file, "%s/%s",
					db_info->bss_parmpath, db_info->name);

	sprintf(search_groups_file, "%s/%s.search_groups",
					db_info->bss_parmpath, db_info->name);

	sprintf(field_types_file, "%s/%s.field_types",
					db_info->bss_parmpath, db_info->name);

	if ((ptf = fopen(pathname_file, "w")) == NULL) {
		fprintf(stderr, "Can't open [%s] for writing\n", pathname_file);
		return_code = FALSE;
	}
	else if ((prf = fopen(parameter_file, "w")) == NULL) {
		fprintf(stderr, "Can't open [%s] for writing\n", parameter_file);
		return_code = FALSE;
	}
	else if ((sgf = fopen(search_groups_file, "w")) == NULL) {
		fprintf(stderr, "Can't open [%s] for writing\n", search_groups_file);
		return_code = FALSE;
	}
	else if ((ftf = fopen(field_types_file, "w")) == NULL) {
		fprintf(stderr, "Can't open [%s] for writing\n", field_types_file);
		return_code = FALSE;
	}
	else {
		fprintf(ptf, "installation_root=%s\n", db_info->installation_root);
		fprintf(ptf, "db_source=%s\n", db_info->db_source);
		fprintf(ptf, "bss_parmpath=%s\n", db_info->bss_parmpath);
		fprintf(ptf, "exchange_format_filename=%s\n",
						db_info->exchange_format_filename);

		fclose (ptf);

		fprintf(prf, "name=%s\n", db_info->name);
		fprintf(prf, "lastbibvol=%d\n", db_info->lastbibvol);
		fprintf(prf, "bib_basename=%s\n", db_info->bib_basename);
		fprintf(prf, "bib_dir=%s\n", db_info->bib_dir);
		fprintf(prf, "bibsize=%d\n", db_info->bibsize);
		fprintf(prf, "real_bibsize=%d\n", db_info->real_bibsize);
		fprintf(prf, "display_name=%s\n", db_info->display_name);
		fprintf(prf, "explanation=%s\n", db_info->explanation);
		fprintf(prf, "nr=%d\n", db_info->nr);
		fprintf(prf, "nf=%d\n", db_info->nf);
		
		for (nf = 0; nf < db_info->nf; nf++) {
			fprintf(prf, "f_abbrev=%s\n", db_info->f_abbrev[nf]);
			fprintf(ftf, "%d %s\n", nf + 1, db_info->field_name[nf]);
		}
		
		fprintf(prf, "rec_mult=%d\n", db_info->rec_mult);
		fprintf(prf, "fixed=%d\n", db_info->fixed);
		fprintf(prf, "db_type=%s\n", db_info->db_type);
		fprintf(prf, "has_lims=%d\n", db_info->has_lims);
		fprintf(prf, "maxreclen==%d\n", db_info->maxreclen);
		fprintf(prf, "ni=%d\n", db_info->ni);
		
		for (ni = 0; ni < db_info->ni; ni++) {
			fprintf(prf, "last_ixvol=%d\n", db_info->last_ixvol[ni]);
			fprintf(prf, "ix_stem=%s\n", db_info->ix_stem[ni]);
			fprintf(prf, "ix_volsize=%d\n", db_info->ix_volsize[ni]);
			fprintf(prf, "ix_type=%d\n", db_info->ix_type[ni]);

			fprintf(sgf, "%s %d %d %s %s %s %s -1\n",
							(index_info + ni)->index_name,
							(index_info + ni)->v1, (index_info + ni)->v2,
							(index_info + ni)->keyword_type,
							(index_info + ni)->stemming_function,
							(index_info + ni)->gsl_fname,
							(index_info + ni)->index_fields);
		}
		
		fclose(prf);
		fclose(ftf);
		fclose(sgf);
	}

	return return_code;
}


int setup_parameters(struct db_param_info *db_info,
										 struct search_groups *index_info) {
	/*
	**  During this function the data will be written to two
	**  files:
	**
	**  1) <installation_root>/param_data/db_info.file
	**  2) <installation_root>/param_data/search_groups.file
	**
	**  If either the database creation or indexing process
	**  fails the program can be rerun and these files re-read,
	**  edited if necessary, and then the database creation /
	**  indexing process re-started
	*/

	int debugging = TRUE;
	int return_code = TRUE;

	int nf;
	int ni;

	char ip_buffer[K_BUFFER];
	char env_variable[SMALL_BUFFER];
	char env_value[SMALL_BUFFER];

	char db_avail_file[SMALL_BUFFER];
	FILE *dba;

	/*
	**  Enter database parameters
	*/
	
	printf("Setting Database Index Parameters.\n\n");
	
	printf("The four directories:\n\n");
	printf("1. The installation directory (OKAPI_ROOT).\n");
	printf("2. The directory of the exchange format data file ");
	printf("   (DB_SOURCE).\n");
	printf("3. The location of the database parameter files ");
	printf("   (BSS_PARMPATH).\n");
	printf("4. The location of the database files and indexes ");
	printf("   (BIBFILES_DIR).\n\n");
	printf("need to be specified as full pathnames.\n\n");
		
	printf("1. OKAPI_ROOT: ");

	sprintf(db_info->installation_root, "%s", input_string());

	if (debugging) {
		fprintf(stderr, "OKAPI ROOT DIRECTORY = [%s]\n",
						db_info->installation_root);
	}

	printf("2. DB_SOURCE: ");
	sprintf(db_info->db_source, "%s", input_string());
		
	if (debugging) {
		fprintf(stderr,
						"DATABASE SOURCE DIRECTORY = [%s]\n", db_info->db_source);
	}
		
	printf("3. BSS_PARMPATH: ");
	sprintf(db_info->bss_parmpath, "%s", input_string());
		
	if (debugging) {
		printf("BSS_PARMPATH DIRECTORY = [%s]\n", db_info->bss_parmpath);
	}
		
	sprintf(env_variable, "BSS_PARMPATH");
	set_env(env_variable, db_info->bss_parmpath);
	
	printf("4. BIBFILES_DIR: ");
	sprintf(db_info->bib_dir, "%s", input_string());
	
	if (debugging) {
		fprintf(stderr, "BIBFILES DIRECTORY = [%s]\n", db_info->bib_dir);
	}
	
	/*
	**  Make sure that <bss_parmpath>/db_avail can be appended to
	*/
	
	sprintf(db_avail_file, "%s/db_avail", db_info->bss_parmpath);
	
	if ((dba = fopen(db_avail_file, "a")) == NULL) {
		fprintf(stderr, "Can't open [%s] for appending\n", db_avail_file);
		return_code = FALSE;
	}
	else {
		printf("DATABASE NAME: ");
		sprintf(db_info->name, "%s", input_string());

		/*
		**  Write database name to end of <bss_parmpath>/db_avail
		*/

		fprintf(dba, "%s *\n", db_info->name);
		fclose(dba);

		printf("ENTER EXCHANGE FORMAT FILENAME: ");
		sprintf(db_info->exchange_format_filename, "%s", input_string());
		
		db_info->lastbibvol = 0;
		
		sprintf(db_info->bib_basename, "%s.bib", db_info->name);
		
		db_info->bibsize = 2047;
		db_info->real_bibsize = 104857600;
		
		printf("DATABASE DISPLAY_NAME: ");
		
		sprintf(db_info->display_name, "%s", input_string());
		
		printf("DESCRIPTION OF DATABASE: ");
		
		sprintf(ip_buffer, "%s", input_string());
		
		if (strlen(ip_buffer) == 0) {
			sprintf(db_info->explanation, "%s", db_info->display_name);
		}
		else {
			sprintf(db_info->explanation, "%s", ip_buffer);
		}
		
		db_info->nr = 10240;
		
		printf("NUMBER OF FIELDS PER RECORD: ");
		
		sprintf(ip_buffer, "%s", input_string());
		db_info->nf = atoi(ip_buffer);
		
		if (db_info->nf <= 0) {
			db_info->nf = 1;
			printf("There must be at least one field\n");
		}
		
		for (nf = 0; nf < db_info->nf; nf++) {
			printf("FIELD%3d -- ABBREVIATION: ", nf);
			sprintf(db_info->f_abbrev[nf], "%s", input_string());
			printf("              FIELD NAME: ", nf);
			sprintf(db_info->field_name[nf], "%s", input_string());
		}
		
		db_info->rec_mult = 4;  /* Allows dbs upto 8 Gb */
		db_info->fixed = 0;
		db_info->has_lims = 0;
		
		printf("DATABASE TYPE (t = text, a = ai): ");
		
		sprintf(db_info->db_type, "text");
		
		sprintf(ip_buffer, "%s", input_string());
		
		if (tolower(ip_buffer[0] == 'a')) {
			sprintf(db_info->db_type, "ai");
		}
		
		/*
		**  Set db_info->ix_type
		**
		**  text = 8, ai = 9
		*/
		
		db_info->maxreclen = 20 * SMALL_BUFFER;
		
		printf("NUMBER OF INDEXES: ");
		
		sprintf(ip_buffer, "%s", input_string());
		db_info->ni = atoi(ip_buffer);
		
		if (db_info->ni > 0) {
			printf("\nFOR EACH INDEX YOU MUST PROVIDE:\n\n");
			printf("1. INDEX NAME.\n");
			printf("2. KEYWORD TYPE.\n");
			printf("3. STEMMING FUNCTION\n");
			printf("4. GSL FILENAME (IN [%s] -- PRESS RETURN FOR NONE)\n",
						 db_info->bss_parmpath);
			printf("5. A LIST OF 0 SEPARATED FIELDS (ANY ORDER) TO BE INDEXED.\n\n");
			printf("   E.g. 2 0 3 0 5 0 7 0 4 0\n");
			
			for (ni = 0; ni < db_info->ni; ni++) {
				db_info->last_ixvol[ni] = 0;
				
				sprintf(db_info->ix_stem[ni],
								"%s/%s",
								db_info->bib_dir, db_info->name);
				
				db_info->ix_volsize[ni] = 2047;
				db_info->ix_type[ni] = (strcmp(db_info->db_type, "ai") == 0) ? 9 : 8;
				
				
				printf("INDEX NAME: ");
				sprintf((index_info + ni)->index_name, "%s", input_string());
				
				(index_info + ni)->v1 = 0;
				(index_info + ni)->v2 = ni;
				
				printf("KEYWORD TYPE: ");
				sprintf((index_info + ni)->keyword_type, "%s", input_string());
				
				printf("STEMMING FUNCTION: ");
				sprintf((index_info + ni)->stemming_function, "%s", input_string());
				
				printf("GSL FILENAME ('empty' IF NONE : ");
				sprintf(ip_buffer, "%s", input_string());
				
				if (strcmp(ip_buffer, "empty") == 0) {
					sprintf((index_info + ni)->gsl_fname, "gsl.empty");
				}
				else {
					sprintf((index_info + ni)->gsl_fname, "%s", ip_buffer);
				}
				
				printf("ZERO SEPARATED LIST OF INDEX FIELDS: ");
				sprintf((index_info + ni)->index_fields, "%s", input_string());
			}
		}
		
		if (return_code) {
			return_code =  write_parameter_files(db_info, index_info);
		}
	}

	return return_code;
}

int convert_db(struct db_param_info *db_info) {
	int debugging = TRUE;
	int return_code = TRUE;

	int norecs = 100;
	char os_command[SMALL_BUFFER];

	char env_variable[SMALL_BUFFER];

	sprintf(env_variable, "BSS_PARMPATH");
	set_env(env_variable, db_info->bss_parmpath);

	/*
	**  convert_runtime is called as:
	**
	**    convert_runtime -c <BSS_PARMPATH> <db_name> < <db_source>/<exchf>
	**
	**  where <exchf> is the exchange format filename
	*/

	sprintf(os_command,
					"%s/bin/convert_runtime -c %s %s < %s/%s",
					db_info->installation_root,
					db_info->bss_parmpath,
					db_info->name,
					db_info->db_source,
					db_info->exchange_format_filename);

	if (debugging) {
		fprintf(stderr, "%s\n", os_command);
	}

	if (system(os_command) < 0) {
		perror(os_command);
		return_code = FALSE;
	}

	return return_code;
}

int make_index(struct db_param_info *db_info) {
	int debugging = TRUE;
	int return_code = TRUE;
	int ni;

	char os_command[SMALL_BUFFER];
	char env_variable[SMALL_BUFFER];

	sprintf(env_variable, "BSS_PARMPATH");
	set_env(env_variable, db_info->bss_parmpath);

	for (ni = 0; ni < db_info->ni; ni++) {
		sprintf(os_command,
						"%s/bin/ix1 -c %s %s %d | %s/bin/ixf -c %s %s %d",
						db_info->installation_root,
						db_info->bss_parmpath,
						db_info->name,
						ni,
						db_info->installation_root,
						db_info->bss_parmpath,
						db_info->name,
						ni);

		if (debugging) {
			fprintf(stderr, "%s\n", os_command);
		}

		if (system(os_command) < 0) {
			perror(os_command);
			return_code = FALSE;
		}

		if (!return_code) {
			break;
		}
	}

	return return_code;
}


int create_db_index(int argc, char *argv[]) {
	int debugging = TRUE;
	
	int buffer_size;
	
	char installation_root[SMALL_BUFFER];
	char db_source[SMALL_BUFFER];
	char bss_parmpath[SMALL_BUFFER];
	char bib_dir[SMALL_BUFFER];

	struct db_param_info *db_info;
	struct search_groups *index_info;

	buffer_size = sizeof(struct db_param_info);
	
	if ((db_info = (struct db_param_info *) malloc(buffer_size)) == NULL){
		malloc_error("params.c", "main", "db_info", buffer_size);
	}	
	
	buffer_size = (MAX_INDEXES + 1) * sizeof(struct search_groups);
	
	if ((index_info = (struct search_groups *) malloc(buffer_size)) == NULL) {
		malloc_error("params.c", "main", "search_groups", buffer_size);
	}	
	
	if (setup_parameters(db_info, index_info)) {
		if (write_parameter_files(db_info, index_info)) {
			printf("Parameter file written OK\n");

			convert_db(db_info);
			make_index(db_info);
		}
		else {
			printf("ERROR writing parameter file.\n");
		}
	}
	else {
		printf("ERROR setting up parameters.\n");
	}
}

