/*
**  interface.cc
**
**  mg - 14 Jan 1998
*/

#define G_EXTERN extern

#include "inx-defs.h"
#include "GLOB.h"

#define MAX_TCL_LINES 256

#define STR_OKAPI_TITLE "CISR - Indexing Interface"
#define STR_OKAPI_DEL "{ puts \"Close is disable... Return to continue. \" }"
#define STR_OKAPI_POSITION  " +200+80"

G_EXTERN int GetAndCheckVars();
G_EXTERN int Do_AppDeConstruct();

G_EXTERN "C" {
  void exit();
  int isatty();
  int read();
  char *strrchr();
}


/*
**  The main window for the application.
**  If NULL then the application no longer exists
*/

static Tk_Window enqWD;

/*
**  Used to assemble lines of terminal input into Tcl tclcmds
*/

static Tcl_DString tclcmd;

/*
**  Non-zero means standard input is a terminal-like device.
**  Zero means it's a file.
*/

static int tty;

static char errorExitCmd[] = "exit 1";

/*
** Command-line options:
*/

static int synchronize = 0;
static char *fileName = NULL;
static char *name = NULL;
static char *display = NULL;
static char *geometry = NULL;
static char *db_name = NULL;

static Tk_ArgvInfo argTable[] = {
  {
    "-file", TK_ARGV_STRING, (char *) NULL, (char *) &fileName,
    "File from which to read commands"
  },
  {
    "-geometry", TK_ARGV_STRING, (char *) NULL, (char *) &geometry,
    "Initial geometry for window"
  },
  {
    "-display", TK_ARGV_STRING, (char *) NULL, (char *) &display,
    "Display to use"
  },
  {
    "-name", TK_ARGV_STRING, (char *) NULL, (char *) &name,
    "Name to use for application"
  },
  {
    "-db", TK_ARGV_CONSTANT, (char *) 1, (char *) &db_name,
    "Use synchronous mode for display server"
  },
  {
    "-sync", TK_ARGV_CONSTANT, (char *) 1, (char *) &synchronize,
    "Use synchronous mode for display server"
  },
  {
    (char *) NULL, TK_ARGV_END, (char *) NULL, (char *) NULL,
    (char *) NULL
  }
};

/*
**  Forward declarations for procedures defined later in this file:
*/

static void Prompt _ANSI_ARGS_((Tcl_Interp *interp, int partial));
static void StdinProc _ANSI_ARGS_((ClientData clientData, int mask));
G_EXTERN int OKAPI_TclInit(Tcl_Interp *theInterp);
G_EXTERN void Enquire_LogClient(char *);

void usage(char progname[]) {
	printf("\n\n*******************************************************\n");
	printf("**                                                   **\n");
	printf("**  Insufficent parameters passed to the program.    **\n");
	printf("**                                                   **\n");
  printf("**  Please rerun as:                                 **\n");
	printf("**                                                   **\n");
  printf("**  %s <topic_no> <user_code>                    **\n",
				 progname);
	printf("**                                                   **\n");
	printf("*******************************************************\n\n");
  exit(0);
}


void no_display()
{
	printf("\n\n*******************************************************\n");
	printf("**                                                   **\n");
	printf("**  There is no DISPLAY environment variable set.    **\n");
	printf("**                                                   **\n");
	printf("**  Please set it as, e.g.                           **\n");
	printf("**                                                   **\n");
	printf("**       setenv DISPLAY pace.is.ac.uk:0.0            **\n");
	printf("**                                                   **\n");
	printf("**  and then re-run the program.                     **\n");
	printf("**                                                   **\n");
	printf("*******************************************************\n\n");
}

void no_topicno_file(char trec_topics[])
{
	printf("\n\n*******************************************************\n");
	printf("**                                                   **\n");
	printf("**          There is no file [%s] in       **\n",
				 trec_topics);
	printf("**               your home directory.                **\n");
	printf("**                                                   **\n");
	printf("**      This file must exist and contain a list      **\n");
	printf("**        of valid topic numbers, 1 per line.        **\n");
	printf("**                                                   **\n");
	printf("*******************************************************\n\n");
	exit(0);
}



/*
**  Main program
*/

int main(int argc, char *argv[])
{
	int run_no = 0;

	int topic;
	int no_trec_topics = 0;
	int topic_ok = FALSE;

  int code;
  int i;
	int next_profile;
	
  char *args;
  char *parameter;
  char *message;
  char buf[20];

	char os_command[SMALL_BUFFER];

  char filename[SMALL_BUFFER];
	char dbname[SMALL_BUFFER];
	char temp[SMALL_BUFFER];
	char userid[32];

	/*
	char *fullName;
	FILE *f;
	*/

  FILE *fp;

	/*
	**  Check that the DISPLAY variable has been set.
	**  If not display error message and terminate the program.
	*/
	
	if (getenv("DISPLAY") == NULL) {
		no_display();
		exit(0);
	}
	
	/*
	**  Set up name of LOGS_DIR and create it if necessary.
	**
	**  1. Create LOGFILES_ROOTDIR
	*/

	/*
	printf("OKAPI_ROOT = [%s]\n", getenv("OKAPI_ROOT"));
	printf("Tcl scripts dir = [%s]\n", getenv("OKAPI_TCL_SCRIPTS_DIR"));
	*/

	sprintf(LOGS_DIR, "%s", getenv("INDEXER_LOGS"));
	sprintf(os_command, "mkdir %s", LOGS_DIR);

	if (system(os_command) < 0) {
		perror(os_command);
	}

  theInterp = Tcl_CreateInterp();
	
#ifdef TCL_MEM_DEBUG
  Tcl_InitMemory(theInterp);
#endif
	
  /*
  **  Parse command-line arguments.
  */
	
  if (Tk_ParseArgv(theInterp, (Tk_Window) NULL, &argc, argv, argTable, 0)
      != TCL_OK) {
    fprintf(stderr, "%s\n", theInterp->result);
    exit(1);
  }

  /*
  **  Thien's code for selecting initialisation parameters
  **  from the command line arguments.
  */
	
	switch (argc) {
  	case 1 :
			sprintf(DB_NAME, "NONE");
			break;
  	case 2 :
			sprintf(DB_NAME, "%s", argv[1]);
			break;
  	default :
			usage(argv[0]);
			break;
	}

	if (db_name != NULL) {
		printf("DB_NAME = [%s]\n", db_name);
	}
	else {
    parameter = argv[0];
	}
		
  name = strrchr(parameter, '/');
  
  if (name != NULL) {
    name++;
  }
  else {
    name = parameter;
  }
	
  /*
  **  If a display was specified, put it into the DISPLAY
  **  environment variable so that it will be available for
  **  any sub-processes created by us.
  */

  
  if (display != NULL) {
    Tcl_SetVar2(theInterp, "env", "DISPLAY", display, TCL_GLOBAL_ONLY);
  }
	
  /*
  **  Initialize the Tk application.
  */
	
  enqWD = TkCreateMainWindow(theInterp, display, name, "Tk");

  if (enqWD == NULL) {
    fprintf(stderr, "%s\n", theInterp->result);
    exit(1);
  }
	
  if (synchronize) {
    XSynchronize(Tk_Display(enqWD), True);
  }
	
  Tk_GeometryRequest(enqWD, 800, 750);
	
  /*
  **  Make command-line arguments available in the Tcl variables
  **  "argc" and "argv".  Also set the "geometry" variable from
  **  the geometry specified on the command line.
  */

  args = Tcl_Merge(argc - 1, argv + 1);
  Tcl_SetVar(theInterp, "argv", args, TCL_GLOBAL_ONLY);
  ckfree(args);
  sprintf(buf, "%d", argc-1);
	
  Tcl_SetVar(theInterp, "argc", buf, TCL_GLOBAL_ONLY);
  Tcl_SetVar(theInterp, "argv0",
						 (fileName != NULL) ? fileName : argv[0], TCL_GLOBAL_ONLY);
	
  if (geometry != NULL) {
    Tcl_SetVar(theInterp, "geometry", geometry, TCL_GLOBAL_ONLY);
  }
	
  /*
  **  Set the "tcl_interactive" variable.
  */
	
  tty = isatty(0);
  Tcl_SetVar(theInterp, "tcl_interactive",
						 ((fileName == NULL) && tty) ? "1" : "0", TCL_GLOBAL_ONLY);
	
  /*
  **  Invoke application-specific initialization.
  */
	
  if (OKAPI_TclInit(theInterp) != TCL_OK) {
    fprintf(stderr, "Tcl_AppInit failed: %s\n", theInterp->result);
  }
	
  /*
  ** Invoke the script specified on the command line, if any.
  */
	
  if (fileName != NULL) {
    /*
    **  code = Tcl_VarEval(theInterp, "source ", fileName, (char *) NULL);
    */
		
    /*
		**  code = Tcl_VarEval(theInterp, "source ", TOPIC_NAME, (char *) NULL);
		*/
		

		if (code != TCL_OK) {
			goto error;
    }
    else {
      printf("theInterp->result [%s]\n", theInterp->result);
    }
		
    tty = 0;
  }
  else {
    /*
    **  Commands will come from standard input, so set up an event
    **  handler for standard input.  If the input device is aEvaluate the
    **  .rc file, if one has been specified, set up an event handler
    **  for standard input, and print a prompt if the input
    **  device is a terminal.
    */

    if (tcl_RcFile != NULL) {
			/*
      Tcl_DString buffer;
			*/

      char *fullName;
			FILE *f;

			/*
			fullName = Tcl_TildeSubst(theInterp, tcl_RcFile, &buffer);
			*/

			fullName = tcl_RcFile;
			printf("Configuration file [%s]\n", fullName);
			
      if (fullName == NULL) {
				fprintf(stderr, "%s\n", theInterp->result);
      }
      else {
				f = fopen(fullName, "r");
				if (f != NULL) {
					code = Tcl_EvalFile(theInterp, fullName);
					
					if (code != TCL_OK) {
						fprintf(stderr, "%s\n", theInterp->result);
					}
					else {
						fprintf(stderr, "%s\n", theInterp->result);
					}
					
					fclose(f);
				}
      }
			
			/*
      Tcl_DStringFree(&buffer);
			*/
    }


		
    Tk_CreateFileHandler(0, TK_READABLE, StdinProc, (ClientData) 0);
		
    if (tty) {
      Prompt(theInterp, 0);
    }
  }
	
  fflush(stdout);
  Tcl_DStringInit(&tclcmd);
	
	if (strcmp(DB_NAME, "NONE") != 0) {
		/*
		**  Read parameter files
		*/
	}
	
  /*
  **  Run-time position
	**

	Tk_MoveWindow(enqWD, FULL_WINDOW_OFFSET_X, FULL_WINDOW_OFFSET_Y);

	**
	*/

	/*
	**  Loop infinitely, waiting for commands to execute.
	**  When there are no windows left, Tk_MainLoop returns and we exit.
	*/

  if (GetAndCheckVars() == TCL_OK) {
    /*
		**  initialisation();
		**
    **  profile_selection();
    */

    Tk_MainLoop();
  }

  /*
  **  Don't exit directly, but rather invoke the Tcl "exit" command.
  **  This gives the application the opportunity to redefine "exit"
  **  to do additional cleanup.
  */

  Tcl_Eval(theInterp, "exit");
  exit(1);

error:
  message = Tcl_GetVar(theInterp, "errorInfo", TCL_GLOBAL_ONLY);

  if (message == NULL) {
    message = theInterp->result;
  }

  fprintf(stderr, "%s\n", message);
  Tcl_Eval(theInterp, errorExitCmd);

  /*
  **  Need 'return 1' only to prevent compiler warnings.
  */
  return 1;
}


int GetAndCheckVars ()
{
  /*
  **  GetAndCheckVars() --
  **  
  **  Needed Vars to communicate between the Internal and External works
  */

  return TCL_OK;
}

/*
**   ARGSUSED
*/

static void StdinProc(ClientData clientData, int mask) /* Not used. */
{
  /*
  **  StdinProc --
  **
  **  This procedure is invoked by the event dispatcher whenever
  **  standard input becomes readable.  It grabs the next line of
  **  input characters, adds them to a command being assembled,
  **  and executes the command if it's complete.
  **
  **  Results:  NONE.
  **
  **  Side effects:  Could be almost arbitrary, depending on
  **                 the command that's typed.
  */

  char input[TCL_BUFFER_SIZE];
  static int gotPartial = 0;
  char *cmd;
  int code, count;
  
  count = read(fileno(stdin), input, TCL_BUFFER_SIZE);
  
  if (count <= 0) {
    if (!gotPartial) {
      if (tty) {
				Tcl_Eval(theInterp, "exit");
				exit(1);
      }
      else {
				Tk_DeleteFileHandler(0);
      }
      return;
    }
    else {
      count = 0;
    }
  }

  cmd = Tcl_DStringAppend(&tclcmd, input, count);

  if (count != 0) {
    if ((input[count-1] != '\n') && (input[count-1] != ';')) {
      gotPartial = 1;
      goto prompt;
    }

    if (!Tcl_CommandComplete(cmd)) {
      gotPartial = 1;
      goto prompt;
    }
  }

  gotPartial = 0;
  
  /*
  **  Disable the stdin file handler while evaluating the command;
  **  otherwise if the command re-enters the event loop we might
  **  process commands from stdin before the current command is
  **  finished.  Among other things, this will trash the text of the
  **  command being evaluated.
  */
  
  Tk_CreateFileHandler(0, 0, StdinProc, (ClientData) 0);
  
  if (strncmp(cmd, "exit", 4)== 0) {
    Do_AppDeConstruct();
  }

  code = Tcl_RecordAndEval(theInterp, cmd, 0);
  Tk_CreateFileHandler(0, TK_READABLE, StdinProc, (ClientData) 0);
  Tcl_DStringFree(&tclcmd);

  if (*theInterp->result != 0) {
    if ((code != TCL_OK) || (tty)) {
      printf("%s\n", theInterp->result);
    }
  }

  /*
  **  Output a prompt.
  */
  
 prompt:
  if (tty) {
    Prompt(theInterp, gotPartial);
  }
}



static void Prompt(Tcl_Interp *interp, int partial)
{
  /*
  **  Prompt
  **
  **  Issue a prompt on standard output, or invoke a script
  **  to issue the prompt.
  **
  **  Results:  NONE.
  **
  **  Side effects:  A prompt gets output, and a Tcl script
  **                 may be evaluated in interp.
  */

  char *promptCmd;
  int code;
  
  promptCmd = Tcl_GetVar(interp, partial ? "tcl_prompt2" :
			 "tcl_prompt1", TCL_GLOBAL_ONLY);
  
  if (promptCmd == NULL) {
  defaultPrompt:
    if (!partial) {
      fputs("% ", stdout);
    }
  }
  else {
    code = Tcl_Eval(interp, promptCmd);

    if (code != TCL_OK) {
      Tcl_AddErrorInfo(interp,
		       "\n    (script that generates prompt)");
      fprintf(stderr, "%s\n", interp->result);
      goto defaultPrompt;
    }
  }
  fflush(stdout);
}




