/*
**  inx-defs.h -- 15 Jan 1998
*/

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <malloc.h>
#include <signal.h>
#include <stdlib.h>
#include <defines.h>
#include <math.h>
#include <sys/time.h>

/*
**  Libraries for reading directory information
**  with readdir() and examining it with stat()
*/

#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>

#include <tcl.h>
#include <tk.h>

#include <errno.h>

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

#ifndef NULL
#define NULL 0
#endif

#ifndef SEEK_SET
#define SEEK_SET 0
#endif

/*
**  Parameters for array sizes for various objects
*/

#define SMALL_BUFFER 256
#define K_BUFFER 1024
#define OS_COMMAND_BUFFER_SIZE 1024
#define TCL_BUFFER_SIZE (4 * 1024)

#define MAX_FIELDS 16
#define MAX_INDEXES 16

G_EXTERN char DB_NAME[SMALL_BUFFER];
G_EXTERN char LOGS_DIR[SMALL_BUFFER];

/*
**  NTD's defs for tk programs.
**
**  Name of a user-specific startup script
*/

G_EXTERN char tcl_RcFile[SMALL_BUFFER];

#ifndef ENQ_DEC_DOT_H
#define ENQ_DEC_DOT_H

/*
**  XXX
*/

EXTERN int OKAPI_Init();
EXTERN int OKAPI_Exit();

/*
**  QUERY
*/

/*
**  EXTERN int QM_Init();
**  EXTERN int QM_Exit();
*/

/*
**  ENQ_UTILS
*/

EXTERN int UTILS_Init();
EXTERN int UTILS_Exit();

/*
**  ENQUIRE
*/

EXTERN int Do_AppDeConstruct();

#endif

/*
**  End of NTD's defs
*/

/*
**  Structure definition for database parameter info
*/

typedef struct db_param_info {
	char installation_root[SMALL_BUFFER];
	char db_source[SMALL_BUFFER];
	char bss_parmpath[SMALL_BUFFER];
	char exchange_format_filename[SMALL_BUFFER];

	char name[SMALL_BUFFER];

	int lastbibvol;
	char bib_basename[SMALL_BUFFER];  /*  <db_name>.bib  							*/
	char bib_dir[SMALL_BUFFER];       /*
																		**  Ex: /project/okapi/bibfiles/<db_name>
																		*/
	int bibsize;                      /*  Disk req'ments in MBytes         */
	int real_bibsize;                 /*
																		**  Actual bibsize -
																		**  Set by convert_runtime program
																		*/

	char display_name[SMALL_BUFFER];
	char explanation[SMALL_BUFFER];
	int nr;                           /*
																		**  Set dummy value of 10240; real
																		**  value set by convert_runtime
																		*/
	int nf;                           /*  Number of fields                    */
	char f_abbrev[MAX_FIELDS][SMALL_BUFFER];    /*  Field name abbreviations  */
	char field_name[MAX_FIELDS][SMALL_BUFFER];  /*  Field names               */
	int rec_mult;                     /*
																		**  Treat m/c words as rec_mult butes
																		**  long. A good default value is 4.
																		**  This will allow a database of up
																		**  to 8 GBytes to be indexed.
																		*/
	int fixed;                        /*  defaults to 0                       */
	char db_type[SMALL_BUFFER];       /*  text or ai  												*/
	int has_lims;                     /*  Defaults to zero                    */
	int maxreclen;                    /*
																		**  Set dummy value, e.g. 20480
																		**  Real value set by convert_runtime
																		*/
	int ni;                           /*  No. of indexes                      */
	int last_ixvol[MAX_INDEXES];
	char ix_stem[MAX_INDEXES][SMALL_BUFFER];
                                    /*  <bib_dir>/<db_name>                 */
	int ix_volsize[MAX_INDEXES];      /*  For each index. defaults to 2047    */
	int ix_type[MAX_INDEXES];         /*  
																		**  Same for each index.
																		**  At the moment text=8, ai=9
																		*/
} db_param_info;

/*
**  Structure to hold 'search_groups info
*/

typedef struct search_groups {
	char index_name[SMALL_BUFFER];
	int v1;  /*  rubbish       */
	int v2;  /*  index number  */
	char keyword_type[SMALL_BUFFER];
	char stemming_function[SMALL_BUFFER];
	char gsl_fname[SMALL_BUFFER];
	char index_fields[SMALL_BUFFER];
} search_groups;
