##  Okapi Indexing Interface
##
##  TCL procedures Project
##
##  Jan 1998
##
## Generates an error dialog box
##

proc get_date { part } {
	##
	##  The Unix date command returns the date in the format, e.g.
	##
	##  Tue Mar  3 11:10:57 GMT 1998
	##
	##  case F -- returns full date: 03-Mar-98 11:10:57
	##  case T -- returns time only: 11:10:57

	set now [ exec date ]

	##  Pick out day

	set pos [ string first " " $now ]
	incr pos -1
	set day [ string range $now 0 $pos ]

	##  month_name

	incr pos 2
	set now [ string range $now $pos end ]

	set pos [ string first " " $now ]
	incr pos -1
	set month_name [ string range $now 0 $pos ]

	##  month_day

	incr pos 2
	set now [ string range $now $pos end ]

	if { [ string range $now 0 0 ] == " " } {
		set now [ string range $now 1 end ]
	}

	set pos [ string first " " $now ]
	incr pos -1
	set month_day [ string range $now 0 $pos ]

	if { $month_day < 10 } {
		set month_day "0$month_day"
	}

	##  hour

	incr pos 2
	set now [ string range $now $pos end ]

	set pos [ string first ":" $now ]
	incr pos -1
	set hour [ string range $now 0 $pos ]

	##  mins

	incr pos 2
	set now [ string range $now $pos end ]

	set pos [ string first ":" $now ]
	incr pos -1
	set mins [ string range $now 0 $pos ]

	## secs

	incr pos 2
	set now [ string range $now $pos end ]

	set pos [ string first " " $now ]
	incr pos -1
	set secs [ string range $now 0 $pos ]

	##  year

	set pos [ string last " " $now ]
	incr pos 3

	set year [ string range $now $pos end ]

	if { $part == "F" } {
		##  date and time

		set now "$month_day-$month_name-$year $hour:$mins:$secs"
	} else {
		##  time only
		set now "$hour:$mins:$secs"
	}

	return $now
}

##  Write history file entries
##

proc write_history { entry } {
  global logging
  global lgf

	if { $logging == "process_history" || $logging == "full_history" } {
		set now [ get_date T ]
		
		puts $lgf "$now:$entry"
		
		if { $entry == "quit:OK" } {
			
			set now [ get_date F ]
			
			puts $lgf "------------------------------------------------------------"
			puts $lgf "FINISHED INDEXING RUN: $now"
			puts $lgf "------------------------------------------------------------"
			
			##  Close log file
			
			close $lgf
			
		} else {
			flush $lgf
		}
	}
}

##  Display message when:
##
##  (a) process is lengthy, or
##  (b) a decision is required

proc display_process_message { w text } {
  global menu_x
  global menu_y

	## 1. Create the top-level window and divide it
	##    into top and bottom parts.

	catch {
		destroy $w
	}

	toplevel $w \
			-class Dialog \
			-width 600 \
			-height 200

	##  The "wm transient $w ." statement stops the
	##  display of the X-Windows title bar.
	##
	##  Don't use this just in case window needs to be moved by the user
	##
	##	wm transient $w .
	##	wm iconname $w Dialog

	frame $w.top \
			-relief raised \
			-bd 1 \
			-bg White

	pack $w.top \
			-side top \
			-fill both

	frame $w.bot \
			-relief raised \
			-bd 1 \
			-bg LightGrey

	pack $w.bot \
			-side bottom \
			-fill both

	## 2. Fill the top part with the message.

	message $w.msg \
			-width 10i \
			-justify center \
			-text $text \
			-bg White

	pack $w.msg \
			-in $w.top \
			-side right \
			-expand true \
			-fill both

	## 3. Create an Ok button at the bottom of the dialog.

	if { $w == ".quit_message" } {

		button $w.ok_button \
				-text "Ok" \
				-command "exit" \
				-bg NavyBlue \
				-fg LightGoldenRodYellow \
				-activebackground LightGoldenRodYellow \
				-activeforeground NavyBlue
		
		button $w.cancel_button \
				-text "No" \
				-command "cancel_command $w" \
				-bg NavyBlue \
				-fg LightGoldenRodYellow \
				-activebackground LightGoldenRodYellow \
				-activeforeground NavyBlue

		pack $w.ok_button \
				-in $w.bot \
				-side left \
				-expand true \
				-padx 3m \
				-pady 3m \
				-ipadx 2m \
				-ipady 1m

		pack $w.cancel_button \
				-in $w.bot \
				-side left \
				-expand true \
				-padx 3m \
				-pady 3m \
				-ipadx 2m \
				-ipady 1m

	} elseif { $w == ".ask_to_save" } {

		button $w.ok_button \
				-text "Ok" \
				-command "save_current_values $w" \
				-bg NavyBlue \
				-fg LightGoldenRodYellow \
				-activebackground LightGoldenRodYellow \
				-activeforeground NavyBlue
		
		button $w.cancel_button \
				-text "No" \
				-command "cancel_command $w" \
				-bg NavyBlue \
				-fg LightGoldenRodYellow \
				-activebackground LightGoldenRodYellow \
				-activeforeground NavyBlue

		pack $w.ok_button \
				-in $w.bot \
				-side left \
				-expand true \
				-padx 3m \
				-pady 3m \
				-ipadx 2m \
				-ipady 1m

		pack $w.cancel_button \
				-in $w.bot \
				-side left \
				-expand true \
				-padx 3m \
				-pady 3m \
				-ipadx 2m \
				-ipady 1m

	} elseif { $w == ".convert_runtime" || $w == ".create_indexes" } {

		button $w.ok_button \
				-text "Ok" \
				-command "remove_window $w" \
				-bg NavyBlue \
				-fg LightGoldenRodYellow \
				-activebackground LightGoldenRodYellow \
				-activeforeground NavyBlue

		pack $w.ok_button \
				-in $w.bot \
				-side left \
				-expand true \
				-padx 3m \
				-pady 3m \
				-ipadx 2m \
				-ipady 1m

	} else {

		button $w.ok_button \
				-text "Ok" \
				-command "press_return $w" \
				-bg NavyBlue \
				-fg LightGoldenRodYellow \
				-activebackground LightGoldenRodYellow \
				-activeforeground NavyBlue
		
		pack $w.ok_button \
				-in $w.bot \
				-side left \
				-expand true \
				-padx 3m \
				-pady 3m \
				-ipadx 2m \
				-ipady 1m

	}

	## 4. Withdraw the window, then update all the geometry
	##    information so we know how big it wants to be, then
	##    center the window in the display and de-iconify it.

	wm withdraw $w
	update idletasks

	wm geom $w +$menu_x+$menu_y
	wm deiconify $w

	## 5. Set a grab and claim the focus too.

	set oldFocus [focus]
	grab $w
	focus $w
}


proc cancel_command { w } {
	##  Just destroy the pop-up window
	destroy $w
}


proc press_return { w } {
		##  puts "THIS IS THE PRESS_RETURN PROCEDURE"
		destroy $w
		##  .term_entry_box delete 0 end
}


## Initial quit box - option to cancel
##

proc tk_quit1 { w title text bitmap default args } {
	##  Modify this to use as final quit popup window
	##
	global tk_priv
	global indbut

	## 1. Create the top-level window and divide it
	##    into top and bottom parts.

	catch { destroy $w}

	toplevel $w \
			-class Dialog

	wm title $w $title
	wm iconname $w Dialog

	frame $w.top \
			-relief raised \
			-bd 1 \
			-bg White

	pack $w.top \
			-side top \
			-fill both

	frame $w.bot \
			-relief raised \
			-bd 1 \
			-bg LightGrey

	pack $w.bot \
			-side bottom \
			-fill both

	## 2. Fill the top part with bitmap and message.

	message $w.msg \
			-width 3i \
			-text $text \
			-font -*-helvetica-bold-r-normal-*-120-* \
			-bg White

	pack $w.msg \
			-in $w.top \
			-side right \
			-expand 1 \
			-fill both \
			-padx 5m \
			-pady 5m

	if { $bitmap != ""} {
		label $w.bitmap -bitmap $bitmap
		pack $w.bitmap \
				-in $w.top \
				-side left \
				-padx 5m \
				-pady 5m
	}

	## 3. Create a row of buttons at the bottom of the dialog.

	set i 0

	foreach but $args {
		button $w.button$i \
				-text $but \
				-command "set tk_priv(button) $i" \
				-bg LightGrey \
				-fg Black \
				-activebackground LightGoldenRodYellow
		
		pack $w.button$i \
				-in $w.bot \
				-side left \
				-expand 1 \
				-padx 3m \
				-pady 3m \
				-ipadx 2m \
				-ipady 1m
		
		incr i
	}

	## 4. Withdraw the window, then update all the geometry
	##    information so we know how big it wants to be, then
	##    center the window in the display and de-iconify it.

	wm withdraw $w
	update idletasks
	set x [expr [winfo screenwidth $w]/2 - [winfo reqwidth $w]/2 \
			- [winfo vrootx [winfo parent $w]]]
	set y [expr [winfo screenheight $w]/2 - [winfo reqheight $w]/2 \
			- [winfo vrooty [winfo parent $w]]]
	wm geom $w +$x+$y
	wm deiconify $w

	## 5. Set a grab and claim the focus too.

	set oldFocus [focus]
	grab $w
	focus $w

	## 6. Wait for the user to respond, then restore the focus
	##    and return the index of the selected button.

	tkwait variable tk_priv(button)
	destroy $w
	focus $oldFocus
	set indbut $tk_priv(button)
}


##  Puts a busy cursor in the interface when something that
##  takes a long time is happening.
##
##  'busy2' requires 'busy1'

proc busy1 { } {
  global errorInfo

	set busy1 {
		.app .root
	}

	set list [ winfo children . ]

	while { $list != "" } {

		set next {}

		foreach w $list {
			set class [ winfo class $w ]
			set cursor [ lindex [ $w config -cursor ] 4 ]
			
			if { [ winfo toplevel $w ] == $w || $cursor != "" } {
				lappend busy1 [ list $w $cursor ]
			}
			
			set next [ concat $next [ winfo children $w ] ]
		}
		
		set list $next
	}

	foreach w $busy1 {
		catch {
			[ lindex $w 0 ] config -cursor watch
		}
	}

	update idletasks

	set error [ catch {	uplevel eval [ list term_entry ] } result ]
	set ei $errorInfo

	foreach w $busy1 {
		catch {
			[ lindex $w 0 ] config -cursor [ lindex $w 1 ]
		}
	}

	if $error {
		error $result $ei
	} else {
		return $result
	}
}

proc busy2 { name1 name2 name3 name4 name5 name6 } {
  global errorInfo

	set busy1 {}
	set busy2 {}
	set list [ winfo children . ]

	while { $list != "" } {

		set next {}

		foreach w $list {
			set class [ winfo class $w ]
			set cursor [ lindex [ $w config -cursor ] 4 ]
			
			if { [winfo toplevel $w] == $w || $cursor != "" } {
				lappend busy1 [ list $w $cursor ]
			}
			if { $cursor == "" && $class != "Frame" } {
				lappend busy2 [ list $w xterm ]
			}
			
			set next [ concat $next [ winfo children $w ] ]
		}
		set list $next
	}

	set busy3 [ concat $busy1 $busy2 ]

	foreach w $busy3 {
		catch {[lindex $w 0] config -cursor {watch red white}}
	}

	update idletasks

	set error [ catch {
		uplevel eval [list $name1 $name2 $name3 $name4 $name5 ]
	} result ]

	set ei $errorInfo

	foreach w $busy1 {
		catch {
			[ lindex $w 0 ] config -cursor [ lindex $w 1 ]
		}
	}

	foreach w $busy2 {
		catch {
			[ lindex $w 0 ] config -cursor ""
		}
	}

	if $error {
		error $result $ei
	} else {
		return $result
	}
}


##  timer -- displays time in a widget

proc timer name {
	toplevel .watch
	label .watch.label -text $name
	label .watch.value -text "Time"
	pack .watch.label .watch.value -side top
}

##  Procedure for tagging any text which matches
##  a certain pattern so that it may be redisplayed differently.
##  Can be used for highlighting search terms

proc forAllMatches {w pattern script} {
	scan [$w index end] %d numLines

	for {set i 0} {$i < $numLines} {incr i} {
		$w mark set last $i.0
		while {[regexp -indices $pattern \
				[$w get last "last lineend"] indices]} {
			$w mark set first "last + [lindex $indices 0] chars"
			$w mark set last "last + 1 chars + [lindex $indices 1] chars"
			uplevel $script
		}
	}
}


##  procedure to do nothing

proc do_nothing { } {
	##  Do nothing
	##  puts "DO NOTHING"
}



proc processing_message { wname message_text back fore } {
	## 1. Create the top-level window.

	catch {
		destroy $wname
	}

	toplevel $wname -class Dialog
	wm transient $wname .

	frame $wname.message_window \
			-relief raised \
			-bd 1 \
			-bg $back

	pack $wname.message_window \
			-side top \
			-fill both

	## 2. Fill the message_window
	##
	##    font was:
	##
	##      -font -*-helvetica-bold-r-normal-*-120-*
	##
	##  until NTD mucked up the font server!

	message $wname.msg \
			-bg $back \
			-fg $fore \
			-width 3i \
			-justify center \
			-text $message_text \
			-font -*-fixed-bold-r-normal--15-140-*-*-*-*-*-*

	pack $wname.msg \
			-in $wname.message_window \
			-side right \
			-expand 1 \
			-fill both \
			-padx 5m \
			-pady 5m

	## 4. Withdraw the window, then update all the geometry
	##    information so we know how big it wants to be, then
	##    center the window in the display and de-iconify it.

	wm withdraw $wname
	update idletasks

	##  Determine position to display window

	set root_x [ winfo rootx . ]
	set root_y [ winfo rooty . ]
	set root_w [ winfo screenwidth . ]
	set root_h [ winfo screenheight . ]

	set popup_w "400"
	set popup_h "80"
	set popup_x [ expr $root_x + ($root_w - $popup_w)/2 - 100 ]
	set popup_y [ expr $root_y + ($root_h - $popup_h)/2 - 100 ]

	wm geometry $wname 400x${popup_h}+${popup_x}+${popup_y}
	wm deiconify $wname

	## 5. Set a grab and claim the focus too.

	set oldFocus [focus]
	grab $wname
	focus $wname

	## Wait till message is displayed before returning
	##

	tkwait visibility $wname.msg
}


proc disable_close_button { w } {
	wm protocol $w WM_DELETE_WINDOW {
		puts "The close button is disabled."
	}
}

##  Packing and binding procedures
##

proc bind_main_window_buttons { } {
	bind .main_buttons.options {
		<Control-m> "post_menu .options_menu"
	}
}


proc database_menu { w } {
	post_menu $w.database
	cancel_menu $w
}



proc set_array_index { widget_name } {
	##
	##  For multiple records per screen field names end in _<rec_no>

  global array_index

	set pos [string last _ $widget_name]
	incr pos 1
	set array_index [string range $widget_name $pos end]

	##  puts "Array index = $array_index"
}


proc determine_coords { widget_name } {
	##
	##  For multiple records per screen field names end in _<rec_no>

  global menu_x
  global menu_y

	##  puts "WIDGET = $widget_name"

	set menu_x [winfo rootx $widget_name]
	set menu_y [winfo rooty $widget_name]
	set fullpath [winfo containing -displayof $widget_name $menu_x $menu_y]

	if { $fullpath == ".main_buttons.options" } {
		incr menu_x 10
		incr menu_y -100
	} elseif { $fullpath == ".main_buttons.next" } {
		incr menu_x 10
		incr menu_y -75
	} else {
		incr menu_x 20
		incr menu_y 30
	}

	##  puts "FULLPATH = $fullpath ($menu_x, $menu_y)"
}


proc post_menu { menu_name x y } {
	##
	##  First find (x, y) coordinates of top lefthand
	##  corner of options button
	##
  global menu_x
  global menu_y

	##  puts "$menu_name ($menu_x, $menu_y)"

	$menu_name post $menu_x $menu_y
}


#-- Cancels the menus when the 'Cancel' menu option is selected.

proc cancel_menu { menu_name } {
	$menu_name unpost
}


proc set_dbtype { type menu_name } {
  global db_type
  global show_message_menu

	set db_type $type
	write_history "set_dbtype:$db_type"

	cancel_menu $menu_name
}


proc set_ftype { type menu_name } {
  global field_type
  global array_index

	set field_type($array_index) $type
	write_history "set_ftype:$field_type($array_index)"

	cancel_menu $menu_name
}

proc set_kw_type { type menu_name } {
  global kw_type
  global array_index

	set kw_type($array_index) $type
	write_history "set_kw_type:$kw_type($array_index)"

	cancel_menu $menu_name
}

proc set_gsl_file { fname menu_name } {
  global gsl_file
  global array_index

	set gsl_file($array_index) $fname
	write_history "set_gsl_file:$gsl_file($array_index)"

	cancel_menu $menu_name
}

proc set_stemming_function { func menu_name } {
  global stem_fcn
  global array_index

	set stem_fcn($array_index) $func
	write_history "set_stemming_function:$stem_fcn($array_index)"

	cancel_menu $menu_name
}

##  Procedures to unpack the various screens.
##

proc unpack_intro_screen { } {
	##  Unpack widgets for intro_screen

	write_history "unpack_intro_screen"

	pack forget .intro_text_vscroll
	pack forget .intro_text
	pack forget .intro_text_frame
}

proc unpack_db_parameters_screen_1 { } {
  global installation_root
  global bss_parmpath
  global bib_dir
  global exchange_format_filename

	##  Unpack widgets for db_parameters_screen_1

	write_history "unpack_db_parameters_screen_1"

	foreach p1 { root parmp bibf exch } {
		pack forget .paths_$p1.entry_box
		pack forget .paths_$p1.field_label
		pack forget .paths_$p1
	}
}


proc unpack_db_parameters_screen_2 { } {
	##  Unpack widgets for db_parameters_screen_2

	write_history "unpack_db_parameters_screen_2"

	foreach p2 { exch dname dtype desc nf ni } {
		write_history "unpack_db_parameters_screen_2:\
				pack forget .dbinfo_$p2.entry_box"

		pack forget .dbinfo_$p2.entry_box

		if { $p2 == "dtype" } {
			write_history "unpack_db_parameters_screen_2:\
					pack forget .dbinfo_$p2.button_$p2"

			pack forget .dbinfo_$p2.button_$p2
		}

		write_history "unpack_db_parameters_screen_2:\
				pack forget .dbinfo_$p2.field_label"
		write_history "unpack_db_parameters_screen_2:\
				pack forget .dbinfo_$p2"

		pack forget .dbinfo_$p2.field_label
		pack forget .dbinfo_$p2
	}
}


proc unpack_field_entry_screen { } {
	##  Unpack widgets for field_entry_screen

  global start_field
  global end_field
  global field_screen_no

	write_history "unpack_field_entry_screen:$field_screen_no"

	set s $field_screen_no

	for { set f $start_field($s) } { $f <= $end_field($s) } { incr f 1 } {
		write_history "unpack_field_entry_screen:\
				pack forget .fields_$f.field_type_entry_$f"
		write_history "unpack_field_entry_screen:\
				pack forget .fields_$f.field_type_$f"
		write_history "unpack_field_entry_screen:\
				pack forget .fields_$f.2_$f"
		write_history "unpack_field_entry_screen:\
				pack forget .fields_$f.field_abbrev_$f"
		write_history "unpack_field_entry_screen:\
				pack forget .fields_$f.1_$f"
		write_history "unpack_field_entry_screen:\
				pack forget .fields_$f"

		pack forget .fields_$f.field_type_entry_$f
		pack forget .fields_$f.field_type_$f
		pack forget .fields_$f.2_$f
		pack forget .fields_$f.field_abbrev_$f
		pack forget .fields_$f.1_$f
		pack forget .fields_$f
	}
}


proc unpack_index_entry_screen { } {
	##  Unpack widgets for index_entry_screen

  global start_index
  global end_index
  global index_screen_no

	write_history "unpack_index_entry_screen:$index_screen_no"

	set i $index_screen_no

	for { set ix $start_index($i) } { $ix <= $end_index($i) } { incr ix 1 } {
		write_history "unpack_index_entry_screen:\
				pack forget .indexes_$ix.stemf_entry_$ix"
		write_history "unpack_index_entry_screen:\
				pack forget .indexes_$ix.index_stemf_$ix"
		write_history "unpack_index_entry_screen:\
				pack forget .indexes_$ix.stemf_label_$ix"
		write_history "unpack_index_entry_screen:\
				pack forget .indexes_$ix.gsl_entry_$ix"
		write_history "unpack_index_entry_screen:\
				pack forget .indexes_$ix.index_gsl_$ix"
		write_history "unpack_index_entry_screen:\
				pack forget .indexes_$ix.gsl_label_$ix"
		write_history "unpack_index_entry_screen:\
				pack forget .indexes_$ix.kw_entry_$ix"
		write_history "unpack_index_entry_screen:\
				pack forget .indexes_$ix.index_kw_$ix"
		write_history "unpack_index_entry_screen:\
				pack forget .indexes_$ix.kw_label_$ix"
		write_history "unpack_index_entry_screen:\
				pack forget .indexes_$ix.flist_entry_$ix"
		write_history "unpack_index_entry_screen:\
				pack forget .indexes_$ix.flist_label_$ix"
		write_history "unpack_index_entry_screen:\
				pack forget .indexes_$ix.ixname_entry_$ix"
		write_history "unpack_index_entry_screen:\
				pack forget .indexes_$ix.ix_name_label_$ix"
		write_history "unpack_index_entry_screen:\
				pack forget .indexes_$ix"

		pack forget .indexes_$ix.stemf_entry_$ix
		pack forget .indexes_$ix.index_stemf_$ix
		pack forget .indexes_$ix.stemf_label_$ix
		pack forget .indexes_$ix.gsl_entry_$ix
		pack forget .indexes_$ix.index_gsl_$ix
		pack forget .indexes_$ix.gsl_label_$ix
		pack forget .indexes_$ix.kw_entry_$ix
		pack forget .indexes_$ix.index_kw_$ix
		pack forget .indexes_$ix.kw_label_$ix
		pack forget .indexes_$ix.flist_entry_$ix
		pack forget .indexes_$ix.flist_label_$ix
		pack forget .indexes_$ix.ixname_entry_$ix
		pack forget .indexes_$ix.ix_name_label_$ix
		pack forget .indexes_$ix
	}
}

proc unpack_convert_db_screen { } {
	write_history "unpack_convert_db_screen:\
			pack forget .db_convert_button"
	write_history "unpack_convert_db_screen:\
			pack forget .db_convert_text"
	write_history "unpack_convert_db_screen:\
			pack forget .db_convert_text_vscroll"		
	write_history "unpack_convert_db_screen:\
			pack forget .db_convert_bottom_frame"
	write_history "unpack_convert_db_screen:\
			pack forget .db_convert_top_frame"

	pack forget .db_convert_button
	pack forget .db_convert_text
	pack forget .db_convert_text_vscroll		
	pack forget .db_convert_bottom_frame	
	pack forget .db_convert_top_frame

}


proc unpack_create_index_screen { } {
	write_history "unpack_create_index_screen:\
			pack forget .create_index_button"
	write_history "unpack_create_index_screen:\
			pack forget .create_index_text"
	write_history "unpack_create_index_screen:\
			pack forget .create_index_text_vscroll"
	write_history "unpack_create_index_screen:\
			pack forget .create_index_bottom_frame"
	write_history "unpack_create_index_screen:\
			pack forget .create_index_top_frame"

	pack forget .create_index_button
	pack forget .create_index_text
	pack forget .create_index_text_vscroll		
	pack forget .create_index_bottom_frame	
	pack forget .create_index_top_frame

}


proc unpack_finish_screen { } {
	write_history "unpack_finish_screen:\
			pack forget .finish_button"
	write_history "unpack_finish_screen:\
			pack forget .finish_text"
	write_history "unpack_finish_screen:\
			pack forget .finish_text_vscroll"
	write_history "unpack_finish_screen:\
			pack forget .finish_bottom_frame"
	write_history "unpack_finish_screen:\
			pack forget .finish_top_frame"

	pack forget .finish_button
	pack forget .finish_text
	pack forget .finish_text_vscroll		
	pack forget .finish_bottom_frame	
	pack forget .finish_top_frame

}


##  Procedures to create widgets and pack them for the
##  various screens of the indexing process

proc check_existing_history_file { } {
  global logfile
  global lgf

	set now [ get_date F ]

	if { [ file exists $logfile ] } {
		if { [ catch { open $logfile a } lgf ] } {
			puts stderr "CANNOT OPEN $logfile for appending"
			puts stderr "There will be no log kept."
			set lgf "NO_HISTORY"
		} else {
			puts stderr "OPENED $logfile for appending"
			puts $lgf "------------------------------------------------------------"
			puts $lgf "NEW INDEXING RUN: $now"
			puts $lgf "------------------------------------------------------------"
		}
	} elseif { [ catch { open $logfile w } lgf ] } {
		puts stderr "CANNOT OPEN $logfile for appending"
		puts stderr "There will be no log kept."
		set lgf "NO_HISTORY"
	} else {
		puts stderr "OPENED $logfile for writing"
		puts $lgf "------------------------------------------------------------"
		puts $lgf "NEW INDEXING RUN: $now"
		puts $lgf "------------------------------------------------------------"
	}
}


proc intro_screen { } {
	##  First screen displayed.
	##
	##  Always disable the previous button

  global installation_root
  global intro_label
  global db_params1_button

	set filename "$installation_root/docs/intro.txt"

	##  The following procedures opens a history file.
	##  If it exists it opens it for appending, otherwise
	##  it opens a brand new file for writing.

	check_existing_history_file

	##  This is the first procedure called when the application
	##  is run. Check for existence of "$installation_root/IndexLogs"

	write_history "intro_screen:pack_widgets"

	##  Change header title

	.top_header_label configure -text $intro_label

	##  Re-configure and re-bind <previous> and <next> buttons

	set screen "intro"
	set previous ""
	set next $db_params1_button

	pack_main_buttons $screen $previous $next

	##  bind "Next" button
	##

	bind .main_buttons.next <Button-1> {
		unpack_intro_screen
		db_parameters_screen_1
	}

	bind . <Control-n> {
		unpack_intro_screen
		db_parameters_screen_1
	}

	##  Pack widgets for intro_screen
	##

	pack .intro_text_frame \
			-in .upper \
			-side top \
			-fill x

  bind .intro_text <Enter> {
    focus .intro_text_vscroll
	}

	pack .intro_text_vscroll \
			-in .intro_text_frame \
			-side right \
			-fill y

	pack .intro_text \
			-in .intro_text_frame \
			-expand 1 \
			-fill both

	load_file .intro_text $filename
}


proc load_default_db { } {
  global bss_parmpath
  global db_name
  global menu_name

	set now [ get_date T ]

	set db_params_file "$bss_parmpath/$db_name"

	if { [ file exists $db_params_file ] } {
		write_history "load_default_db:$db_params_file:file_exists"
		read_current_values $menu_name
	} else {
		write_history "load_default_db:$db_params_file:no_such_file"
	}
}


proc db_parameters_screen_1 { } {
  global installation_root_label
  global bss_parmpath_label
  global bib_dir_label
  global exchange_format_label
  global paths_label

  global installation_root
  global bss_parmpath
  global bib_dir
  global exchange_format_filename

  global intro_button
  global db_params2_button

	##  Change header title

	write_history "db_parameters_screen_1:pack_widgets"

	.top_header_label configure -text $paths_label

	##  Re-configure and re-bind <previous> and <next> buttons

	set screen "db_params1"
	set previous_button $intro_button
	set next_button $db_params2_button

	pack_main_buttons $screen $previous_button $next_button

  bind .main_buttons.previous <Button-1> {
    unpack_db_parameters_screen_1
		intro_screen
	}

	bind . <Control-p> {
    unpack_db_parameters_screen_1
		intro_screen
	}

	bind .main_buttons.next <Button-1> {
		load_default_db
    unpack_db_parameters_screen_1
		db_parameters_screen_2
	}

	bind . <Control-n> {
		load_default_db
    unpack_db_parameters_screen_1
		db_parameters_screen_2
	}

	##  Pack widgets for db_parameters_screen_1

	foreach p1 { root parmp bibf exch } {
		pack .paths_$p1 \
				-in .upper \
				-side top \
				-fill x

		pack .paths_$p1.field_label \
			-side top

		pack .paths_$p1.entry_box \
			-side top \
			-fill x \
			-expand true
	}
}


proc test_nf { } {
  global nf

	write_history "test_nf:$nf"

	if { $nf == 0 } {
		determine_coords .main_buttons.next
		set w ".error_message"
		set message "The number of fields per record is set to ZERO."

		display_process_message $w $message
	} else {
		unpack_db_parameters_screen_2
		set_no_field_entry_screens
		field_entry_screen
	}
}

proc db_parameters_screen_2 { } {
  global db_name_label
  global display_name_label
  global db_type_label
  global nf_label
  global ni_label
  global nf
  global dbase_label

  global db_params1_button
  global field_entry_button

	write_history "db_parameters_screen_2"

	##  Change header title

	.top_header_label configure -text $dbase_label

	##  Re-configure and re-bind <previous> and <next> buttons

	set screen "db_params2"
	set previous_button $db_params1_button
	set next_button $field_entry_button

	pack_main_buttons $screen $previous_button $next_button

	bind .main_buttons.previous <Button-1> {
		unpack_db_parameters_screen_2
		db_parameters_screen_1
	}

	bind . <Control-p> {
		unpack_db_parameters_screen_2
		db_parameters_screen_1
	}

	bind .main_buttons.next <Button-1> {
		test_nf
	}

	bind . <Control-n> {
		test_nf
	}

	##  Pack widgets for db_parameters_screen_2

	foreach p2 { exch dname dtype desc nf ni } {
		write_history "db_parameters_screen_2:pack:.dbinfo_$p2"

		pack .dbinfo_$p2 \
				-in .upper \
				-side top \
				-fill x

		if { $p2 == "nf" } {
			pack .dbinfo_$p2.field_label -side left
		} elseif { $p2 == "ni" } {
			pack .dbinfo_$p2.field_label -side left
		} else {
			pack .dbinfo_$p2.field_label -side top
		}

		if { $p2 == "exch" } {
			pack .dbinfo_$p2.button_$p2 \
					-side left
		}

		if { $p2 == "dname" } {
			pack .dbinfo_$p2.button_$p2 \
					-side left
		}

		if { $p2 == "dtype" } {
			pack .dbinfo_$p2.button_$p2 \
					-side left
		}

		if { $p2 == "nf" } {
			pack .dbinfo_$p2.entry_box \
					-side left \
					-fill x \
					-expand true
		} elseif { $p2 == "ni" } {
			pack .dbinfo_$p2.entry_box \
					-side left \
					-fill x \
					-expand true
		} else {
			pack .dbinfo_$p2.entry_box \
					-side top \
					-fill x \
					-expand true
		}
	}
}


proc set_no_field_entry_screens { } {
  global nf
  global start_field
  global end_field
  global nfields_per_screen
  global field_screen_no
  global no_full_field_screens
  global remaining_fields

	set no_full_field_screens [ expr [ expr $nf - 1 ] / $nfields_per_screen ]
	incr no_full_field_screens 1

	set remaining_fields [ expr $nf % $nfields_per_screen ]

	##  set up start_field & end_field arrays

	for { set sno 0 } { $sno < $no_full_field_screens } { incr sno 1 } {
		set start_field($sno) [ expr $sno * $nfields_per_screen + 1 ]

		if { [ expr $sno + 1 ] == $no_full_field_screens } {
			set end_field($sno) $nf
		} else {
			set end_field($sno) [ expr $start_field($sno) + $nfields_per_screen - 1 ]
		}
	}

	set field_screen_no 0

	write_history "set_no_field_entry_screens:\
			$no_full_field_screens:$remaining_fields:$field_screen_no"
}


proc call_next_field_screen { pressed } {
  global field_screen_no

	if { $pressed == "P" } {
		write_history "call_next_field_screen:previous_screen"

		incr field_screen_no -1
	} else {
		write_history "call_next_field_screen:next_screen"

		incr field_screen_no 1
	}

	field_entry_screen
}


proc test_ni { } {
  global nf
  global ni
  global field_abbrev
  global field_type

	set field_entry_flag 1
	set message ""
	set fmessage ""
	set imessage ""

	##  make sure $ni is not zero

	if { $ni == 0 } {
		set ni 1
	}

	write_history "test_ni:$ni"

	for { set f 1 } { $f <= $nf } { incr f 1 } {
		if { $field_abbrev($f) == "" || $field_type($f) == "" } {
			set field_entry_flag 0
		}
	}

	##  First check that the field entry screen has been completed

	if { $field_entry_flag == 0 } {
		set message "The FIELD ENTRY DATA is incomplete."

		write_history "test_ni:$message"
		
		determine_coords .main_buttons.next
		set w ".error_messages"
		display_process_message $w $message
	} else {
		unpack_field_entry_screen
		set_no_index_entry_screens
		index_entry_screen
	}
}


proc field_entry_screen { } {
	##
	##  $direction determines the previous screen (to be unpacked):
	##
	##  "F" db_parameters_screen_2
	##  "B" index_entry_screen

  global ni
  global field_label
  global start_field
  global end_field
  global nfields_per_screen
  global field_screen_no
  global no_full_field_screens

  global db_params2_button
  global index_entry_button

	write_history "field_entry_screen:$field_screen_no"

	set s $field_screen_no

	##  Change header title

	.top_header_label configure -text $field_label

	##  Re-configure and re-bind <previous> and <next> buttons

	set screen "field_entry"
	set previous_button $db_params2_button
	set next_button $index_entry_button

	pack_main_buttons $screen $previous_button $next_button

	if { $s == 0 } {

		bind .main_buttons.previous <Button-1> {
			unpack_field_entry_screen
			db_parameters_screen_2
		}
		
		bind . <Control-p> {
			unpack_field_entry_screen
			db_parameters_screen_2
		}

	} else {

		bind .main_buttons.previous <Button-1> {
			unpack_field_entry_screen
			call_next_field_screen P
		}
		
		bind . <Control-p> {
			unpack_field_entry_screen
			call_next_field_screen P
		}

	}

	if { $s == [ expr $no_full_field_screens - 1 ] } {

		bind .main_buttons.next <Button-1> {
			test_ni
		}

		bind . <Control-n> {
			test_ni
		}

	} else {

		bind .main_buttons.next <Button-1> {
			unpack_field_entry_screen
			call_next_field_screen N
		}

		bind . <Control-n> {
			unpack_field_entry_screen
			call_next_field_screen N
		}

	}

	##  Pack widgets for field_entry_screen

	for { set f $start_field($s) } {$f <= $end_field($s) } { incr f 1 } {

		pack .fields_$f \
				-in .upper \
				-side top \
				-fill x
		
		pack .fields_$f.1_$f \
				-side left

		pack .fields_$f.field_abbrev_$f \
				-side left \
				-fill x \
				-expand true
		
		pack .fields_$f.2_$f \
				-side left
		
		pack .fields_$f.field_type_$f \
				-side left
		
		pack .fields_$f.field_type_entry_$f \
				-side left \
				-fill x \
				-expand true
	}
}


proc test_convert_runtime { } {
  global nf
  global ni
  global field_abbrev
  global field_type


	set field_entry_flag 1
	set message ""
	set fmessage ""
	set imessage ""

	##  First check that field entry is complete

	for { set f 1 } { $f <= $nf } { incr f 1 } {
		if { $field_abbrev($f) == "" || $field_type($f) == "" } {
			set field_entry_flag 0
		}
	}

	if { $field_entry_flag == 0 } {
		set fmessage "The FIELD ENTRY DATA is incomplete."
		write_history "test_convert_runtime:$fmessage"
	}

	write_history "test_convert_runtime:number_of_indexes:$ni"

	if { $ni == 0 } {
		write_history "test_convert_runtime:Return to the Database Details screen"
		set imessage "The NUMBER OF INDEXES is set to ZERO."
		set imessage "$imessage\nReturn to the \"Database Details\" screen."
	}

	if { $fmessage != "" || $imessage != "" } {
		if { $fmessage != "" } {
			if { $imessage != "" } {
				set message "$fmessage\n\n$imessage"
			} else {
				set message "$fmessage"
			}
		} else {
			set message "$imessage"
		}


		determine_coords .main_buttons.next
		set w ".error_messages"
		display_process_message $w $message
	} else {
		write_history "test_convert_runtime:OK"

		unpack_field_entry_screen
		set_no_index_entry_screens
		index_entry_screen
	}
}



##  Set up global arrays for start and end of index entry screens

proc set_no_index_entry_screens { } {
  global ni
  global start_index
  global end_index
  global nfields_per_screen
  global index_screen_no
  global no_full_index_screens
  global remaining_indexes

	set no_full_index_screens [ expr [ expr $ni - 1 ] / $nfields_per_screen ]
	incr no_full_index_screens 1

	set remaining_indexes [ expr $ni % $nfields_per_screen ]

	##  set up start_index & end_index arrays

	for { set sno 0 } { $sno < $no_full_index_screens } { incr sno 1 } {
		set start_index($sno) [ expr $sno * $nfields_per_screen + 1 ]

		if { [ expr $sno + 1 ] == $no_full_index_screens } {
			set end_index($sno) $ni
		} else {
			set end_index($sno) [ expr $start_index($sno) + $nfields_per_screen - 1 ]
		}
	}

	set index_screen_no 0

	write_history "set_no_index_entry_screens:\
			no_full_index_screens:$remaining_indexes:$index_screen_no"
}

proc call_next_index_screen { pressed } {
  global index_screen_no

	write_history "call_next_index_screen"

	if { $pressed == "P" } {
		incr index_screen_no -1
	} else {
		incr index_screen_no 1
	}

	index_entry_screen
}

proc index_entry_screen { } {
  global ni

  global index_label
  global start_index
  global end_index
  global nfields_per_screen
  global index_screen_no
  global no_full_index_screens

  global field_entry_button
  global convert_db_button

	write_history "index_entry_screen:$index_screen_no"

	set ix $index_screen_no

	##  Change header title

	.top_header_label configure -text $index_label

	##  Re-configure and re-bind <previous> and <next> buttons

	set screen "index_entry"
	set previous_button $field_entry_button
	set next_button $convert_db_button

	pack_main_buttons $screen $previous_button $next_button

	if { $ix == 0 } {

		bind .main_buttons.previous <Button-1> {
			unpack_index_entry_screen
			set_no_field_entry_screens
			set field_screen_no $no_full_field_screens
			call_next_field_screen P
		}

		bind . <Control-p> {
			unpack_index_entry_screen
			set_no_field_entry_screens
			set field_screen_no $no_full_field_screens
			call_next_field_screen P
		}

	} else {

		bind .main_buttons.previous <Button-1> {
			unpack_index_entry_screen
			call_next_index_screen P
		}
		
		bind . <Control-p> {
			unpack_index_entry_screen
			call_next_index_screen P
		}

	}

	if { $ix == [ expr $no_full_index_screens - 1 ] } {

		bind .main_buttons.next <Button-1> {
			ask_to_save .save_menu
			unpack_index_entry_screen
			convert_db_screen
			check_for_existing_files
		}
		
		bind . <Control-n> {
			ask_to_save .save_menu
			unpack_index_entry_screen
			convert_db_screen
			check_for_existing_files
		}

	} else {

		bind .main_buttons.next <Button-1> {
			unpack_index_entry_screen
			call_next_index_screen N
		}
		
		bind . <Control-n> {
			unpack_index_entry_screen
			call_next_index_screen N
		}

	}

	##  Pack widgets for "index_entry_screen"

	for { set i $start_index($ix) } {$i <= $end_index($ix) } { incr i 1 } {
		# create a frame for buttons and entry

		pack .indexes_$i \
				-in .upper \
				-side top \
				-fill x
		pack .indexes_$i.ix_name_label_$i \
				-side left

		pack .indexes_$i.ixname_entry_$i \
				-side left \
				-fill x \
				-expand true
		
		pack .indexes_$i.kw_label_$i \
				-side left
		
		pack .indexes_$i.index_kw_$i -side left

		pack .indexes_$i.kw_entry_$i \
				-side left \
				-fill x \
				-expand true

		pack .indexes_$i.gsl_label_$i \
				-side left
		
		pack .indexes_$i.index_gsl_$i -side left

		pack .indexes_$i.gsl_entry_$i \
				-side left \
				-fill x \
				-expand true

		pack .indexes_$i.stemf_label_$i \
				-side left
		
		pack .indexes_$i.index_stemf_$i -side left

		pack .indexes_$i.stemf_entry_$i \
				-side left \
				-fill x \
				-expand true

		pack .indexes_$i.flist_label_$i \
				-side left
		
		pack .indexes_$i.flist_entry_$i \
				-side left \
				-fill x \
				-expand true

	}
}


proc convert_db_screen { } {
  global convert_label
  global no_full_index_screens
  global index_screen_no
  global installation_root
  global index_entry_button
  global create_indexes_button

	##  Change header title

	.top_header_label configure -text $convert_label

	set filename "$installation_root/docs/convert.txt"

	write_history "convert_db_screen:info_file:$filename"

	##  Re-configure and re-bind <previous> and <next> buttons

	set screen "convert_db"
	set previous_button $index_entry_button
	set next_button $create_indexes_button

	pack_main_buttons $screen $previous_button $next_button

	bind .main_buttons.next <Button-1> {
		unpack_convert_db_screen
		create_index_screen
	}

	bind . <Control-n> {
		unpack_convert_db_screen
		create_index_screen
	}

	bind .main_buttons.previous <Button-1> {
		unpack_convert_db_screen
		set_no_index_entry_screens
		set index_screen_no $no_full_index_screens
		call_next_index_screen P
	}

	bind . <Control-p> {
		unpack_convert_db_screen
		set_no_index_entry_screens
		set index_screen_no $no_full_index_screens
		call_next_index_screen P
	}

	##  Pack widgets for intro_screen

	pack .db_convert_top_frame \
			-in .upper \
			-side top \
			-fill x

	pack .db_convert_bottom_frame \
			-in .upper \
			-side bottom \
			-fill x

	pack .db_convert_text_vscroll \
			-in .db_convert_top_frame \
			-side right \
			-fill y
		
	pack .db_convert_text \
			-in .db_convert_top_frame \
			-expand 1 \
			-fill both

	pack .db_convert_button \
			-in .db_convert_bottom_frame \
			-side top \
			-fill x

	load_file .db_convert_text $filename
}


proc create_index_screen { } {
  global create_index_label
  global no_full_index_screens
  global index_screen_no
  global installation_root

  global convert_db_button
  global finish_button

	##  Change header title

	.top_header_label configure -text $create_index_label

	set filename "$installation_root/docs/indexing.txt"

	write_history "create_index_screen:info_file:$filename"

	##  Re-configure and re-bind <previous> and <next> buttons

	set screen "create_index"
	set previous_button $convert_db_button
	set next_button $finish_button

	pack_main_buttons $screen $previous_button $next_button

	bind .main_buttons.next <Button-1> {
		unpack_create_index_screen
		finish_screen
	}

	bind . <Control-n> {
		unpack_create_index_screen
		finish_screen
	}

	bind .main_buttons.previous <Button-1> {
		unpack_create_index_screen
		convert_db_screen
	}

	bind . <Control-p> {
		unpack_create_index_screen
		convert_db_screen
	}

	##  Pack widgets for intro_screen

	pack .create_index_top_frame \
			-in .upper \
			-side top \
			-fill x

	pack .create_index_bottom_frame \
			-in .upper \
			-side bottom \
			-fill x

	pack .create_index_text_vscroll \
			-in .create_index_top_frame \
			-side right \
			-fill y
		
	pack .create_index_text \
			-in .create_index_top_frame \
			-expand 1 \
			-fill both

	pack .create_index_button \
			-in .create_index_bottom_frame \
			-side top \
			-fill x

	load_file .create_index_text $filename
}


proc finish_screen { } {
  global finish_label
  global no_full_index_screens
  global index_screen_no
  global installation_root
  global create_indexes_button
  global exit_button

	##  Change header title

	.top_header_label configure -text $finish_label

	set filename "$installation_root/docs/finish.txt"

	write_history "finish_screen:info_file:$filename"

	##  Re-configure and re-bind <previous> and <next> buttons

	set screen "finish"
	set previous_button $create_indexes_button
	set next_button $exit_button

	pack_main_buttons $screen $previous_button $next_button

	bind .main_buttons.next <Button-1> {
		determine_coords %W
		quit
	}

	bind . <Control-n> {
		determine_coords %W
		quit
	}

	bind .main_buttons.previous <Button-1> {
		unpack_finish_screen
		create_index_screen
	}

	bind . <Control-p> {
		unpack_finish_screen
		create_index_screen
	}

	##  Pack widgets for intro_screen

	pack .finish_top_frame \
			-in .upper \
			-side top \
			-fill x

	pack .finish_bottom_frame \
			-in .upper \
			-side bottom \
			-fill x

	pack .finish_text_vscroll \
			-in .finish_top_frame \
			-side right \
			-fill y
		
	pack .finish_text \
			-in .finish_top_frame \
			-expand 1 \
			-fill both

	load_file .finish_text $filename
}



proc file_props { filepath } {

}

proc newer { file1 file2 } {
	## Practical Programming in Tcl and Tk, Brent B Welch, P.66

	write_history "newer:$file1:$file2"

	if { ![ file exists $file1 ] } {
		write_history "newer:$file1:does_not_exist"
		return 1
	} elseif { ![ file exists $file2 ] } {
		write_history "newer:$file2:does not exist"
		return 1
	} else {
		expr [ file mtime $file1 ] > [ file mtime $file2 ]
		write_history "newer"
	}
}

proc remove_window { w } {
	destroy $w
}


proc convert_runtime { } {
  global installation_root
  global bss_parmpath
  global bib_dir
  global exchange_format_path
  global exchange_format_filename
  global db_name

	##  First append database name to db_avail

	set db_avail_fname "$bss_parmpath/db_avail"

	if { [ catch { open $db_avail_fname a } fptr ] } {
		puts stderr "CANNOT OPEN $db_avail_fname for appending"
		puts stderr "You must scheck this as you wont be able to"
		puts stderr "the database until it has been added to $db_avail_fname"
	} else {
		puts stderr "OPENED $db_avail_fname for appending"
		puts $fptr "$db_name *"
		close $fptr
	}

	set prog "$installation_root/bin/convert_runtime"
	set p1 "$bss_parmpath"
	set p2 "$db_name"
	set src_file "$exchange_format_path/$exchange_format_filename"

	set  message "\nConverting \"$exchange_format_filename\" .....\n"
	set window_name ".processing"

	set back "NavyBlue"
	set fore "Yellow"
		
	processing_message $window_name $message $back $fore

	write_history "convert_runtime:$prog:$p1:$p2:$src_file"

	set result [ busy2 c.convert_runtime $prog $p1 $p2 $src_file \"\"]

	destroy $window_name

	set window_name ".convert_runtime"

	if { $result == "ERROR" } {
		write_history "convert_runtime:$src_file:$db_name:error"
		set message "\nERROR converting \"$exchange_format_filename\"\n"
	} else {
		write_history "convert_runtime:$src_file:$db_name:ok"
		set message "\n\"$exchange_format_filename\" converted successfully\n"
	}

	display_process_message $window_name $message
}



proc create_indexes { } {
  global installation_root
  global bss_parmpath
  global db_name
  global ni

	set prog1 "$installation_root/bin/ix1 -c"
	set prog2 "$installation_root/bin/ixf -c"

##	set n [ exec $prog $params < $src_file ]

	write_history "create_indexes:$prog1:$prog2:$bss_parmpath:$db_name:$ni"

	set  message "\nCreating Indexes for \"$db_name\" .....\n"
	set window_name ".processing"

	set back "NavyBlue"
	set fore "Yellow"
		
	processing_message $window_name $message $back $fore

	set result [ busy2 c.create_indexes \
			$installation_root $bss_parmpath $db_name $ni \"\"]

	destroy $window_name

	set window_name ".create_indexes"

	if { $result == "ERROR" } {
		write_history "create_indexes:$db_name:error"
		set message "\nERROR CREATING INDEXES FOR \"$db_name\"\n"
	} else {
		write_history "create_indexes:OK"
		set message "\nIndexes successfuully created for \"$db_name\"\n"
	}

	display_process_message $window_name $message
}


proc check_for_existing_files { } {
  global installation_root
  global bss_parmpath
  global bib_dir
  global exchange_format_path
  global exchange_format_filename
  global db_name

	set db_params_file "$bss_parmpath/$db_name"
	set exchange_format_file "$exchange_format_path/$exchange_format_filename"
	set bibfiles_file "$bib_dir/$db_name.bib"


	if { [ file exists $db_params_file ] } {
		puts "check_for_existing_files:$db_params_file:exists"
	} else {
		puts "check_for_existing_files:$db_params_file:does not exist"
	}

	if { [ file exists $exchange_format_file ] } {
		puts "check_for_existing_files:$exchange_format_file:exists"
	} else {
		puts "check_for_existing_files:$exchange_format_file:does not exist"
	}

	if { [ file exists $bibfiles_file ] } {
		puts "check_for_existing_files:$bibfiles_file:exists"
	} else {
		puts "check_for_existing_files:$bibfiles_file:does not exist"
	}
}

########################################################################
##                                                                    ##
##  Database parameter filing procedures                              ##
##                                                                    ##
########################################################################

proc ask_to_save { $menu_name } {
  global db_name

	write_history "ask_to_save"

	set w ".ask_to_save"
	set message "Save parameter files for $db_name?"
	display_process_message $w $message
}


proc save_current_values { menu_name } {

  global installation_root
  global exchange_format_filename
  global bss_parmpath
  global db_name
  global lastbibvol
  global bib_basename
  global bib_dir
  global bibsize
  global real_bibsize
  global display_name
  global explanation

  global nr

  global nf
  global field_abbrev
  global field_type

  global rec_mult
  global fixed
  global db_type
  global has_lims
  global maxreclen

  global ni
  global index_name
  global field_list
  global kw_type
  global gsl_file
  global stem_fcn

  global last_ixvol
  global ix_stem
  global ix_volsize
  global ix_type

  global max_fields
  global max_indexes

	set open_error "cannot open for writing"
	set open_ok "opened for writing"

	set pr_file "$bss_parmpath/$db_name"
	set ft_file "$bss_parmpath/$db_name.field_types"
	set sg_file "$bss_parmpath/$db_name.search_groups"

	if [ catch { open $pr_file w } prf ] {
		write_history "save_current_values:$pr_file:$open_error"
		puts stderr "Cannot open $pr_file for writing"
		set prf "NOT_OPEN"
	} else {
		write_history "save_current_values:$pr_file:$open_ok"
	}

	if [ catch { open $ft_file w } ftf ] {
		write_history "save_current_values:$ft_file:$open_error"
		puts stderr "Cannot open $ft_file for writing"
		set ftf "NOT_OPEN"
	} else {
		write_history "save_current_values:$ft_file:$open_ok"
	}

	if [ catch { open $sg_file w } sgf ] {
		write_history "save_current_values:$sg_file:$open_error"
		puts stderr "Cannot open $sg_file for writing"
		set sgf "NOT_OPEN"
	} else {
		write_history "save_current_values:$sg_file:$open_ok"
	}

	if { $prf != "NOT_OPEN" } {
		##
		##  Write history file entries

		set lastbibvol 0

		write_history "save_current_values:name=$db_name"
		write_history "save_current_values:lastbibvol=$lastbibvol"
		write_history "save_current_values:bib_basename=$db_name.bib"
		write_history "save_current_values:bib_dir=$bib_dir"
		write_history "save_current_values:bibsize=$bibsize"
		write_history "save_current_values:real_bibsize=$real_bibsize"
		write_history "save_current_values:display_name=$db_name"
		write_history "save_current_values:explanation=$display_name"
		write_history "save_current_values:nr=$nr"
		write_history "save_current_values:nf=$nf"
		
		##  Write to parameters file
		##

		puts $prf "name=$db_name"
		puts $prf "lastbibvol=$lastbibvol"
		puts $prf "bib_basename=$db_name.bib"
		puts $prf "bib_dir=$bib_dir"
		puts $prf "bibsize=$bibsize"
		puts $prf "real_bibsize=$real_bibsize"
		puts $prf "display_name=$db_name"
		puts $prf "explanation=$display_name"
		puts $prf "nr=$nr"
		puts $prf "nf=$nf"
	}

	for { set f 1 } { $f <= $nf } { incr f 1 } {
		set field_abbrev($f) [ string toupper $field_abbrev($f) ]
		
		if { $prf != "NOT_OPEN" } {
			write_history "save_current_values:$f:f_abbrev=$field_abbrev($f)"
			puts $prf "f_abbrev=$field_abbrev($f)"
		}

		if { $ftf != "NOT_OPEN" } {
			write_history "save_current_values:$f $field_type($f)"
			puts $ftf "$f $field_type($f)"
		}
	}

	if { $ftf != "NOT_OPEN" } {
		write_history "save_current_values:$ft_file:closed"
		close $ftf
	}

	##  Now set field data to nulls for $nf + 1 to $max_fields

	if { $nf < $max_fields } {
		set startf [ expr $nf + 1 ]
		
		for { set f $startf } { $f <= $max_fields } { incr f 1 } {
			set field_abbrev($f) ""
			set field_type($f) ""
		}
	}

	if { $prf != "NOT_OPEN" } {
		write_history "save_current_values:rec_mult=$rec_mult"
		write_history "save_current_values:fixed=$fixed"
		write_history "save_current_values:db_type=$db_type"
		write_history "save_current_values:has_lims=$has_lims"
		write_history "save_current_values:maxreclen=$maxreclen"
		write_history "save_current_values:ni=$ni"
		
		puts $prf "rec_mult=$rec_mult"
		puts $prf "fixed=$fixed"
		puts $prf "db_type=$db_type"
		puts $prf "has_lims=$has_lims"
		puts $prf "maxreclen=$maxreclen"
		puts $prf "ni=$ni"
	}

	for { set i 1 } {$i <= $ni } { incr i 1 } {
		set ino [ expr $i - 1 ]

		set last_ixvol($i) 0
		set ix_stem($i) "$bib_dir/$db_name"

		if { $db_type == "text" } {
			set ix_type($i) 8
		} else {
			set ix_type($i) 9
		}

		set ix_volsize($i) 2047

		if { $prf != "NOT_OPEN" } {
			write_history "save_current_values:last_ixvol=$last_ixvol($i)"
			write_history "save_current_values:ix_stem=$ix_stem($i)"
			write_history "save_current_values:ix_volsize=$ix_volsize($i)"
			write_history "save_current_values:ix_type=$ix_type($i)"
			
			puts $prf "last_ixvol=$last_ixvol($i)"
			puts $prf "ix_stem=$ix_stem($i)"
			puts $prf "ix_volsize=$ix_volsize($i)"
			puts $prf "ix_type=$ix_type($i)"
		}

		set ixpar "$index_name($i) 1 $ino $kw_type($i) $stem_fcn($i) $gsl_file($i)"

		set temp_field_list $field_list($i)
		set pos [ string first " " $temp_field_list ]
		set field_params ""

		while { $pos > 0 } {
			incr pos -1
			set next_field [ string range $temp_field_list 0 $pos]
			set field_params "$field_params$next_field 0 "
			set chars [ string length $next_field ]
			incr pos [ expr $chars + 1 ]
			set temp_field_list [ string range $temp_field_list $pos end ]
			set pos [ string first " " $temp_field_list ]
		}

		set field_params "$field_params$temp_field_list 0 -1"

		if { $sgf != "NOT_OPEN" } {
			write_history "save_current_values:$ixpar $field_params"
			puts $sgf "$ixpar $field_params"
		}
	}

	if { $prf != "NOT_OPEN" } {
		close $prf
	}

	if { $sgf != "NOT_OPEN" } {
		close $sgf
	}

	if { $menu_name != ".ask_to_save" } {
		cancel_menu $menu_name
	} else {
		destroy $menu_name
		validate_gsl_files
	}
}


proc read_current_values { menu_name } {
  global installation_root
  global exchange_format_filename
  global bss_parmpath
  global db_name
  global lastbibvol
  global bib_basename
  global bib_dir
  global bibsize
  global real_bibsize
  global display_name
  global explanation

  global nr

  global nf

  global field_abbrev
  global field_type

  global rec_mult
  global fixed
  global db_type
  global has_lims
  global maxreclen

  global ni

  global last_ixvol
  global ix_stem
  global ix_volsize
  global ix_type

  global index_name
  global field_list
  global kw_type
  global gsl_file
  global stem_fcn

  global field_entry_button

  global max_fields
  global max_indexes

	set open_error "cannot open for reading"
	set open_ok "opened for reading"

	set pr_file "$bss_parmpath/$db_name"
	set ft_file "$bss_parmpath/$db_name.field_types"
	set sg_file "$bss_parmpath/$db_name.search_groups"

	set ix_count 0

	if { [ file exists $pr_file ] } {
		if [ catch { open $pr_file r } prf ] {
			set message "Cannot open $pr_file for reading"
			write_history "read_current_values:$pr_file:$open_error"
			display_processing_message ".file_error" $message
		} else {
			write_history "read_current_values:$pr_file:$open_ok"

			set field_count 0

			##  Set up some dummy values for all params in case not in file

			set lastbibvol 0
			set bib_basename "$db_name.bib"
			set bib_dir "$installation_root/bibfiles"
			set bibsize 2047
			set real_bibsize 2047
			set display_name $db_name
			set explanation $db_name
			set nr 1000
			set nf 0
			set rec_mult 4
			set fixed 0
			set has_lims 0
			set maxreclen 2047

			while { [gets $prf next_line ] >= 0 } {

				set pos [string first = $next_line]
				incr pos -1
				set param [string range $next_line 0 $pos]
				incr pos 2
				set value [string range $next_line $pos end]

				write_history "read_current_values:$pr_file:$param:$value"
				
				if { $param == "name" } {
					set db_name $value
				} elseif { $param == "lastbibvol" } {
					set lastbibvol $value
					puts $lastbibvol
				} elseif { $param == "bib_basename" } {
					set lastbibvol $value
				} elseif { $param == "bib_dir" } {
					set bib_dir $value
				} elseif { $param == "bibsize" } {
					set bibsize $value
				} elseif { $param == "real_bibsize" } {
					set real_bibsize $value
				} elseif { $param == "display_name" } {
					set display_name $value
				} elseif { $param == "explanation" } {
					set explanation $value
				} elseif { $param == "nr" } {
					set nr $value
				} elseif { $param == "nf" } {
					set nf $value
				} elseif { $param == "f_abbrev" } {
					incr field_count 1
					set field_abbrev($field_count) $value
				} elseif { $param == "db_type" } {
					set db_type $value
				} elseif { $param == "ni" } {
					set ni $value
				} elseif { $param == "last_ixvol" } {
					incr ix_count 1
					set last_ixvol($ix_count) $value
				} elseif { $param == "ix_stem" } {
					set ix_stem($ix_count) $value
				}	elseif { $param == "ix_volsize" } {
					set ix_volsize($ix_count) $value
				} elseif { $param == "ix_type" } {
					set ix_type($ix_count) $value
				}
			}
			close $prf
		}
	} else {
		##
		##  No main parameter file

		write_history "read_current_values:$pr_file:no main parameter file"
	}

	if { [ file exists $ft_file ] } {
		if [ catch { open $ft_file r } ftf ] {
			puts stderr "Cannot open $ft_file for reading"
			write_history "read_current_values:$ft_file:$open_error"
		} else {
			write_history "read_current_values:$ft_file:$open_ok"

			set field_count 0
			while { [gets $ftf next_line ] >= 0 } {
				incr field_count 1
				set pos [ string first " " $next_line ]
				incr pos 1
				set next_line [ string range $next_line $pos end ]
				set field_type($field_count) $next_line
				write_history "read_current_values:$ft_file:\
						$field_count:$field_type($field_count)"
			}
			close $ftf
		}
	} else {
		##  No field types file so empty arrays
		write_history "read_current_values:$ft_file:no field types file"

		for { set f 0 } {$f <= $max_fields } { incr f 1 } {
			set field_abbrev($f) ""
			set field_type($f) ""
		}
	}

	if { [ file exists $sg_file ] } {
		if [ catch { open $sg_file r } sgf ] {
			puts stderr "Cannot open $sg_file for reading"
			write_history "read_current_values:$sg_file:$open_error"
		} else {
			write_history "read_current_values:$sg_file:$open_ok"

			set index_count 0
			while { [gets $ftf next_line ] >= 0 } {
				incr index_count 1

				write_history "read_current_values:$sg_file:$next_line"

				set pos [ string first " " $next_line ]
				incr pos -1
				set index_name($index_count) [ string range $next_line 0 $pos ]

				##  go past dummy fields 1 1
				incr pos 6
				set next_line [ string range $next_line $pos end ]

				set pos [ string first " " $next_line ]
				incr pos -1
				set kw_type($index_count) [ string range $next_line 0 $pos ]

				incr pos 2
				set next_line [ string range $next_line $pos end ]

				set pos [ string first " " $next_line ]
				incr pos -1
				set stem_fcn($index_count) [ string range $next_line 0 $pos ]

				incr pos 2
				set next_line [ string range $next_line $pos end ]

				set pos [ string first " " $next_line ]
				incr pos -1
				set gsl_file($index_count) [ string range $next_line 0 $pos ]

				incr pos 2
				set next_line [ string range $next_line $pos end ]

				set field_list($index_count) ""
				set pos [ string first " " $next_line ]

				while { $pos > 0 } {
					incr pos -1
					set fno [ string range $next_line 0 $pos ]

					if { $fno > 0  } {
						if { $field_list($index_count) == "" } {
							set field_list($index_count) "$field_list($index_count)$fno"
						} else {
							set field_list($index_count) "$field_list($index_count) $fno"
						}
					}

					set chars [ expr [ string length $fno ] + 3 ]
					incr pos $chars

					set next_line [ string range $next_line $pos end ]
					set pos [ string first " " $next_line ]
				}
			}
		}

		close $sgf
	} else {
		##  No search groups file - empty arrays

		write_history "read_current_values:$sg_file:no search groups file"

		for { set i 0 } {$i <= $max_indexes } { incr i 1 } {
			set index_name($i) ""
			set field_list($i) ""
			set kw_type($i) ""
			set gsl_file($i) ""
			set stem_fcn($i) ""
		}	
	}

	cancel_menu $menu_name
}


proc set_exchange_format_filename { pname sub_menu } {
  global exchange_format_filename
  global show_message_menu

	write_history "set_exchange_format_filename:$pname"

	set exchange_format_filename $pname

	set w ".exchange_format_file"
	set title ""
	set message1 "Set exchange format source file to"
	set message "$message1:\"$exchange_format_filename\""

	if { $show_message_menu == "on" } {
		display_process_message $w $message
	}
}

proc source_data_list { sub_menu } {
	##
	##  Make up a sub-menu of source data file names by reading the contents
	##  of the <exchange_format_path> and forming a sub-menu of these names.
	##
	##  Procedure to list all existing databases on the basis
	##  of their parameter files. Possible files that will
	##  exist are:
	##
	##  a) <db_name>
	##  b) <db_name>.field_types
	##  c) <db_name>.search_groups
	##
	##  (a) must exist. (b) and (c) may exist if (a) exists.

  global exchange_format_path
		
	set parameters_pattern "$exchange_format_path/*"

	##  1) delete all previous sub-menu entries else the contents of
	##     PROFILES_DIR will be appended to the previous list.

	$sub_menu delete 0 "last"

	##  2) The function 'glob' returns a list for each filename
	##     that matches the pattern. e.g.
	##
	##              exchange_format_path = /project/okapi/OkapiNet/databases
	##        parameters_pattern = <exchange_format_path>/*
	##
	##     The variable pname is set equal to each entry in turn
	##
	##  3) Against each sub-menu entry set the command:
	##
	##     load_profile $pname
	##
	##     This function will call the reciprocal function
	##     in 'gui.cc' which will load the profile and then
	##     search the current database

	foreach f [glob $parameters_pattern] {
		set path_len [ string length $exchange_format_path ]
		incr path_len
		set pname [ string range $f $path_len end ]

		write_history "source_data_list:$pname"

		$sub_menu add command \
				-label "  $pname  " \
				-command "set_exchange_format_filename $pname $sub_menu"
	}

	$sub_menu add separator

	$sub_menu add command \
      -label "  None Of These  " \
      -command "cancel_menu $sub_menu"
}

proc load_db_params { dname sub_menu } {
	##
	##  $dname is returned from pop-up menu list selection
	##  First set the global variable db_name to this value.

  global db_name

	write_history "load_db_params:$dname"

	set db_name $dname
	read_current_values $sub_menu

##	set w ".loaded_params"
##	set message1 "Loaded parameter files for database"
##	set message  "$message1 \"$db_name\"."
##
##	display_process_message $w $message
}


proc file_test { pname } {
	##  First set variable file_ok to 0. It will be set to 1 if
	##  the file is a valid parameter file

  global bss_parmpath

	set full_pname "$bss_parmpath/$pname"
	set file_ok 0

	##  Test to see if it is readable
	##

	if { [ file readable $full_pname ] } {

		write_history "file_test:$full_pname:readable"
		
		##
		##  Test to see if it has a "field_types" or "search_groups" extension
		
		set file_extension [ file extension $full_pname ]
		
		if { $file_extension == ".field_types" } {
			write_history "file_test:$full_pname.field_types:exists"
			
			set file_ok 0
		} elseif { $file_extension == ".search_groups" } {
			write_history "file_test:$full_pname.search_groups:exists"
			
			set file_ok 0
		} else {
			##
			##  Determine length of file. Must be at least 7 bytes long
			##  if it is to start with "name=<db_name><nl>"
			
			set flength [ file size $full_pname ]
			
			if { $flength > 7 } {
				if { [ file isfile $full_pname ] } {
					set file_id [ open $full_pname r ]
					set line1 [ gets $file_id ]
					
					if { [ string length $line1 ] > 5 } {
						set param [ string range $line1 0 4 ]
						
						if { $param == "name=" } {
							set file_ok 1
						}
					}
					
					close $file_id
				}
			}
		}
	} else {
		##  "full_pname" cannot be opened for reading"
		write_history "file_test:$full_pname:not readable"
	}

	return $file_ok
}


proc database_list { sub_menu } {
	##
	##  Make up a sub-menu of (converted) database names by reading
	##  the contents of the <bss_parmpath> and forming a sub-menu
	##  of the names.
	##

  global bss_parmpath
		

	set parameters_pattern "$bss_parmpath/*"

	write_history "database_list:pattern:$parameters_pattern"

	##  1) delete all previous sub-menu entries else the contents of
	##     PROFILES_DIR will be appended to the previous list.

	$sub_menu delete 0 "last"

	##  2) The function 'glob' returns a list for each filename
	##     that matches the pattern. e.g.
	##
	##              bss_parmpath = /project/okapi/OkapiNet/databases
	##        parameters_pattern = <bss_parmpath>/*
	##
	##     The variable pname is set equal to each entry in turn
	##
	##  3) Against each sub-menu entry set the command:
	##
	##     load_profile $pname
	##
	##     This function will call the reciprocal function
	##     in 'gui.cc' which will load the profile and then
	##     search the current database

	foreach f [glob $parameters_pattern] {
		set path_len [ string length $bss_parmpath ]
		incr path_len
		set pname [ string range $f $path_len end ]

		set valid_file [ file_test $pname ]

		if { $valid_file == "1" } {
			write_history "database_list:pattern:$pname"

			$sub_menu add command \
					-label "  $pname  " \
					-command "load_db_params $pname $sub_menu"
		}
	}

	$sub_menu add separator

	$sub_menu add command \
      -label "  None Of These  " \
      -command "cancel_menu $sub_menu"
}


proc gsl_file_list { sub_menu } {
	##
	##  Make up a sub-menu of gsl file names by reading the contents
	##  of the <bss_parmpath> and forming a sub-menu of those
	##  that contain the string 'gsl'.

  global bss_parmpath

	set parameters_pattern "$bss_parmpath/*gsl*"

	write_history "gsl_file_list:$parameters_pattern"

	##  1) delete all previous sub-menu entries else the contents of
	##     PROFILES_DIR will be appended to the previous list.

	$sub_menu delete 0 "last"

	##  2) The function 'glob' returns a list for each filename
	##     that matches the pattern. e.g.
	##
	##              bss_parmpath = /project/okapi/OkapiNet/databases
	##        parameters_pattern = <bss_parmpath>/*gsl*
	##
	##     The variable pname is set equal to each entry in turn

	foreach f [glob $parameters_pattern] {
		set path_len [ string length $bss_parmpath ]
		incr path_len
		set pname [ string range $f $path_len end ]

		##  set valid_file [ file_test $pname ]

		set valid_file 1

		if { $valid_file == "1" } {

			write_history "gsl_file_list:$pname"

			if { $pname == "gsl.empty" } {
				$sub_menu add command \
						-label "  NONE  " \
						-command "set_gsl_file $pname $sub_menu"
			} else {
				$sub_menu add command \
						-label "  $pname  " \
						-command "set_gsl_file $pname $sub_menu"
			}
		}
	}

	$sub_menu add separator

	$sub_menu add command \
      -label "  None Of These  " \
      -command "cancel_menu $sub_menu"
}


########################################################################
##                                                                    ##
##  Procedures from TREC interfaces that might be useful              ##
##                                                                    ##
########################################################################

##  1. Load file into a text widget

proc load_file { w file } {
	##
	##  Ensure window is not read only

	$w configure -state normal

	$w delete 1.0 end

	set fptr [open $file]

	while { ![eof $fptr]} {
		$w insert end [read $fptr 1000]
	}

	close $fptr

	write_history "load_file:$file:OK"
}


##  Quit application

proc quit { } {
	set w ".quit_message"
	set message "Quit -- Click OK to confirm\n"

	write_history "quit:OK"

##	display_process_message $w $message

	quit2
}

proc quit2 { } {

	exit
}


proc pack_main_buttons { screen previous_button next_button } {
	##  Values of $screen :
	##
	##    intro
	##    db_params1
	##    db_params1
	##    field_entry
	##    index_entry
	##    convert_db
	##    create_index
	##    finish

	##  First unpack all buttons

  global menu_name
  global menu_x
  global menu_y

	write_history "pack_main_buttons:unpack_old_screen_buttons"

	pack forget .main_buttons.previous
	pack forget .main_buttons.options
	pack forget .main_buttons.next

	write_history "pack_main_buttons:$screen"

	if { $screen != "intro" } {

		pack .main_buttons.previous \
				-in .main_buttons \
				-side left \
				-expand 1 \
				-fill x \
				-ipadx 4 \
				-ipady 2
		
		pack .main_buttons.options \
				-in .main_buttons \
				-side left \
				-expand 1 \
				-fill x \
				-ipadx 4 \
				-ipady 2
		
		pack .main_buttons.next \
				-in .main_buttons \
				-side left \
				-expand 1 \
				-fill x \
				-ipadx 4 \
				-ipady 2

		.main_buttons.previous configure \
				-state normal \
				-text "PREVIOUS:     $previous_button"

		.main_buttons.next configure \
				-state normal \
				-text "NEXT:     $next_button"

	} else {

		pack .main_buttons.previous \
				-in .main_buttons \
				-side left \
				-expand 1 \
				-fill x \
				-ipadx 4 \
				-ipady 2
		
		pack .main_buttons.options \
				-in .main_buttons \
				-side left \
				-expand 1 \
				-fill x \
				-ipadx 4 \
				-ipady 2
		
		pack .main_buttons.next \
				-in .main_buttons \
				-side left \
				-expand 1 \
				-fill x \
				-ipadx 4 \
				-ipady 2

		.main_buttons.previous configure \
				-state disabled \
				-text "PREVIOUS: "

		.main_buttons.next configure \
				-state normal \
				-text "NEXT:     $next_button"
		
	}

	bind .main_buttons.options <Button-1> {
		determine_coords %W
		post_menu $menu_name $menu_x $menu_y
	}

	##  Button functionality set in the calling procedure
}

										
proc gsl_error { line_count full_line error_message errf_id errf_open } {
	set lc_len [ string length $line_count ]

	write_history "gsl_error:$error_message at line $line_count"
	write_history "gsl_error:$full_line"

	if { $lc_len == 1 } {
		set message "   $error_message"
	} elseif { $lc_len == 2 } {
		set message "    $error_message"
	} elseif { $lc_len == 3 } {
		set message "     $error_message"
	} else {
		set message "      $error_message"
	}

	if { $errf_open == 0 } {
		puts $errf_id "$line_count: \[$full_line\]"
		puts $errf_id "$message"
	}

	puts "$line_count: \[$full_line\]"
	puts "$message"
}


proc validate_gsl { filename } {
	##
	##  All lines must start with a valid GSL class code [FGHIS]
	##  or a # symbol indicating a comment
	##
	##  Each line must end in ",:" followed by zero, one or more
	##  whitespace characters.
	##
	##  Only "S" lines may contain more than one term.
	##
	##  Terms must start and end with an alphanumeric character.
	##  Other than that they may contain and of the printable
	##  ASCII character set.
	##
	##  Sample lines
	##
	##  Fsince,:
	##  GLouisiana State University,:
	##  Halong,:
	##  Hdon't,:
	##  Itherefore,:
	##  Scba, cost benefit analysis,:
	##  Scbi, confederation of british industries,:
	##  Sccl, common comand language,:
	##  Scd-rom, cd rom,:
	##
	##  file_ok =  0  --  OK
	##  file_ok = -1  --  ERROR

  global gsl_codes
  global valid_gsl

	set line_count 0

	write_history "validate_gsl:$filename"

	if { [ file exists $filename ] } {
		puts "GSL FILENAME = \[$filename\] EXISTS"
		
		if [ catch { open $filename r } gsl_file_id ] {
			puts stderr "Cannot open $filename for reading"
			set message "Cannot open $filename for reading"
			set w ".gsl_error_file"
			display_process_message $w $message
		} else {
			set file_ok 0
			
			##  Set error filename - check can open for writing
			
			puts "GSL FILENAME = \[$filename\] EXISTS for reading"
			
			set error_file "$filename.errors"
			
			##			if [ catch { open $error_file w } errf_id ] {
			##				set errf_open 0
			##			} else {
			##				set errf_open -1
			##				set message "Cannot open $error_file for writing"
			##				puts stderr "Cannot open $error_file for writing"
			##				set w ".gsl_error_file"
			##				display_process_message $w $message
			##			}
					
			set errf_id [ open $error_file w ]
			set errf_open 0
			
			while { [gets $gsl_file_id next_line ] >= 0 } {
				incr line_count 1
				
				##  First trim whitespace, if any,  from end of line
				
				set next_line [ string trimright $next_line ]
				set full_line $next_line
				
				##  puts "$line_count: \[$full_line\]"
				
				##  Pick out first character on line.
				##  Must be a legal gsl class code or a # (comment character)
				
				set fc [ string range $next_line 0 0 ]
				
				if { $fc != "#" } {
					##  Not a comment
					set legal_code [ string first $fc $gsl_codes ]
					
					if { $legal_code != -1 } {
						##  Legal GSL Class Code. Check rest of line
						##  First remove GSL Class Code
						
						set next_line [ string range $next_line 1 end ]
						
						##  Now check that the characters ",:" occur only once in the
						##  line and they are the last two characters in the line.
						
						set pos1 [ string first ",:" $next_line ]
						set pos2 [ string last ",:" $next_line ]
						
						if { $pos1 == $pos2 && $pos1 > 0 } {
							##  Line contains ",:" only once and has at least
							##  1 character before the ","
							
							set line_len [ string length $next_line ]
							set test_pos [ expr $line_len - 2 ]
							
							if { $pos1 == $test_pos } {
								##  line terminated ok
								##  Set $next_line to the terms only
								
								incr pos1 -1
								set next_line [ string range $next_line 0 $pos1 ]
								
								##  Check that first and last characters are alphanumerics
								
								set line_len [ string length $next_line ]
								set first [ string range $next_line 0 0 ]
								set lc [ expr $line_len - 1 ]
								set last [ string range $next_line $lc $lc ]
								
								set fres [regexp {[ 0-9A-Za-z]+} $first fmatch]
								set lres [regexp {[ 0-9A-Za-z]+} $last lmatch]
								
								if { $fres == 1 || $lres == 1 } {
									if { $fc != "S" } {
										
										if { $fc == "G" } {
											set result [regexp {[ -~]+} $next_line matched]
										} else {
											set result [regexp {[ -~]+} $next_line matched]
										}
										
										if { $result == 1 && $matched == $next_line } {
											##  next_line contains only alphanumerics - ok
										} else {
											##  incorrect entry
											set message "Incorrect entry"
											gsl_error $line_count $full_line \
													$message $errf_id $errf_open
											
										}
									} else {									
										##  (S)ynonym groups - terms delimited by ", "
									}
								} else {
									##  Terms don't start or end with alphanumerics
									
									if { $fc != "S" && $fc != "P" } {
										puts "f = $first, L = $last"
										set message "Starts or ends with a non-alphanumeric char"
										gsl_error $line_count $full_line $message \
												$errf_id $errf_open
									}
								}
							} else {
								set message "Illegally terminated line."
								gsl_error $line_count $full_line $message \
										$errf_id $errf_open
								set file_ok -1
							}
						} else {
							## line contains no term
							set message "Line contains no term"
							gsl_error $line_count $full_line $message \
									$errf_id $errf_open
							set file_ok -1
						}
					} else {
						##  Invalid GSL Class Code
						
						puts "$line_count: $next_line"
						set lc_len [ string length $line_count ]
						set message "*  Invalid GSL class code"
						gsl_error $line_count $full_line $message \
								$errf_id $errf_open
						
						set file_ok -1
					}
				} else {
					##  puts "$line_count: A comment"
				}
			}
		}
		
		close $gsl_file_id
		
		if { $errf_open } {
			close $errf_id
		}
	}

	if { $file_ok == 0 } {
		write_history "validate_gsl:$filename:OK"
		set valid_gsl 0
	} else {
		write_history "validate_gsl:$filename:ERROR"
		set valid_gsl -1
	}
}


proc validate_gsl_files { } {
  global bss_parmpath
  global ni
  global gsl_file
  global valid_gsl

	set message "Validating GSL files\n"
	set error_message "ERRORS were found in:\n"

	write_history "validate_gsl_files"

	set error_count 0

	set back "NavyBlue"
	set fore "Yellow"
	set window_name ".validate_gsl_files"

	processing_message $window_name $message $back $fore

	set validation_message "Validating\n"

	for { set i 1 } {$i <= $ni } { incr i 1 } {
		set filename "$bss_parmpath/$gsl_file($i)"
		set validation_message "$validation_message\n$gsl_file($i)"

		processing_message $window_name $validation_message $back $fore

		puts "validate_gsl $filename"
		validate_gsl $filename

		if { $valid_gsl == -1 } {
			set message "$message\n$i) $gsl_file($i)"
			incr error_count 1
		}
	}

	destroy $window_name

	if { $error_count > 0 } {
		set message "$message\n\nErrors in file \[$filename.errors\]"
		set w ".gsl_file_error"
		display_process_message $w $message
	}
}


