/*
**  extract.c
**
**  Program to extract all terms from a single user rel.
*/

#define G_EXTERN extern

#include "defs.h"

#define OLD_TERM 0
#define NEW_TERM 1
#define OLD_AND_NEW 2

#undef RESULTS_BUFFER_SIZE
#define RESULTS_BUFFER_SIZE 75 * 1024

int parse_show_file(char *, struct query_terms *, int);
void write_new_file(struct query_terms *, int, int, int, int *);

/*
**  External function headers
*/

G_EXTERN void show_time(char *, int);
G_EXTERN int calc_little_r(int);
G_EXTERN double calc_wgt(int, int, char);
G_EXTERN double calc_rsv(double, int r, char);
G_EXTERN void terminate_run(char *, char *, int);
G_EXTERN int ReadTermsetFile(int);
G_EXTERN void sort_terms(struct query_terms *, int, int);
G_EXTERN void malloc_error(char *, char *, char *, int);
G_EXTERN void malloc_success(char *, char *, char *, int);

void extract_terms(int rj, int previous_rel,
									 int full_doclen, int part_doclen) {

	int debugging = atoi(getenv("EXTRACT_TERMS_DEBUG"));

  /*
  **  Two files, one consisting of the whole document and one of
  **  the 'best' passage, are written in the show_record() function.
  **
  **  The file used will depend on the relj, R_FULL or R_PASSAGE.
  **
  **  The appropriate file is parsed and the extracted terms read
  **  into the structure 'extracted_terms[]'.
  **
  **  The original termset, already in the structure termset[],
  **  is merged with these extracted terms. The merged terms are
  **  written to the termset file.
  **
  **  After the merge process this file is read back into the
  **  structure termset[].
	**
  **  Character and term counters for for loops
  */

  int c;
  int i;

	/*
	**  Extra charcaters for the pre- and post- display of a record
	**  for which the full document is too long but passage is ok
	*/

	int buffer_size;

	int extra_chars = 500;

  int chars_read;
  int length;
  int old_r = 0;
  int no_extracted_terms;

  int old = 0;
  int new = 0;

  int term_no = TERM_NO;
  int bss_set = NONE_ASSIGNED;
  double weight = (double) NONE_ASSIGNED;
  double rsv = (double) NONE_ASSIGNED;
  int np = NONE_ASSIGNED;
  int little_r = 1;
  int term_flag = OLD_TERM;

	int termset_size = 0;

  /*
  **  The following variable, 'value', is set (for each term) to
  **  1 or 0 according to whether or not the term is in the current.
  **  rel. This value is written as the BIGRth value to the term
  **  occurrence matrix in the termset file.
  */
  
  int value = 0;

  char temp[MAX_TERM_LENGTH];

  char attribute;
  char term_type = EXTRACT_TERM;

  char *results_ptr;
  
  char parsed[MAX_TERM_LENGTH];
  char source[MAX_TERM_LENGTH];
  char gsl_class;
  char operation = NO_OP;

  char passage_file[FILENAME_SIZE];
  char full_minus_passage_file[FILENAME_SIZE];

	int return_code;

	char bss_command[SMALL_BUFFER];
	char bss_result[SMALL_BUFFER];

  struct query_terms *extracted_terms;

  show_time("extract_terms", START);

	buffer_size = atoi(getenv("MAX_TERMS_PER_DOC")) * sizeof(struct query_terms);
	
	if ((extracted_terms = (struct query_terms *) malloc(buffer_size)) == NULL) {
		malloc_error("ExtractTerms()", "extract_terms()",
								 "extracted_terms", buffer_size);
	}
	else if (debugging) {
		malloc_success("ExtractTerms()", "extract_terms()",
									 "extracted_terms", buffer_size);
	}
	
	/*
	**  Open termset_file to write merged terms to
	*/
	
	tf = fopen(termset_file, "w");

	if (FEEDBACK) {
		/*
		**  Relevance feedback required
		**
		**  Set up filenames
		*/

		if (debugging) {
			fprintf(stderr, "extract_terms() :: RELEVANCE FEEDBACK ON\n");
		}
		
		sprintf(passage_file, "%s.pass", show_file);
		sprintf(full_minus_passage_file, "%s.fminusp", show_file);
		
		EXTRACTED_TERMS = TRUE;
		
		/*
		**  Set up appropriate file to extract terms from.
		**  The function only gets called for:
		**
		**     R_FULL : 0     0 -> F    show_file
		**              2     P -> F    full_minus_pass_file
		**              3     N -> F    show_file
		**
		**  R_PASSAGE : 0     0 -> P    passage_file
		**              1     F -> P    full_minus_pass_file
		**              3     N -> P    passage_file
		**
		**      R_NOT : 1     F -> N    show_file
		**              2     P -> N    passage_file
		*/
		
		switch (rj) {
			
		  case R_FULL :
			
				switch (previous_rel) {
				
			    case 0 : case 3 :
						/*
						**  0 -> F, N -> F
						*/
						no_extracted_terms =
							parse_show_file(show_file, extracted_terms, full_doclen);
						break;

				  case 2 : 
						/*
						**  P -> F
						*/
						no_extracted_terms =
							parse_show_file(full_minus_passage_file,
															extracted_terms,	full_doclen);
						
						break;
				}

				break;
			
	    case R_PASSAGE :

			  switch (previous_rel) {

				  case 0 : case 3 :
						/*
						**  0 -> P, N -> P
						*/
						no_extracted_terms =
							parse_show_file(passage_file, extracted_terms,
															part_doclen + extra_chars);
						break;
						
			    case 1 :
						/*
						**  F -> P
						*/
						no_extracted_terms =
							parse_show_file(full_minus_passage_file,
															extracted_terms, full_doclen);
						break;
				}
				
				break;
				
	    case R_NOT :

			  switch (previous_rel) {

			    case 1 :
						/*
						**  F -> N
						*/
						no_extracted_terms =
							parse_show_file(show_file, extracted_terms, full_doclen);
						break;
						
			    case 2 :
						/*
						**  P -> N
						*/
						no_extracted_terms =
							parse_show_file(passage_file, extracted_terms,
															part_doclen + extra_chars);
				}
				
				break;
		}

		/*
		**  Now merge the original terms -- termset[] -- with
		**  the new terms -- extracted_terms[]
		**
		**  Merged terms are written to FILE tf (termset_file)
		*/
		
		while (old < TERM_NO) {
			if (new > no_extracted_terms) {
				/*
				**  We've reached the end of the extracted terms.
				**  => these are all old terms which do not exist
				**  in the current rel => little_r correct, value = 0
				*/
				
				value = 0;
				term_flag = OLD_TERM;
				
				for (i = old; i < TERM_NO; i++) {
					/*
					**  First recalculate weight and rsv since, although
					**  little_r is unchanged, BIGR has altered.
					*/
					
					(termset + i)->weight =
						calc_wgt((termset + i)->np,
										 (termset + i)->little_r,
										 (termset + i)->term_type);
					
					(termset + old)->rsv =
						calc_rsv((termset + i)->weight,
										 (termset + i)->little_r,
										 (termset + i)->term_type);
					
					write_new_file(termset, i, value, term_flag, &termset_size);
				}
				
				break;
			}
			
			/*
			**  If we get here there are still terms in both
			**  termset[] and extracted_terms[]
			*/
			
			if (strcmp((termset + old)->parsed,
								 (extracted_terms + new)->parsed) < 0) {
				/*
				**  Exists in termset[] but not in extracted_terms[]
				**
				**  There are two possibilities:
				**
				**  1) It is either a single word or a GSL phrase.
				**     => it is not in the rel,
				**        little_r is correct, set value = 0
				**
				**  2) It is a user-defined phrase and so might be in the rel.
				**     Recalculate little_r. If little_r increases then the
				**     phrase is in the rel so set value = 1.
				**
				**  We will in fact set value by default to zero. It will only
				**  be changed to 1 after case (2).
				**
				**  Need to look at (extracted_terms + new)-> again
				**  so don't increment nt.
				*/

				term_flag = OLD_TERM;
				value = 0;
				old_r = (termset + old)->little_r;
				
				if (strchr('ABS', (termset + old)->operation) != NULL) {
					(termset + old)->little_r =
						calc_little_r((termset + old)->bss_set);
					
					value = ((termset + old)->little_r > old_r) ? 1 : 0;
				}
				
				(termset + old)->weight =
					calc_wgt((termset + old)->np,
									 (termset + old)->little_r,
									 (termset + old)->term_type);
				
				(termset + old)->rsv =
					calc_rsv((termset + old)->weight,
									 (termset + old)->little_r,
									 (termset + old)->term_type);
				
				write_new_file(termset, old, value, term_flag, &termset_size);
				old++;
			}
			else if (strcmp((termset + old)->parsed,
											(extracted_terms + new)->parsed) > 0) {
				/*
				**
				**  The term '(extracted_terms + new)->' is not in the termset[]
				**  file => write it to file if it is an index term.
				**
				**  Need to look at termset[old] again
				**    => don't increment old.
				**
				**  Since this is a new extracted term we must generate
				**  a set for the term and set (little_r, value) = (1, 1)
				*/
				
				if (strlen((extracted_terms + new)->parsed) > 0) {
					sprintf(bss_command, "f t=%s", (extracted_terms + new)->parsed);
					
					if ((return_code = i0(bss_command, bss_result)) < 0) {
						terminate_run("extract_terms()", bss_command, return_code);
					}
					
					term_flag = NEW_TERM;
					value = 1;
					
					results_ptr = bss_result;
					sscanf(results_ptr, "S%d np=%d",
								 &(extracted_terms + new)->bss_set,
								 &(extracted_terms + new)->np);
					
					if (debugging) {
						fprintf(stderr,
										"extracted term:: np=%d\n", (extracted_terms + new)->np);
					}
					
					if ((extracted_terms + new)->np > 0) {
						/*
						**  Only write to file if it is an extracted term, i.e. np > 0
						*/
						(extracted_terms + new)->little_r = 1;
						
						(extracted_terms + new)->weight =
							calc_wgt((extracted_terms + new)->np,
											 (extracted_terms + new)->little_r,
											 (extracted_terms + new)->term_type);
						
						(extracted_terms + new)->rsv =
							calc_rsv((extracted_terms + new)->weight,
											 (extracted_terms + new)->little_r,
											 (extracted_terms + new)->term_type);
						
						if (debugging) {
							fprintf(stderr,
											"extracted term:: wgt = %.3f, rsv = %.3f\n",
											(extracted_terms + new)->weight,
											(extracted_terms + new)->rsv);
						}
						
						write_new_file(extracted_terms, new, value,
													 term_flag, &termset_size);
					}
				}
				new++;
			}
			else {
				/*
				**  (termset + old)->parsed == (extracted_terms + new)->parsed
				** 
				**  In both termset and extracted terms =>
				**    must both be single terms or a GSL phrase.
				**
				**  The appropriate file of extracted terms will be opened
				**  as described above, according to values passed to the
				**  program (from make_rj()) in 'previous_rel' and 'rj'
				**
				**  The action taken will depend on the values of the
				**  current and previous rels.
				*/
				
				term_flag = OLD_AND_NEW;
				old_r = (termset + old)->little_r;
				
        switch (rj) {
	  		  case R_FULL :

					  switch (previous_rel) {

					    case 0 : case 2 : case 3 :
								/*
								**  0 -> F, P -> F, N -> F
								**  Increment little_r, set value to 1 (it's in the rel)
								*/
								(termset + old)->little_r++;
								value = 1;
								break;
								
					    case 1 :
								/*
								**  F -> F
								*/
								break;
								
						}
						
						break;

			    case R_PASSAGE :

					  switch (previous_rel) {

					    case 0 : case 3 :
								/*
								**  0 -> P, N -> P
								**  Increment little_r, set value to 1 (it's in the rel)
								*/
								(termset + old)->little_r++;
								value = 1;
								break;
								
					    case 1 : 
								/*
								**  F -> P
								**  Decrement little_r since we are considering extracted
								**  terms that are not in the passage. Set value to zero.
								*/
								
								(termset + old)->little_r--;
								value = 0;
								
								/*
								**  Check to make sure that little_r has not dropped
								**  below zero. It shouldn't have done but .....
								*/
								
								if ((termset + old)->little_r < 0) {
									(termset + old)->little_r = 0;
								}
								
								/*
								**  If little_r == 0 term has been effectively removed
								**  from the list of candidate query terms. Thus
								**  set operation field to zero
								*/
								
								if (((termset + old)->little_r == 0) &&
										(toupper((termset + old)->term_type) != USER_TERM)) {
									(termset + old)->operation =
										tolower((termset + old)->operation);
								}
								
								break;

					    case 2 :
								/*
								**  P -> P    do nothing.
								*/
								break;
						}
						
  			  case R_NOT :

					  switch (previous_rel) {

					    case 0 : case 3 :
								/*
								**  0 -> N, N -> N,   do nothing.
								*/
								break;

					    case 1 : case 2 :
								/*
								**  F -> N
								**  Decrement little_r since we are considering extracted
								**  terms that are not in the passage. Set value to zero.
								*/
								
								(termset + old)->little_r--;
								
								/*
								**  Check to make sure that little_r has not dropped
								**  below zero. It shouldn't have done but .....
								*/

								if ((termset + old)->little_r < 0) {
									(termset + old)->little_r = 0;
								}
								
								/*
								**  If little_r == 0 term has been effectively removed
								**  from the list of candidate query terms. Thus
								**  set operation field to zero
								*/
								
								if (((termset + old)->little_r == 0) &&
										(toupper((termset + old)->term_type) != USER_TERM)) {
									(termset + old)->operation =
										tolower((termset + old)->operation);
								}
								
								break;
						}
						
						break;
				}
				
				(termset + old)->weight =
					calc_wgt((termset + old)->np,
									 (termset + old)->little_r,
									 (termset + old)->term_type);
				
				(termset + old)->rsv =
					calc_rsv((termset + old)->weight,
									 (termset + old)->little_r,
									 (termset + old)->term_type);
				
				if (debugging) {
					fprintf(stderr, "extract_terms():: wgt = %.3f, rsv = %.3f\n",
									(termset + old)->weight, (termset + old)->rsv);
				}
				
				write_new_file(termset, old, value, term_flag, &termset_size);
				old++;
				new++;
			}
		}
		
		if (new < no_extracted_terms) {
			/*
			**  Seen all the old termset[] but there are
			**  still some left in extracted_terms[].
			**
			**  These are all new terms that are in the current rel.
			**
			**  => generate set etc. and set value to 1
			*/
			
			for (i = new; i < no_extracted_terms; i++) {
				if (strlen((extracted_terms + i)->parsed) > 0) {
					(extracted_terms + i)->term_type = EXTRACT_TERM;
					sprintf(bss_command, "f t=%s", (extracted_terms + i)->parsed);
					
					if ((return_code = i0(bss_command, bss_result)) != 0) {
						terminate_run("extract_terms()", bss_command, return_code);
					}
					
					term_flag = NEW_TERM;
					value = 1;
					
					results_ptr = bss_result;
					
					sscanf(results_ptr, "S%d np=%d",
								 &(extracted_terms + i)->bss_set,
								 &(extracted_terms + i)->np);
					
					if ((extracted_terms + i)->np > 0) {
						/*
						**  Only write term to file if it is an index term, i.e. np > 0
						*/
						(extracted_terms + i)->little_r = 1;
						
						(extracted_terms + i)->weight =
							calc_wgt((extracted_terms + i)->np,
											 (extracted_terms + i)->little_r,
											 (extracted_terms + i)->term_type);
						
						(extracted_terms + i)->rsv =
							calc_rsv((extracted_terms + i)->weight,
											 (extracted_terms + i)->little_r,
											 (extracted_terms + i)->term_type);
						
						if (debugging) {
							fprintf(stderr, "extract_terms():: wgt = %.3f, rsv = %.3f\n",
											(termset + i)->weight, (termset + i)->rsv);
						}
						
						write_new_file(extracted_terms, i, value,
													 term_flag, &termset_size);
					}
				}
			}
		}
	}
	else {
		fprintf(stderr, "extract_terms() :: RELEVANCE FEEDBACK OFF\n");
		/*
		**  BIGR has been updated before this function is called.
		**  Re-calculate little_r, weight & RSV for all members
		**  of termset[]
		*/

		for (i = 0; i < TERM_NO; i++) {
			term_flag = OLD_TERM;
			value = 0;
			old_r = (termset + i)->little_r;

			(termset + i)->little_r =
				calc_little_r((termset + i)->bss_set);
			
			value = ((termset + i)->little_r > old_r) ? 1 : 0;
				
			(termset + i)->weight =
				calc_wgt((termset + i)->np,
								 (termset + i)->little_r,
								 (termset + i)->term_type);
			
			(termset + i)->rsv =
				calc_rsv((termset + i)->weight,
								 (termset + i)->little_r,
								 (termset + i)->term_type);
				
			write_new_file(termset, i, value, term_flag, &termset_size);
		}
	}

	fclose(tf);
		
	/*
	**  Re-read termset_file back into termset_file[]
	**
	**  FALSE indicates that this is not an old query being read in
	*/
	
	TERM_NO = ReadTermsetFile(FALSE);
	
  show_time("extract_terms", FINISH);
}


int parse_show_file(char show_file[],
										struct query_terms *extracted_terms,
										int text_length) {
  /*
  **  Integer loop variables
  */

	int debugging = atoi(getenv("PARSE_SHOW_FILE_DEBUG"));

  int c;
  int i;
  int t;

  int distinct_terms;

  int bss_set;
  int length;
  int little_r;
  int no_terms;
  int np;
  int rsv = NONE_ASSIGNED;
  int term_count = 0;
  int term_no = TERM_NO;
  double weight = (double) NONE_ASSIGNED;

  int chars_read;
	int buffer_size;

  char attribute;
  char term_type = EXTRACT_TERM;
  char gsl_class;
  char operation = NO_OP;

  char temp[MAX_TERM_LENGTH];
  char parsed[MAX_TERM_LENGTH];
  char previous[MAX_TERM_LENGTH];
  char source[MAX_TERM_LENGTH];

  char *results_ptr;
  char text_line[RESULTS_BUFFER_SIZE];

	int return_code;

	char *bss_command;
	char *bss_result;

  struct query_terms *reldoc_terms;

  FILE *sfile;

	buffer_size = atoi(getenv("MAX_TERMS_PER_DOC")) * sizeof(struct query_terms);

	if ((reldoc_terms = (struct query_terms *) malloc(buffer_size)) == NULL) {
		malloc_error("ExtractTerms()", "parse_show_file",
								 "reldoc_terms", buffer_size);
	}
	else if (debugging) {
		malloc_success("ExtractTerms()", "parse_show_file",
									 "reldoc_terms", buffer_size);
	}
	
  /*
  **  Open show_file for reading.
  */

  if ((sfile = fopen(show_file, "r")) == NULL) {
    sprintf(ERROR_MESSAGE,
						"EXTRACT: Can't open file '%s' for reading",
						show_file);
    printf("%s\n", ERROR_MESSAGE);
    exit(-1);
  }

  text_line[0] = '\0';

  /*
  **  Read terms in show_file and make into one string
  **  to be stored in 'text_line', e.g.
  **
  **      air traffic controllers at Kennedy Airport .....
  **
  **  This string is 'superparsed' and the results, one line
  **  per term, is written to ..... as:
  **
  **    term_no, bss_set, np, little_r, weight, rsv,
  **    term_type, attribute, source, parsed, operation
  **
  **  Note: these are set to:
  **
  **  bss_set, np, weight, rsv      NONE_ASSIGNED
  **
  **   little_r = 1    (this is an extracted term => occurs in at
  **                   least 1 document).
  **  term_type = EXTRACTED_TERM
  **  attribute = gsl_class
  **  operation = NO_OP
  */

  while (!feof(sfile)) {
    if (fgets(temp, MAX_TERM_LENGTH, sfile) == NULL) {
      break;
    }
    else {
      temp[strlen(temp) - 1] = '\0';
      if (strlen(temp) > 0) {
				sprintf(text_line + strlen(text_line), "%s ", temp);
      }
    }
  }
	
  fclose(sfile);

  /*
  **  Now superparse text_line.
  **
  **  The output from superparse is in the form:
  **
  **     337 t=prison c=N s=Prison t=guard c=N s=Guards ...
  **
  **  where 337 is the number of terms parsed.
  */

	/*
	**  First allocate buffers for bss_command and bss_result.
	*/

	buffer_size = strlen(text_line) + 10;

	if ((bss_command = (char *) malloc(buffer_size)) == NULL) {
		malloc_error("extract.c", "parse_show_file", "bss_command", buffer_size);
	}

	/*
	**  Allocate buffers for bss_result. The length of bss_result
	**  buffer is determined by the structure of the result returned.
	**
	**  Results are of the form:
	**
	**     12345 t=air c=N s=air t=traffic c=N s=traffic .....
	**
	**  The first item is the number of terms followed by a
	**  space; take 8 chars for this.
	**
	**  Assuming an average word length of 4 chars (an under-estimate).
	**  => no_terms = text_length / 4
	**
	**  Each term requires 10 extra characters to cater for the
	**  't= ', 'c=G ' and 's= '.
	*/

	buffer_size =
		2 * text_length + 6 + 10 * (int) (0.5 + ((float) text_length / 4.0));

	if ((bss_result = (char *) malloc(buffer_size)) == NULL) {
		malloc_error("extract.c", "parse_show_file", "bss_result", buffer_size);
	}

  sprintf(bss_command, "sp t=%s", text_line);

  if ((return_code = i0(bss_command, bss_result)) != 0) {
    terminate_run("parse_show_file()", bss_command, return_code);
  }
    
  results_ptr = bss_result;
  sscanf(results_ptr, "%d %n", &no_terms, &chars_read);
  results_ptr += chars_read;

  for (t = 0; t < no_terms; t++) {
    /*
    **  'results_ptr' is now in the form:
    **
    **     t=<term> c=<gsl_class> s=<source> t=.....
    **
    **  First remove 't='
    */
    
    results_ptr += 2;
    
    /*
    **  Now determine length of 'parsed' (can't use sscanf since
    **  parsed may be a gsl term including a space.
    */

    length = strstr(results_ptr, " c=") - results_ptr;
    memcpy(parsed, results_ptr, length);
    parsed[length] = '\0';

    /*
    **  move past the space preceding 'c=' then read gsl_class
    */

    results_ptr += length + 1;
    sscanf(results_ptr, "c=%c s=%n", &gsl_class, &chars_read);

    if (strchr(LEGAL_GSL, gsl_class) == NULL) {
      /*
      **  If it's not a legal gsl class code then something has
      **  gone wrong. It is likely that the rest of the output
      **  from superparse is wrong so discard it all.
      **
      **  This will only happen very rarely so it is not necessary
      **  to incorporate any further validation.
      */
      break;
    }
    else {
      results_ptr += chars_read;
      
      /*
      **  Otherwise the length of 'source' is determined in a similar
      **  way to that of 'term' (i.e. can't use sscanf since 'source'
      **  may contain a space.
      */
      
      if (t == no_terms - 1) {
				/*
				**  If this is the last term there will be no more ' t=' in
				**  the string and 'source' will be the remaining string plus
				**  a '\n' character. Remove the '\n' from end of string.
				*/
				sprintf(source, "%s", results_ptr);
				length = strlen(source);
				source[length] = '\0';
				
				/*
				**  Check there are no more trailing spaces or newline
				**  characters. If so remove them.
				*/
				for (c = length - 1; c > 0; c--) {
					if ((source[c] != ' ') && (source[c] != '\n')) {
						break;
					}
					else {
						source[c] = '\0';
					}
				}
      }
      else {
				length = strstr(results_ptr, " t=") - results_ptr;
				memcpy(source, results_ptr, length);
				source[length] = '\0';
				/*
				**  Move past the space preceding the next t=
				*/
				results_ptr += length + 1;
      }
      
      /*
      **  DON'T save terms with gsl class codes of
      **
      **    H: stop term,
      **    F: other function word (stopped)
      **
      **    STOP_TERMS = "FH"
      **
      **  Don't generate a set yet as (i) term might exist more than
      **  once in the string, and (ii) the term might already be in
      **  the existing termset.
      */
      
      if (strchr(STOP_TERMS, gsl_class) == NULL) {
				if (strlen(parsed) > 0) {
					bss_set = NONE_ASSIGNED;
					np = NONE_ASSIGNED;
					attribute = gsl_class;
					little_r = 1;
					
					(reldoc_terms + term_count)->term_no = NONE_ASSIGNED;
					(reldoc_terms + term_count)->bss_set = bss_set;
					(reldoc_terms + term_count)->np = np;
					(reldoc_terms + term_count)->little_r = little_r;
					(reldoc_terms + term_count)->weight = weight;
					(reldoc_terms + term_count)->rsv = rsv;
					(reldoc_terms + term_count)->term_type = term_type;
					(reldoc_terms + term_count)->attribute = attribute;
					(reldoc_terms + term_count)->operation = operation;
					sprintf((reldoc_terms + term_count)->source, "%s", source);
					sprintf((reldoc_terms + term_count)->parsed, "%s", parsed);
					
					term_count++;
				}
      }
    }
  }

  /*
  **  Now sort reldoc_terms into ascending parsed order.
  */

  sort_terms(reldoc_terms, term_count, TSORT_PSD);

  /*
  **  Now eliminate duplicate parsed terms from reldoc_terms
  */
	
  sprintf(previous, "abcdefghijklmnopqrstuvwxyz");
  t = 0;
	
  for (i = 0; i < term_count; i++) {
    if (strcmp((reldoc_terms + i)->parsed, previous) != 0) {
      (extracted_terms + t)->term_no = t;
      (extracted_terms + t)->bss_set = (reldoc_terms + i)->bss_set;
      (extracted_terms + t)->np = (reldoc_terms + i)->bss_set;
      (extracted_terms + t)->little_r = (reldoc_terms + i)->bss_set;
      (extracted_terms + t)->weight = (reldoc_terms + i)->bss_set;
      (extracted_terms + t)->rsv = (reldoc_terms + i)->bss_set;
      (extracted_terms + t)->term_type = (reldoc_terms + i)->term_type;
      (extracted_terms + t)->attribute = (reldoc_terms + i)->attribute;

      sprintf((extracted_terms + t)->source, "%s",
							(reldoc_terms + i)->source);

      sprintf((extracted_terms + t)->parsed, "%s",
							(reldoc_terms + i)->parsed);

      (extracted_terms + t)->operation = (reldoc_terms + i)->operation;

			if (debugging) {
				fprintf(stderr, "PSF:: [%d:%d:%d:%d:%f:%f:%c:%c:%s:%s:%c]\n",
								(extracted_terms + t)->term_no,
								(extracted_terms + t)->bss_set,
								(extracted_terms + t)->np,
								(extracted_terms + t)->little_r,
								(extracted_terms + t)->weight,
								(extracted_terms + t)->rsv,
								(extracted_terms + t)->term_type,
								(extracted_terms + t)->attribute,
								(extracted_terms + t)->source,
								(extracted_terms + t)->parsed,
								(extracted_terms + t)->operation);
			}

      t++;
    }
    sprintf(previous, "%s", (reldoc_terms + i)->parsed);
  }
	
  /*
  **  Function returns 't', the number of distinct, extracted terms.
  */
	
  distinct_terms = t;

	free(reldoc_terms);
	free(bss_command);
	free(bss_result);

  return distinct_terms;
}


void write_new_file(struct query_terms *next_term, int index,
										int value, int term_flag, int *termset_size) {

	int debugging = atoi(getenv("WRITE_NEW_FILE_DEBUG"));

  int relj;
	
	if (*termset_size < MAX_NUMBER_TERMS - 1) {
		fprintf(tf, "%d:%d:%d:%d:%.3f:%.3f:%c:%c:%s:%s:%c",
						(next_term + index)->term_no,
						(next_term + index)->bss_set,
						(next_term + index)->np,
						(next_term + index)->little_r,
						(next_term + index)->weight,
						(next_term + index)->rsv,
						(next_term + index)->term_type,
						(next_term + index)->attribute,
						(next_term + index)->source,
						(next_term + index)->parsed,
						(next_term + index)->operation);
		
		if (debugging) {
			fprintf(stderr, "write_new_file():: %d:%d:%d:%d:%f:%f:%c:%c:%s:%s:%c",
							(next_term + index)->term_no,
							(next_term + index)->bss_set,
							(next_term + index)->np,
							(next_term + index)->little_r,
							(next_term + index)->weight,
							(next_term + index)->rsv,
							(next_term + index)->term_type,
							(next_term + index)->attribute,
							(next_term + index)->source,
							(next_term + index)->parsed,
							(next_term + index)->operation);
		}

		/*
		**  Now write term occurrence matrix to file
		*/
		
		if (BIGR == 1) {
			(next_term + index)->relj[0] = value;
			fprintf(tf, ":%d\n", (next_term + index)->relj[0]);

			if (debugging) {
				fprintf(stderr, ":%d]\n", (next_term + index)->relj[0]);
			}
		}
		else {
			for (relj = 0; relj < BIGR - 1; relj++) {
				if (term_flag == NEW_TERM) {
					(next_term + index)->relj[relj] = 0;
				}
				
				fprintf(tf, ":%d", (next_term + index)->relj[relj]);
			}
			
			(next_term + index)->relj[relj] = value;
			fprintf(tf, ":%d\n", (next_term + index)->relj[relj]);

			if (debugging) {
				fprintf(stderr, ":%d]\n", (next_term + index)->relj[relj]);
			}
		}
		
		fflush(tf);
		(*termset_size)++;
	}
	else {
		printf("\n\n*****************************************************\n");
		printf("**  THE MAXIMUM NUMBER OF TERMS HAS BEEN EXCEEDED  **\n");
		printf("**     THE TERMSET BUFFER IS SET TO 5000 TERMS     **\n");
		printf("**             THE RUN WILL BE HALTED              **\n");
		printf("*****************************************************\n\n");
		exit(0);
	}

	if ((MAX_NUMBER_TERMS - (*termset_size)) < 1000) {
		printf("WARNING: CURRENT TERMSET SIZE = %5d terms\n", *termset_size);
	}
}


