/*
**  fileterms.c
**
**  Write/read termset functions. Moved from utilities.c, Apr 97
*/

#define G_EXTERN extern

#include "defs.h"

G_EXTERN int calc_little_r(int bss_set);
G_EXTERN double calc_wgt(int np, int r, char term_type);
G_EXTERN double calc_rsv(double weight, int r, char term_type);
G_EXTERN void sort_terms(struct query_terms *, int, int);
G_EXTERN void sort_user_rels(struct user_rj *, int, int);
G_EXTERN int old_split_up(char *, struct query_terms *, int);
G_EXTERN void bss_find(char *, int *, int *);
G_EXTERN void terminate_run(char *, char *, int);
G_EXTERN void form_phrase(struct user_input *, int);
G_EXTERN void check_against_rels(int);
G_EXTERN void current_time(char *);
G_EXTERN void hf_header(int, char *, int);
G_EXTERN void malloc_error(char *, char *, char *, int);
G_EXTERN void malloc_success(char *, char *, char *, int);


FILE *OpenFile(char file_name[], char file_mode[]) {

	FILE *fptr;

	char message[8];

	sprintf(message, "create");

	if (file_mode[0] == 'a') {
		sprintf(message, "reopen");
	}

	if ((fptr = fopen(file_name, file_mode)) == (FILE *)0) {
		fprintf(stderr, "Can't %s <%s> file '%s'\n",	message, file_name);
		exit(-1);
	}

	return fptr;
}


void write_termset_file(struct query_terms *query, int no_terms) {
  int next_term;
  int relj;
	int term_count = 0;
	
  tf = fopen(termset_file, "w");
	
  /*
  **  First sort terms into ascending order of parsed.
  */

  sort_terms(query, no_terms, TSORT_PSD);

  for (next_term = 0; next_term < no_terms; next_term++) {
    if ((query + next_term)->np > 0) {
			(query + next_term)->term_no = term_count++;
      fprintf(tf, "%d:%d:%d:%d:%.3f:%.3f:%c:%c:%s:%s:%c",
							(query + next_term)->term_no,
							(query + next_term)->bss_set,
							(query + next_term)->np,
							(query + next_term)->little_r,
							(query + next_term)->weight,
							(query + next_term)->rsv,
							(query + next_term)->term_type,
							(query + next_term)->attribute,
							(query + next_term)->source,
							(query + next_term)->parsed,
							(query + next_term)->operation);
			
      /*
      **  Now write rels matrix info
      */
			
      if (BIGR > 0) {
				for (relj = 0; relj < BIGR; relj++) {
					fprintf(tf, ":%d", (query + next_term)->relj[relj]);
				}
      }
			
      fprintf(tf, "\n");
    }
  }
	
  fclose(tf);
}


void WriteUserTerms(struct query_terms *query, int no_terms) {
	/*
	**  Function to clear relevance feeback by just writing
	**  user entered terms to the termset file and then
	**  re-reading them back into termset[].
	**
	**  Relevance information is discarded so set BIGR to 0 (by
	**  calling function) which in effect starts from scratch.
	*/

	int debugging = atoi(getenv("WRITE_USER_TERMS_DEBUG"));

  int next_term;
	int term_count = 0;
	int relj;
	
  tf = fopen(termset_file, "w");
	
  /*
  **  First sort terms into ascending order of parsed.
  */

  sort_terms(query, no_terms, TSORT_PSD);

  for (next_term = 0; next_term < no_terms; next_term++) {
		if (toupper((query + next_term)->term_type) == 'U') {
			if ((query + next_term)->np > 0) {
				(query + next_term)->term_no = term_count++;
				(query + next_term)->little_r = 0;

				(query + next_term)->weight =
					calc_wgt((query + next_term)->np,
									 (query + next_term)->little_r ,
									 (query + next_term)->term_type);

				(query + next_term)->rsv =
					calc_rsv((query + next_term)->weight,
									 (query + next_term)->little_r ,
									 (query + next_term)->term_type);

				fprintf(tf, "%d:%d:%d:%d:%.3f:%.3f:%c:%c:%s:%s:%c\n",
								(query + next_term)->term_no,
								(query + next_term)->bss_set,
								(query + next_term)->np,
								(query + next_term)->little_r,
								(query + next_term)->weight,
								(query + next_term)->rsv,
								(query + next_term)->term_type,
								(query + next_term)->attribute,
								(query + next_term)->source,
								(query + next_term)->parsed,
								(query + next_term)->operation);
				
				if (debugging) {
					fprintf(stderr, "%d) [%s] (%c)\n",
									(query + next_term)->term_no,
									(query + next_term)->source,
									(query + next_term)->operation);
				}
				
				/*
				**  BIGR = 0 => no rels matrix information to write
				*/
			}
		}
	}
	
  fclose(tf);
}


int ReadTermsetFile(int reading_old_query) {
	/*
	**  reading_old_query = TRUE => we are reading in the terms from an
	**  archived query which means we need to generate a set for each
	**  term before we reweight them.
	**
	**  If (reading_old_query == FALSE)  =>  sets already exist
	*/

	int buffer_size;
	int valid = FALSE;
	
	int chars_read;
	int length;
	
	int no_terms;
	int np;
	int rj;
	int set;
  int term_no = 0;
	int next_term;
	
  char temp[MAX_TERM_LENGTH];

	int return_code;
	char bss_command[SMALL_BUFFER];
	char bss_result[SMALL_BUFFER];
	char *rbuf;
	
  char not_found_list[MAX_TERM_LENGTH];
	
  struct user_input *user_query;

	buffer_size = sizeof(struct user_input);

	if ((user_query = (struct user_input *) malloc(buffer_size)) == NULL) {
		malloc_error("fileutils.c", "ReadTermsetFile", "user_query", buffer_size);
	}			 
	
  if ((tf = fopen(termset_file, "r")) != NULL) {
		/*
		**  a 'termset_file' exists
		*/

		while (!feof(tf)) {
			if (fgets(temp, MAX_TERM_LENGTH, tf) == NULL) {
				break;
			}
			else {
				/*
				**  If we are loading an old query then there
				**  will be no sets. Once these are found we need
				**  to reweight the terms.
				*/
				
				temp[strlen(temp) - 1] = '\0';
				
				/*
				**  valid = split_up(temp, termset, term_no);
				*/
				
				valid = old_split_up(temp, termset, term_no);
				
				if (valid) {
					/*
					**  String stored in termset_file is ok.
					*/
					
					switch (reading_old_query) {
					  case TRUE :
						
							/*
							**  First need to find a set for each term or phrase
							*/
							
							(termset + term_no)->term_no = term_no;
						
							if (strchr(PHRASE_OPS, (termset + term_no)->operation) == NULL) {
								/*
								**  A single term
								*/
								
								bss_find((termset + term_no)->parsed, &set, &np);
								(termset + term_no)->bss_set = set;
								(termset + term_no)->np = np;
								
								if (np > 0) {
									(termset + term_no)->little_r = 
										calc_little_r((termset + term_no)->bss_set);
									
									(termset + term_no)->weight =
										calc_wgt((termset + term_no)->np,
														 (termset + term_no)->little_r,
														 (termset + term_no)->term_type);
									
									(termset + term_no)->rsv =
										calc_rsv((termset + term_no)->weight,
														 (termset + term_no)->little_r,
														 (termset + term_no)->term_type);
								}
								else {
									(termset + term_no)->little_r = NONE_ASSIGNED;
									(termset + term_no)->weight =
										(termset + term_no)->rsv =
										(double) NONE_ASSIGNED;
									
									/*
									**  Add term to 'not_found_list'
									**
									**  First add linefeed to 'not_found_list' if it is not empty
									*/
									
									if (strlen(not_found_list) > 0) {
										sprintf(not_found_list + strlen(not_found_list), "\n");
									}
									
									sprintf(not_found_list + strlen(not_found_list),
													"\"%s\" WAS NOT FOUND.",
													(termset + term_no)->source);
								}
							}
							else {
								/*
								**  Form a phrase
								*/
								sprintf(not_found_list, "\0");
								
								sprintf(user_query->source_terms,
												"%s", (termset + term_no)->source);
								
								sprintf(user_query->parsed_phrase,
												"%s", (termset + term_no)->parsed);
								
								sprintf(bss_command, "sp t=%s", user_query->source_terms);
								
								if ((return_code = i0(bss_command, bss_result)) != 0) {
									terminate_run("ReadTermsetFile()",
																bss_command, return_code);
								}
								
								rbuf = bss_result;
								
								sscanf(rbuf, "%d %n", &no_terms, &chars_read);
								rbuf += chars_read;
								
								user_query->no_terms = no_terms;
								
								for (next_term = 0; next_term < no_terms; next_term++) {
									rbuf += 2;
									length = strstr(rbuf, " c=") - rbuf;
									memcpy(user_query->parsed[next_term], rbuf, length);
									user_query->parsed[next_term][length] = '\0';
									
									rbuf += length + 1;
									sscanf(rbuf,
												 "c=%c s=%n",
												 &user_query->gsl_class[next_term], &chars_read);
									
									rbuf += chars_read;
									
									if (next_term < no_terms - 1) {
										length = strstr(rbuf, " t=") - rbuf;
										memcpy(user_query->source[next_term], rbuf, length);
										user_query->source[next_term][length] = '\0';
										rbuf += length + 1;
									}
									else {
										length = strlen(rbuf) - 1;
										sprintf(user_query->source[next_term], "%s", rbuf);
										user_query->source[next_term][length] = '\0';
									}
								}
								
								form_phrase(user_query, FALSE);
								
								(termset + term_no)->term_no = term_no;
								(termset + term_no)->bss_set = user_query->phrase_set;
								(termset + term_no)->np = user_query->phrase_np;
								(termset + term_no)->little_r = user_query->phrase_r;
								(termset + term_no)->weight = user_query->phrase_wgt;
								(termset + term_no)->rsv = user_query->phrase_rsv;
								(termset + term_no)->term_type = USER_TERM;
								(termset + term_no)->attribute = tolower(USER_PHRASE);
								
								if (user_query->phrase_np > 0) {
									/*
									**  Phrase exists - termset modified.
									*/
									NEW_ITERATION = TRUE;
									
									/*
									**  Now check to see if term is in any of the rels.
									**  Set (termset)->relj[] to:
									**
									**    NONE_ASSIGNED if the relj is no longer in the database
									**    0 if not in relj
									**    1 if in relj
									*/
									
									check_against_rels(term_no);
								}
								else {
									sprintf(not_found_list,
													"\"%s\" WAS NOT FOUND.",
													(termset + term_no)->source);
								}
							}

							break;
						
            case FALSE :
							/*
							**  Sets already exist.
							**  Only need to reweight etc. for phrases.
							*/
							if (strchr(PHRASE_OPS, (termset + term_no)->operation) != NULL) {
								/*
								**  Calculate r, wgt & rsv for phrases
								*/
								(termset + term_no)->little_r = 
									calc_little_r((termset + term_no)->bss_set);
								
								(termset + term_no)->weight =
									calc_wgt((termset + term_no)->np,
													 (termset + term_no)->little_r,
													 (termset + term_no)->term_type);
								
								(termset + term_no)->rsv =
									calc_rsv((termset + term_no)->weight,
													 (termset + term_no)->little_r,
													 (termset + term_no)->term_type);
							}
							
							break;
					}
					term_no++;
				}
				else {
					printf("ERROR READING termset_file\n");
				}
			}
		}
		
		fclose(tf);
	}
	else {
		printf("TERMSET FILE = [%s] ..... ", termset_file);
		printf("DOES NOT EXIST.\n");
		term_no = NONE_ASSIGNED;
	}	
			
  return term_no;
}


void write_terms_to_history_file(int no_terms, int command_no)
{
  int current_term;
  char elapsed[RUNTIME_LENGTH];
	
  /*
  **  Set NEW_ITERATION to FALSE to indicate that no new terms
  **  have been added to the termset[] either by the user
  **  or by extracting terms from user rels.
  */
  
  NEW_ITERATION = FALSE;
  
  /*
  **  Update ELAPSED_TIME
  */

  current_time(elapsed);
  
  for (current_term = 0; current_term < TERM_NO; current_term++) {
    hf_header(command_no, "query", PARAMS);
    
    fprintf(hf, ":%d:%d:%d:%d:%d:%.3f:%.3f:%c:%c:%s:%s:%c\n",
						ITERATION_NO,
						(termset + current_term)->term_no,
						(termset + current_term)->bss_set,
						(termset + current_term)->np,
						(termset + current_term)->little_r,
				    (termset + current_term)->weight,
						(termset + current_term)->rsv,
						(termset + current_term)->term_type,
						(termset + current_term)->attribute,
						(termset + current_term)->source,
						(termset + current_term)->parsed,
						(termset + current_term)->operation);
		
    fflush(hf);
  }
}


void write_rels_file(int no_rels)
{
  /*
  **  First sort (if necessary) into relno order
  */
  
  int next_rel;
	
  if (no_rels > 1) {
    sort_user_rels(user_rels, no_rels, RELNO_SORT);
  }
	
  if ((rf = fopen(rels_file, "w")) == NULL) {
		printf("Can't open [%s] for writing", rels_file);
		exit(0);
	}
	
  for (next_rel = 0; next_rel < no_rels; next_rel++) {
    fprintf(rf, "%d:%d:%d:%d:%d:%f:%d:%d:%c:%s:%s:%s\n",
						(user_rels + next_rel)->relno,
						(user_rels + next_rel)->set,
						(user_rels + next_rel)->abs_recno,
						(user_rels + next_rel)->set_recno,
						(user_rels + next_rel)->rj,
						(user_rels + next_rel)->weight,
						(user_rels + next_rel)->offset,
						(user_rels + next_rel)->length,
						(user_rels + next_rel)->sflag,
						(user_rels + next_rel)->docid,
						(user_rels + next_rel)->title,
						(user_rels + next_rel)->date);
  }
	
  fclose(rf);
}    


void RemoveTempfiles(int query_pid)
{
  /*
	**  Remove temporary files from Temp_dir
	*/

  sprintf(OS_COMMAND, "%s %s/*%d*", RM, Temp_dir, query_pid);

  if (system(OS_COMMAND) < 0) {
		perror(OS_COMMAND);
	}
}


FILE *test_open(char filename[], int rw, char *function)
{
	char file_mode[10];

  FILE *fptr;
	
	switch (rw) {
  	case WRITE :
			fptr = fopen(filename, "w");
			sprintf(file_mode, "WRITING");
			break;
  	case READ :
			fptr = fopen(filename, "r");
			sprintf(file_mode, "READING");
			break;
  	case APPEND :
			fptr = fopen(filename, "a");
			sprintf(file_mode, "APPENDING");
			break;
  	default :
			sprintf(file_mode, "UNKNOWN FILE MODE");
			break;
	}

  if (fptr == NULL) {
    printf("%s() :: Can't open file [%s] for %s.",
					 function, filename, file_mode);
    printf("\n\nLeaving program\n");
		exit(0);
	}
	else if (strcmp(file_mode, "UNKNOWN FILE MODE") == 0) {
		printf("%s()  --  %s.", file_mode);
    printf("\n\nLeaving program\n");
    exit(0);
	}
	else {
		printf("%s() :: File [%s] opened for %s.", filename, file_mode);
	}
}


int ReadRelsFile()
{
	/*
	**  Called when a user loads an archived query.
	**  Adds reljs to BIGR_SET and SEEN_SET if it
	**  exists in the current database.
	*/

	int char_length;
	int chars_read;

	int next_rel = 0;

	int set;
	int np;

	int relno;
	int bss_set;
	int abs_recno;
	int bss_set_recno;
	int rj;
	double weight;
	int offset;
	int length;
	char sflag;

	char temp[MAX_DOCID_LENGTH + MAX_PHRASE_LENGTH + MAX_DATE_LENGTH];

	char docid[MAX_DOCID_LENGTH];
  char title[MAX_PHRASE_LENGTH];
  char date[MAX_DATE_LENGTH];

	char *in;

	if ((rf = fopen(rels_file, "r")) != NULL) {
		while (!feof(rf)) {
			if (fgets(temp, MAX_TERM_LENGTH, rf) == NULL) {
				break;
			}
			else {
				temp[strlen(temp) - 1] = '\0';
				in = temp;

				/*
				**  Parse 'in'.
				*/
				
				sscanf(in, "%d:%d:%d:%d:%d:%f:%d:%d:%c:%n",
							 &relno, &bss_set, &abs_recno, &bss_set_recno,
							 &rj, &weight, &offset, &length, &sflag, &chars_read);
				
				in += chars_read;
				
				while (*in == ' ') {
					in++;
				}
				
				char_length = strchr(in, ':') - in;
				
				memcpy(docid, in, char_length);
				docid[char_length] = '\0';
				
				in += char_length + 1;
				
				while (*in == ' ') {
					in++;
				}
				
				char_length = strlen(in) - 10;
				memcpy(title, in, char_length);
				title[char_length] = '\0';
				in += char_length + 1;
				sprintf(date, "%s", in);
				
				(user_rels + next_rel)->relno = relno;
				(user_rels + next_rel)->set = bss_set;
				(user_rels + next_rel)->abs_recno = abs_recno;
				(user_rels + next_rel)->set_recno = bss_set_recno;
				(user_rels + next_rel)->rj = rj;
				(user_rels + next_rel)->weight = weight;
				(user_rels + next_rel)->offset = offset;
				(user_rels + next_rel)->length = length;
				(user_rels + next_rel)->sflag = sflag;
				sprintf((user_rels + next_rel)->docid, "%s", docid);
				sprintf((user_rels + next_rel)->title, "%s", title);
				sprintf((user_rels + next_rel)->date, "%s", date);
			}
			next_rel++;
		}
		
		fclose(rf);
	}

	return next_rel;
}



