/*
**  header.c
**
**  Functions for parsing the show header.
**  Moved from utilities.c 28-Apr-97.
*/

#define G_EXTERN extern

#include "defs.h"

#undef RESULTS_BUFFER_SIZE
#define RESULTS_BUFFER_SIZE 75 * 1024

G_EXTERN void terminate_run(char *, char *, int);

void determine_doc_length(int set, int norec,
													int *header_length, int *doc_length,
													int *passage_length) {

	int debug = atoi(getenv("DETERMINE_DOC_LENGTH_DEBUG"));

	int dir_rec;
	int no_dir_recs;
	int dir_rec_len;
	int no_passage_recs;

	int pass_len0;
	int pass_len1;

	int return_code;

	char bss_command[SMALL_BUFFER];
	char bss_result[RESULTS_BUFFER_SIZE];

	char *rbuf;

	sprintf(bss_command,
					"show s=%d f=%d r=%d",
					set, atoi(getenv("HEADER_SHOW_FORMAT")), norec);
			
	if ((return_code = i0(bss_command, bss_result)) != 0) {
		terminate_run("determine_doc_length()", bss_command, return_code);
	}
	else {
		rbuf = bss_result;

		if (debug) {
			fprintf(stderr, "determine_doc_length() :: rbuf = [%s]\n", rbuf);
		}
	}

	/*
	**  Now pick out header information.
	**
	**  The first 144 bytes (= 92+3+19+19+3+8) take us to the
	**  length of the text field value. There will always be
	**  3 directory records but I have read the number from
	**  the header just in case (the last directory record
	**  will always be the text field).
	*/
	
	/*
	**  Read header length then move to the number
	**  of directory records field.
	*/
	
	sscanf(rbuf, "%8d", header_length);

	rbuf += 92;

	if (debug) {
		fprintf(stderr, "determine_doc_length() :: rbuf = [%s]\n", rbuf);
	}

	/*
	**  Read the number of directory records.
	*/
	
	sscanf(rbuf, "%3d", &no_dir_recs);
	rbuf += 3;

	if (debug) {
		fprintf(stderr, "determine_doc_length() :: rbuf = [%s]\n", rbuf);
	}

	/*
	**  Now determine the total length of the document.
	*/
	
	*doc_length = 0;

	for (dir_rec = 0; dir_rec < no_dir_recs; dir_rec++) {
		/*
		**  Move past field_no & offset fields the read field length
		*/
		rbuf += 11;
		sscanf(rbuf, "%8d", &dir_rec_len);
		*doc_length += dir_rec_len;
		rbuf += 8;
	}

	if (debug) {
		fprintf(stderr, "determine_doc_length() :: rbuf = [%s]\n", rbuf);
	}

	/*
	**  Move on to the passage records
	*/
	
	sscanf(rbuf, "%6d", &no_passage_recs);

	if (debug) {
		fprintf(stderr, 
						"determine_doc_length() :: no_passage_recs = [%d]\n", 
						no_passage_recs);
	}

	rbuf += 6;
			
	if (no_passage_recs == 2) {
		/*
		**  Read lengths of both passages and determine the shortest
		*/
		
		rbuf += 18;
		sscanf(rbuf, "%8d", &pass_len0);

		rbuf += 26;
		sscanf(rbuf, "%8d", &pass_len1);

		*passage_length = (pass_len0 >= pass_len1) ? pass_len1 : pass_len0;
	}
	else {
		*passage_length = *doc_length;
	}

	if (debug) {
		fprintf(stderr, "determine_doc_length() :: %d (%d)\n",
						*doc_length, *passage_length);
	}
}


void parse_header(char *bss_result, char *returned_doc,
									struct header *show_head,
									struct dir_records *doc_fields,
									struct passage_records *passage,
									struct highlight_records *hl_terms,
									int *next_term) {
	/*
	**  Extracts header information from a record
	**  shown with format 3 or 259
	*/

	int debug = atoi(getenv("PARSE_HEADER_DEBUG"));

	int max_doclength;
	int doc_length = 0;

  int chars_read;
  int length;
  int num_hl = 0;
  int offset;
  int v;
  int which_part;
	int part_only;

  char code;
	char weight_string[32];

  char *all;

	int PASSAGE;
	int FULLDOC;

	/*
	**  Parameters defined in .okapi_rc
	*/

	int dir_rec_len = atoi(getenv("DIRECTORY_REC_LEN"));
	int pas_rec_len = atoi(getenv("PASSAGE_REC_LEN"));
	int hil_rec_len = atoi(getenv("HIGHLIGHT_REC_LEN"));
  
	/*
	**  DOC_THRESHOLD set in ~/.okapi_rc for easy modification.
	*/

	max_doclength = atoi(getenv("DOC_THRESHOLD"));

  all = bss_result;

	if (debug) {
		fprintf(stderr, "parse_header() :: all = [%s]\n", all);
	}
	
  sscanf(all, "%8d%32c%6d%10d%10d%n",
				 &show_head->msg_len, &show_head->reserved,
				 &show_head->setnum, &show_head->set_recno,
				 &show_head->abs_recno, &chars_read);

	if (debug) {
		fprintf(stderr, "parse_header() :: show_head->\n\n");
		fprintf(stderr, "msg_len: %d, setnum: %d, set_recno: %d, abs_recno: %d\n",
						show_head->msg_len, show_head->setnum,
						show_head->set_recno, show_head->abs_recno);
	}

	all += chars_read;

	sscanf(all, "%10s%n", weight_string, &chars_read);

	all += chars_read;

	if (strchr(weight_string, '.') == NULL) {
		show_head->weight = (double) NONE_ASSIGNED;
	}
	else {
		show_head->weight = atof(weight_string);
	}

	sscanf(all, "%10c%6d%3d%n",
				 &show_head->unused, &show_head->data_offset,
				 &show_head->no_dir_recs, &chars_read);
	
	if (debug) {
		fprintf(stderr, "weight: %.3f, data_offset: %d\n",
						show_head->weight, show_head->data_offset);
	}

  /*
  **  directory records are the fields e.g. TITLE, etc.
  */
  
  all += chars_read;
	
  /*
  **  Pick out field information
  */

	if (debug) {
		fprintf(stderr, 
						"parse_header() :: no_dir_recs: %d\n\n", show_head->no_dir_recs);
	}
	
  for (v = 0; v < show_head->no_dir_recs; v++) {
    sscanf(all, "%3d%8d%8d%n", 
					 &(doc_fields + v)->field_no,
					 &(doc_fields + v)->offset,
					 &(doc_fields + v)->length,
					 &chars_read);
		 
		if (chars_read != dir_rec_len) {
			fprintf(stderr,
							"parse_header() :: chars_read = %d, dir_rec_len = %d\n",
							chars_read, dir_rec_len);
		}

		doc_length += (doc_fields + v)->length;

		if (debug) {
			fprintf(stderr, 
							"field_no: %d,  offset: %d,  length: %d\n",
							(doc_fields + v)->field_no,
							(doc_fields + v)->offset,
							(doc_fields + v)->length);
		}

		all += dir_rec_len;
  }
	
  /*
  **  Read passage information
  **
  **  Now read the number of passage records. This will be 0 or 2.
  */
	
  sscanf(all, "%6d%n", &show_head->no_passage_recs, &chars_read);
  all += chars_read;
	
	if (debug) {
		fprintf(stderr, 
						"parse_header() :: no_passage_recs: %d\n\n",
						show_head->no_passage_recs);
	}
	
  if (show_head->no_passage_recs == 2) {
    /*
    **  There are 2 passage records.
    */
    
    for (v = 0; v < show_head->no_passage_recs; v++) {
      /*
      **  The highest weighted passage will come first (whether
			**  the sub-passage or the full document.
      */
			
			sscanf(all, "%10lf%8d%8d%n", 
						 &(passage + v)->weight,
						 &(passage + v)->offset,
						 &(passage + v)->length,
						 &chars_read);

			if (chars_read != pas_rec_len) {
				fprintf(stderr,
								"parse_header() :: chars_read = %d, pas_rec_len = %d\n",
								chars_read, pas_rec_len);
			}

			all += pas_rec_len;

			if (debug) {
				fprintf(stderr, 
								"weight: %.3f,  offset: %d,  length: %d\n",
								(passage + v)->weight,
								(passage + v)->offset,
								(passage + v)->length);
			}

			/*
			**  Sometimes a very short sub-passage is found that has a weight
			**  very close to zero. If this is the case then we should set
			**  the passage offset and fields to zero.
			*/

			if ((passage + v)->weight <= 0.0) {
				(passage + v)->weight = 0.0;
				(passage + v)->offset = 0;
				(passage + v)->length = 0;

			}

			if (debug) {
				fprintf(stderr, 
								"weight: %.3f,  offset: %d,  length: %d\n",
								(passage + v)->weight,
								(passage + v)->offset,
								(passage + v)->length);
			}

    }
	}
	else {
		/*
		**  Set up passage_records structure.
		**  Values might be needed elsewhere.
		*/
		for (v = 0; v < 2; v++) {
			(passage + v)->weight = show_head->weight;
			(passage + v)->offset = show_head->data_offset;
			(passage + v)->length = doc_length;
		}
	}
  
	PASSAGE =
		((passage + 0)->length > (passage + 1)->length) ? 1 : 0;
	
	FULLDOC = (PASSAGE == 0) ? 1 : 0;

	part_only = ((passage + FULLDOC)->length >= max_doclength &&
							 (passage + PASSAGE)->length < max_doclength);

  /*
  **  Read highlight information
  */
  
  sscanf(all, "%6d%n", &show_head->no_highlight_recs, &chars_read);
  all += chars_read;
  
  num_hl = 0;

	if (debug) {
		fprintf(stderr, 
						"parse_header() :: no_highlight_recs: %d\n\n",
						show_head->no_highlight_recs);
	}

  if (show_head->no_highlight_recs > 0) {
		for (v = 0; v < show_head->no_highlight_recs; v++) {
			sscanf(all, "%c%8d%8d%n", &code, &offset, &length, &chars_read);
			
			if (chars_read != hil_rec_len) {
				fprintf(stderr,
								"parse_header() :: chars_read = %d, hil_rec_len = %d\n",
								chars_read, hil_rec_len);
			}

			all += hil_rec_len;
			
      if (part_only) {
				/*
				**  Showing passage only
				*/

				if ((offset >= (passage + PASSAGE)->offset) &&
						(offset <= ((passage + PASSAGE)->offset +
												(passage + PASSAGE)->length))) {
					/*
					**  If we're only showing part of the document set next_term
					**  to the first highlight term in the passage.
					*/
					if (*next_term == 0) {
						(*next_term) = num_hl;
					}
				}
      }
      
      (hl_terms + num_hl)->code = code;
			(hl_terms + num_hl)->offset = offset;
			(hl_terms + num_hl)->length = length;

			if (debug) {
				fprintf(stderr, 
								"code: %c,  offset: %d,  length: %d\n",
								(hl_terms + num_hl)->code,
								(hl_terms + num_hl)->offset,
								(hl_terms + num_hl)->length);
			}			

			num_hl++;
    }
  }
  
  show_head->no_highlight_recs = num_hl;
  
	sprintf(returned_doc, "%s", all);
}
