/*
**  hitlist.c
**
**  Displays the next 50 docs starting at offset ???
**
**  A global variable, 'short_offset' declared in 'defs.h' will be used
**  to keep track of the last starting record in the set.
**
**  May 8th - Changed hitlist from a listbox to a text widget.
*/

#define G_EXTERN extern

#include "defs.h"

#undef MAX_DISPLAY_LENGTH
#define MAX_DISPLAY_LENGTH 63

#undef TITLE_BUFFER_SIZE
#define TITLE_BUFFER_SIZE 512

typedef struct terms_in_doc {
  int count;
  char source[REASONABLE_TERM_LENGTH];
  char stem[REASONABLE_TERM_LENGTH];
};

int check_user_rels(char *, double);
void construct_doclength_field(char *, int, int, int, int);
void construct_title(char *);

G_EXTERN int MapWeight(double, double);

G_EXTERN void current_time(char *);

G_EXTERN void show_time(char *, int);
G_EXTERN void determine_doc_length(int, int, int *, int *, int *);
G_EXTERN void terminate_run(char *, char *, int);
G_EXTERN void parse_header(char *, char *,
													 struct header *, struct dir_records *,
													 struct passage_records *,
													 struct highlight_records *,
													 int *);

G_EXTERN void hf_header(int, char *, int);
G_EXTERN void sort_user_rels(struct user_rj *, int, int);
G_EXTERN void find_docset(char *, int *, int *);
G_EXTERN void malloc_error(char *, char *, char *, int);
G_EXTERN void malloc_success(char *, char *, char *, int);

void build_hitlist(int show_set, int set_np, double max_weight,
										 int window, char hitrels_file[], int *last_row,
										 struct hitlist_info row_index[]) {
  /*
  **  window determines which display the list is being formed for:
  **
  **  1) HITS_WINDOW  -  hitlist
  **  2) RELS_WINDOW  -  rels pool window
  */

	int debug = atoi(getenv("BUILD_HITLIST_DEBUG"));
	int part_debug = atoi(getenv("TERM_OCCURRENCE_DEBUG"));

	int return_code;

	int buffer_size;

	int max_doclength;
	int title_len;

  int already_seen;
  int bp_length = 0;
	
  int c;
  int chars_read;

  int doc_length;
  double doc_weight;

  int end_doc;
  int exists = 0;

  int first_hlt = 0;

  int found_first = FALSE;

	int header_length;

  int hi_on = FALSE;
  int hitline = 0;
  int hit_index = 0;
  int hit_term_count = 0;
  int hit_length = 0;

  int i;
  int k;
  int length;

	int last_record = atoi(getenv("MAX_RECS_TO_SHOW"));

  int n = 0;
  int next_char;
  int next_hl;
  int next_term;
  int new_docs = 0;
  int norec;

	int no_blanks;
	int no_title_chars = 200;

	int nparsed;

	int num_length = 1;

  int part_only = FALSE;
  int passage_length;
  int total_chars = 0;
  int total_length;

  int set_recno;
  int seq_no = NONE_ASSIGNED;
  int start_doc = 0;
  int term_count = 0;
  int term_length = 0;
	int title_count = 0;

	int FULLDOC;
	int PASSAGE;

	int page_size = atoi(getenv("CHARS_PER_PAGE"));

  char *all_doc;
	char *returned_doc;
  char *in;
  char *temp_ptr;
  char *title_string;
	char *tsp;

  char title_line[512];

	char start_tag;
	char end_tag;
	char rjcode;

  char *docid;
	char doclength_field[16];
  char *full_term;

  char hitlist_line[MAX_PHRASE_LENGTH];
  char word[MAX_PHRASE_LENGTH];
  char stem[MAX_PHRASE_LENGTH];
  char blanks[MAX_DISPLAY_LENGTH + 1];
  char elapsed[RUNTIME_LENGTH];

	char bss_command[SMALL_BUFFER];
	char *bss_result;

  struct terms_in_doc term_info[MAX_NUMBER_TERMS];

  /*
  **  Show header data
  */

  struct header *show_head;
  struct dir_records *doc_fields;
  struct passage_records *passage;
  struct highlight_records *hl_terms;

  FILE *hlf;

	int HITS_WINDOW = 0;
	int RELS_WINDOW = 1;

  show_time("build_hitlist", START);

	if (debug) {
		fprintf(stderr,
						"build_hitlist() :: set = %d, np = %d, max_weight = %.3f\n",
						show_set, set_np, max_weight);
		fprintf(stderr,
						"build_hitlist() :: window = %d, last_row = %d\n",
						window, *last_row);
		fprintf(stderr,
						"build_hitlist() :: hitrels_file = [%s]\n",	hitrels_file);
		
	}

  memset(blanks, ' ', MAX_DISPLAY_LENGTH);
  blanks[MAX_DISPLAY_LENGTH] = '\0';

	/*
	**  DOC_THRESHOLD set in ~/.okapi_rc for easy modification.
	*/

	max_doclength = atoi(getenv("DOC_THRESHOLD"));

  /*
  **  HITLIST: Always show docs in the range [0 .. MAX_RECS_TO_SHOW - 1]
  **           Set end_doc to np in set if this is < MAX_RECS_TO_SHOW
  **     RELS: Show all docs from user_rels judged YES or MAYBE.
  */
	
  hlf = fopen(hitrels_file, "w");
	
  if (window == HITS_WINDOW) {
    /*
    **  hitlist window
		**
		**  Allocate memory for hitlist_info
    */

		/*
		**  Allocate storage for header data
		*/
		
		buffer_size = sizeof(struct header);
		
		if ((show_head = (struct header *) malloc(buffer_size)) == NULL) {
			malloc_error("hitlist.c", "build_hitlist",
									 "show_head", buffer_size);
		}
		else if (debug) {
			malloc_success("hitlist.c", "build_hitlist",
										 "show_head", buffer_size);
		}

		buffer_size = (params_file.nf + 1) * sizeof(struct dir_records);
		
		if ((doc_fields =
				 (struct dir_records *) malloc(buffer_size)) == NULL) {
			malloc_error("hitlist.c", "build_hitlist",
									 "doc_fields", buffer_size);
		}
		else if (debug) {
			malloc_success("hitlist.c", "build_hitlist",
										 "doc_fields", buffer_size);
		}

		buffer_size = MAX_PASSAGES * sizeof(struct passage_records);
		
		if ((passage =
				 (struct passage_records *) malloc(buffer_size)) == NULL) {
			malloc_error("hitlist.c", "build_hitlist",
									 "passage", buffer_size);
		}
		else if (debug) {
			malloc_success("hitlist.c", "build_hitlist",
										 "passage", buffer_size);
		}

		buffer_size = MAX_HIGHLIGHT_TERMS * sizeof(struct highlight_records);
		
		if ((hl_terms =
				 (struct highlight_records *) malloc(buffer_size)) == NULL) {
			malloc_error("hitlist.c", "build_hitlist",
									 "hl_terms", buffer_size);
		}
		else if (debug) {
			malloc_success("hitlist.c", "build_hitlist",
										 "hl_terms", buffer_size);
		}

    FIRST_REC = 0;
    LAST_REC = last_record;
		
    /*
    **  First set up start and end record numbers for docs to show
    */
		
    start_doc = 0;
    end_doc = (set_np < last_record) ? set_np : last_record;
    
		if (debug) {
			fprintf(stderr, "build_hitlist() :: end_doc = [%d]\n", end_doc);
		}

    norec = 0;
    new_docs = 0;
		
    while (new_docs < end_doc) {
      /*
      **  First set up FIRST_REC for choosing recs
      **  from the hitlist window.
      */
			
      if (norec == set_np) {
				break;
      }
			
      if (!found_first) {
				FIRST_REC = norec;
				found_first = TRUE;
      }
			
      /*
      **  Set up show commands depending upon whether hitlist or rels pool
      **
      **  1. First determine whether document (or passage) is too long.
			**
			**  Need to allocate size of results buffer dynamically since a
			**  document that is too long to show in entirety may contain a
			**  passage that is < DOC_THRESHOLD in length.
      */

			if (debug) {
				fprintf(stderr, "show_set = %d, norec = %d\n", show_set, norec);
			}

			determine_doc_length(show_set, norec,
													 &header_length, &doc_length,
													 &passage_length);

			/*
			**  NOTE: passage_length is set to doc_length
			**  if there are no passages
			*/

			if (debug) {
				fprintf(stderr,
								"build_hitlist() :: show_set = %d, norec = %d\n",
								show_set, norec);
				fprintf(stderr,
								"build_hitlist() :: header_len = %d\n", header_length);
				fprintf(stderr, "build_hitlist() :: doc_len = %d, pass_len = %d\n",
								doc_length, passage_length);
			}

			/*
			**  For databases where there is no passage retrieval
			**  passage_length will be set to doc_length.
			**
			**  If passage is too long to show => doc is too long also
			*/
			
      if (passage_length < max_doclength) {
				/*
				**  Show whole document. First allocate
				**  bss_result buffer size.
				*/
				
				if (debug) {
					fprintf(stderr, "build_hitlist() :: RECORD = %d\n", norec);
				}

				buffer_size = header_length + doc_length + 1;
				
				if ((bss_result = (char *) malloc(buffer_size)) == NULL) {
					malloc_error("hitlist.c", "build_hitlist",
											 "bss_result", buffer_size);
				}				 
				else if (debug) {
					malloc_success("hitlist.c", "build_hitlist",
												 "bss_result", buffer_size);
				}

				sprintf(bss_command,
								"show s=%d f=%d r=%d",
								show_set, atoi(getenv("PASSAGE_SHOW_FORMAT")), norec);
				
				/*
				**  Send the command to the bss
				*/
				
				if ((return_code = i0(bss_command, bss_result)) != 0) {
					terminate_run("build_hitlist()", bss_command, return_code);
				}
				
				/*
				**  First read header information.
				**
				**  'all_doc' will point to whole record less header information.
				*/
				
				total_chars = 0;
				
				buffer_size = strlen(bss_result) + 1;

				if ((returned_doc = (char *) malloc(buffer_size)) == NULL) {
					malloc_error("hitlist.c", "build_hitlist",
											 "returned_doc", buffer_size);
				}
				else if (debug) {
					malloc_success("hitlist.c", "build_hitlist",
												 "returned_doc", buffer_size);
				}

				parse_header(bss_result, returned_doc, show_head,
										 doc_fields, passage, hl_terms, &first_hlt);

				all_doc = returned_doc;

				if (debug) {
					fprintf(stderr, "[%s]\n", all_doc);
				}

				/*
				**  'all_doc' contains the whole document less the header.
				**
				**  Determine number of characters to copy to 'title_string'
				**  field(0) + MAX_TITLE_CHARS from the rest.
				**
				**  Pick out the docid and then move onto the start
				**  of the text field (field 3).
				**
				**  First check to see if the document has already been seen
				**  by the searcher. If so don't include it in the current
				**  hitlist but, if it exists in user_rels[] as a positive
				**  rel, set the weight in user_rels to that in the new set.
				**
				**  1. Copy enough characters from all_doc to title_string
				*/
				
				no_title_chars =
					atoi(getenv("MAX_TITLE_CHARS")) + (doc_fields + 0)->length;

				if (debug) {
					fprintf(stderr,
									"build_hitlist() :: no_title_chars = [%d]\n",
									no_title_chars);
				}

				buffer_size = no_title_chars + 1;

				if ((title_string = (char *) malloc(buffer_size)) == NULL) {
					malloc_error("hitlist.c", "build_hitlist",
											 "title_string", buffer_size);
				}
				else if (debug) {
					malloc_success("hitlist.c", "build_hitlist",
												 "title_string", buffer_size);
				}

				memcpy(title_string, all_doc, no_title_chars);
				*(title_string + no_title_chars) = '\0';
					
				/*
				**  First set tsp to point to title_string and then 
				**  manipulate this rather than title_string otherwise
				**  we wont be able to free title_string.
				*/

				tsp = title_string;

				if (debug) {
					fprintf(stderr, "build_hitlist() :: title_string = [%s]\n", tsp);
				}

				/*
				**  2. Pick out docid
				**
				**  Note: Inspec databases don't have line-feeds between fields.
				**        Trec ones do so need to get rid of this.
				*/

				length = (doc_fields + 0) -> length;

				if (debug) {
					fprintf(stderr, "build_hitlist() :: length = [%d]\n", length);
				}

				buffer_size = length + 1;

				if ((docid = (char *) malloc(buffer_size)) == NULL) {
					malloc_error("hitlist.c", "build_hitlist",
											 "docid", buffer_size);
				}
				else if (debug) {
					malloc_success("hitlist.c", "build_hitlist",
												 "docid", buffer_size);
				}

				memcpy(docid, tsp, length);

				/*
				**  "ai" type databases don't contain newline chars.
				**  Other types contain newline chars at the end of each
				**  displayed line. Hence the docid string needs the
				**  newline character to be removed.
				*/

				if (strncmp(params_file.db_type, "ai", 2) == 0) {
					*(docid + length) = '\0';
				}
				else {
					*(docid + length - 1) = '\0';
				}

				if (debug) {
					fprintf(stderr, "build_hitlist() :: docid = [%s]\n", docid);
				}

				/*
				**  Move to the start of field 2.
				*/
				
				tsp += length;
	
				if (debug) {
					fprintf(stderr, "build_hitlist() :: tsp = [%s]\n", tsp);
				}

				/*
				**  Set part_only to TRUE for non-ai databases if
				**  full doc too long but passage ok
				*/

				if (show_head->no_passage_recs == 2) {
					PASSAGE =
						((passage + 0)->length > (passage + 1)->length) ? 1 : 0;
					
					FULLDOC = (PASSAGE == 0) ? 1 : 0;
					
					part_only = ((passage + FULLDOC)->length >= max_doclength &&
											 (passage + PASSAGE)->length < max_doclength);
				}
				else {
					part_only = FALSE;
					PASSAGE = 0;
					FULLDOC = 1;
				}

				/*
				**  Check whether document has been judged yet.
				*/

				already_seen = (USER_RELS_COUNT > 0) ?
					check_user_rels(docid, show_head->weight) :
					FALSE;

				/*
				**  title string contains the characters to go in the
				**  title but it may contain extra white space and LFs.
				*/

				if (!already_seen) {
					/*
					**  First make up the hitlist title string.
					**  Extract rec_no, docid, weight & doc length
					**
					**  Copy enough of all_doc to title_string
					**
					**  Increment hitlist entry count for storing first
					**  and last rows in (row_index -> seq_no)
					*/


					seq_no++;
					new_docs++;
					set_recno = show_head->set_recno;
					doc_weight = show_head->weight;
					
					/*
					**  Update ELAPSED_TIME
					*/
					
					current_time(elapsed);

					/*
					**  Write history file entry
					*/
					
					hf_header(COMMAND_NO, "hl_title", PARAMS);

					fprintf(hf, ":%d:%d:%d:%s:%10.3f",
									ITERATION_NO, set_recno,
									show_head->abs_recno,	docid, doc_weight);
					
					if (show_head->no_passage_recs == 2) {
						fprintf(hf, ":%d:%d:%d:%d\n",
										(passage + PASSAGE)->offset, (passage + PASSAGE)->length,
										(passage + FULLDOC)->offset, (passage + FULLDOC)->length);
					}
					else {
						fprintf(hf, ":%d:%d:%d:%d\n",
										NONE_ASSIGNED, NONE_ASSIGNED,
										(doc_fields + 0)->offset, doc_length);
					}
					fflush(hf);
					
					/*
					**  Now make up header line to be written to hitrels_file:
					**
					**  <R> <D> <W>  ...  [<P>/]<F> page[s]
					**
					**  where: R = set_recno
					**         D = docid
					**         W = normalised weight
					**         P = display_paslen
					**         F = display_doclen
					**
					**  The [<P>/]<F> page[s] will first be formed in
					**  doclength_field (e.g. 12/15 pages). This will be
					**  a maximum of, say, 16 characters in length.
					*/

					if (show_head->no_passage_recs == 2) {
						construct_doclength_field(doclength_field,
																			show_head->no_passage_recs,
																			(passage + PASSAGE)->length,
																			(passage + FULLDOC)->length,
																			page_size);
					}
					else {
						construct_doclength_field(doclength_field,
																			show_head->no_passage_recs,
																			doc_length,
																			doc_length,
																			page_size);
					}


					/*
					**  The function MapWeight() maps the weight range
					**  [minwt..maxwt] onto the range [MIN_WGT..MAX_WGT]
					**  Don't need to show docid, they will all be FT records
					*/

					sprintf(hitlist_line,	"<S>%2d:  %s  [%d]",
									set_recno + 1, docid, MapWeight(doc_weight, max_weight));

					/*
					**  The "-5" in the next line is to account for
					**  the "<S>" and the ": "
					*/
					
					no_blanks = MAX_DISPLAY_LENGTH + 2 -
						strlen(hitlist_line) -	strlen(doclength_field);
					
					strncat(hitlist_line, blanks, no_blanks);
					sprintf(hitlist_line + strlen(hitlist_line),
									"%s<E>", doclength_field);
					
					fprintf(hlf, "%s\n", hitlist_line);
					
					(row_index + seq_no) -> set_recno = set_recno;
					(row_index + seq_no) -> abs_recno = show_head->abs_recno;
					sprintf((row_index + seq_no) -> docid, "%s", docid);
					(row_index + seq_no) -> weight = doc_weight;
					(row_index + seq_no) -> first_row = hit_index;
					(row_index + seq_no) -> offset[0] = (passage + PASSAGE)->offset;
					(row_index + seq_no) -> offset[1] = (passage + FULLDOC)->offset;
					(row_index + seq_no) -> length[0] = (passage + PASSAGE)->length;
					(row_index + seq_no) -> length[1] = (passage + FULLDOC)->length;
					
					hit_index += 1;					
					hitline = 0;
					hit_term_count = 0;
					
					/*
					**  Now build up title line
					**
					**  Make it into one line for sending to 'construct_title'
					**  Remove blanks from the front of 'title_string'
					*/
					
					temp_ptr = NULL;
					
					while ((temp_ptr = strchr(tsp, '\n')) != NULL) {
						*temp_ptr  = ' ';
					}
					
					/*
					**  Remove prefix spaces
					*/
					
					while (*tsp == ' ') {
						tsp++;
					}
					
					/*
					**  Copy title details to hit_titles[]
					*/
					
					hit_titles[norec].abs_recno = show_head->abs_recno;
					hit_titles[norec].set_recno = norec;
					sprintf(hit_titles[norec].docid, "%s", docid);
					
					/*
					**  Make sure tsp doesn't point to title_line
					*/

					sprintf(title_line, "%s", tsp);
					construct_title(title_line);
					sprintf(hit_titles[norec].title, "%s", title_line);

					if (part_only) {
						total_chars += (passage + PASSAGE)->offset;
						all_doc += total_chars;
						bp_length = (passage + PASSAGE)->length;
						memcpy(bss_result, all_doc, bp_length);
						*(bss_result + bp_length) = '\0';
					}
					else {
						all_doc += ((doc_fields + PASSAGE)->length +
												(doc_fields + 1)->length);
						total_chars += ((doc_fields + PASSAGE)->length +
														(doc_fields + 1)->length);;
						sprintf(bss_result, "%s", all_doc);
						bp_length = strlen(bss_result);
					}
				
					next_hl = 0;
					
					if (part_only) {
						/*
						**  find index of first highlight record in the passage
						*/
						for (next_hl = 0;
								 next_hl < show_head->no_highlight_recs;
								 next_hl++) {
							if ((hl_terms + next_hl)->offset > (passage + PASSAGE)->offset) {
								break;
							}
						}
					}
					
					term_count = 0;
					
					if (debug) {
						fprintf(stderr, "build_hitlist() :: extract_query_term_info\n");
						fprintf(stderr, "                   [%s]\n", bss_result);
					}

					for (k = 0; k < bp_length; k++) {
						if (next_hl < show_head->no_highlight_recs) {
							if ((hl_terms + next_hl)->length > 0) {
								if (total_chars == (hl_terms + next_hl)->offset) {
									hi_on = TRUE;
									next_char = 0;
								}
							}
						}
						
						if (hi_on) {
							if (*(bss_result + k) == '\n') {
								word[next_char++] = ' ';
							}
							else {
								word[next_char++] = *(bss_result + k);
							}
						}
						
						total_chars++;
						
						/*
						**  Check for end of word
						*/
						
						if (((hl_terms + next_hl)->offset +
								 (hl_terms + next_hl)->length) ==	total_chars) {
							word[next_char] = '\0';
							hi_on = FALSE;
							next_hl++;
							next_char = 0;
							
							/*
							**  Check to see if term already found. If so increment
							**  'count' field of 'term_info[]'. If not add term to
							**  'term' field of 'term_info[]' and set 'count' field to 1
							*/

							if (strlen(word) > 0) {
								/*
								**  First stem word
								*/

								sprintf(bss_command, "parse t=%s", word);

								if ((return_code = i0(bss_command, bss_result)) != 0) {
									terminate_run("build_hitlist()",
																bss_command, return_code);
								}

								in = bss_result;
								sscanf(in, "%d t=%s", &nparsed, stem);

								if (term_count == 0) {
									/*
									**  The first term
									*/
									
									sprintf(term_info[term_count].source, "%s", word);
									sprintf(term_info[term_count].stem, "%s", stem);
									term_info[term_count].count = 1;
									term_count++;
								}
								else {
									exists = NONE_ASSIGNED;
									
									for (i = 0; i < term_count; i++) {
										if (strcasecmp(term_info[i].stem, stem) == 0) {
											exists = i;
										}
									}
									
									if (exists == NONE_ASSIGNED) {
										sprintf(term_info[term_count].source, "%s", word);
										sprintf(term_info[term_count].stem, "%s", stem);
										term_info[term_count].count = 1;
										term_count++;
									}
									else {
										term_info[exists].count += 1;
									}
								}

								if (part_debug) {
									fprintf(stderr, "generate_hitlist():: t=[%s], s=[%s]\n",
													stem, word);
								}
							}
						}
					}
					
					/*
					**  Now build remaining hitlist_line[] array entries
					**  to write to hitrels_file.
					*/
					
					in = title_line;
					sprintf(hitlist_line, "");
					
					while (strlen(in) > 0) {
						sscanf(in, "%s%n", word, &chars_read);
						in += chars_read;
						
						if (strlen(hitlist_line) == 0) {
							/*
							**  The first word in the title_line
							*/
							sprintf(hitlist_line, "%s", word);
						}
						else if ((strlen(hitlist_line) + strlen(word) + 1) <
										 MAX_DISPLAY_LENGTH) {
							sprintf(hitlist_line + strlen(hitlist_line), " %s", word);
						}
						else {
							/*
							**  Too long to fit on this line so copy to hitrels_file
							**  and start building ip next line.
							*/
							fprintf(hlf, "%s\n", hitlist_line);
							hit_index++;
							
							/*
							**  Write next line of hitlist info to history file.
							**  First update ELAPSED_TIME
							*/

							current_time(elapsed);

							hf_header(COMMAND_NO, "hl_info", PARAMS);
							fprintf(hf,
											":%d:%d:%d:%s\n",
											ITERATION_NO, set_recno, hitline++, hitlist_line);
							
							sprintf(hitlist_line, "%s", word);
						}
					}
					
					/*
					**  Write the last non-zero length hitlist_line to the history file.
					*/
					
					if (strlen(hitlist_line) > 0) {
						fprintf(hlf, "%s\n", hitlist_line);
						hit_index++;
						
						/*
						**  Update ELAPSED_TIME
						*/

						current_time(elapsed);

						hf_header(COMMAND_NO, "hl_info", PARAMS);
						fprintf(hf, ":%d:%d:%d:%s\n",
										ITERATION_NO, set_recno, hitline++, hitlist_line);
					}
					
					sprintf(hitlist_line, "");
					in = NULL;
					
					/*
					**  Only add term occurrence info to hitlist window items
					**
					**  New record; set line length to zero.
					*/
					
					hit_length = 0;
					
					for (next_term = 0; next_term < term_count; next_term++) {
						/*
						**  Update ELAPSED_TIME
						*/

						current_time(elapsed);

						hf_header(COMMAND_NO, "hl_terms", PARAMS);

						fprintf(hf,
										":%d:%d:%d:%s:%d\n",
										ITERATION_NO, set_recno, hit_term_count++,
										term_info[next_term].source, term_info[next_term].count);
						
						/*
						**  Allocate storage for full_term
						**  First determine length of number for term_count
						*/

						num_length = (int) log10((float) term_info[next_term].count) + 1;

						buffer_size =
							strlen(term_info[next_term].source) + num_length + 4;

						if ((full_term =	(char *) malloc(buffer_size)) == NULL) {
							malloc_error("hitlist.c", "build_hitlist",
													 "full_term", buffer_size);
						}
						else if (debug) {
							malloc_success("hitlist.c", "build_hitlist",
														 "full_term", buffer_size);
						}

						sprintf(full_term,
										"%s (%d)",
										term_info[next_term].source,
										term_info[next_term].count);

						term_length = strlen(full_term);
						
						if (next_term == 0) {
							/*
							**  The first term
							*/
							sprintf(hitlist_line, "<T>%s", full_term);
							hit_length = term_length;
						}
						else if ((hit_length + term_length) > MAX_DISPLAY_LENGTH) {
							/*
							**  Too long for current hitlist_line,
							**  => print 'hitlist_line' to hitrels_file.
							*/
							
							fprintf(hlf, "%s<W>\n", hitlist_line);
							hit_index++;
							
							/*
							**  Add term occurrence info to new (next) hitlist_line
							*/
							
							sprintf(hitlist_line, "<T>%s", full_term);
							hit_length = term_length;
						}
						else {
							/*
							**  Add terms to non-zero length hitlist_line
							*/
							sprintf(hitlist_line + strlen(hitlist_line),
											" %s", full_term);
							
							hit_length += (term_length + 1) ;
						}
						free(full_term);
					}
					
					/*
					**  Last hitlist_line string to be written to hitrels_file
					*/
					
					if (hit_length > 0) {
						fprintf(hlf, "%s<W>\n", hitlist_line);
						hit_index++;
					}
					
					/*
					**  Add a blank line after record
					*/
					
					fprintf(hlf, "\n");
					(row_index + seq_no) -> last_row = hit_index++;
					hitlist_line[0] = '0';
					/*
					**  End of if (!already_seen)
					*/
				}
				else if (debug) {
					printf("build_hitlist: Already seen [%s]\n", docid);
				}

				if (debug) {
					fprintf(stderr,
									"build_hitlist() :: freeing memory for: returned_doc\n");
					fprintf(stderr,
									"                                         title_string\n");
					fprintf(stderr,
									"                                         bss_result\n");
					fprintf(stderr,
									"                                         docid\n");
				}

				free(returned_doc);
				free(title_string);
				free(bss_result);
				free(docid);
      }
      else {
				/*
				**  Even passage too long to show
				*/
				printf("build_hitlist: Record is too long to show\n");
				hf_header(COMMAND_NO, "hl_error", PARAMS);
				fprintf(hf, ":%d:%d:passage & document too long to show\n",
								ITERATION_NO, show_head->abs_recno);
      }

			norec++;
    }


		if (debug) {
			fprintf(stderr, "build_hitlist() :: freeing memory for: show_head\n");
			fprintf(stderr, "                                         doc_fields\n");
			fprintf(stderr, "                                         passage\n");
			fprintf(stderr, "                                         hl_terms\n");
		}

		free(show_head);
		free(doc_fields);
		free(passage);
		free(hl_terms);

    LAST_REC = norec;
		*last_row = hit_index;
  }
  else if (USER_RELS_COUNT > 0) {
    /*
		**  Relspool window
		**
		**  Build hitlist window from user_rels[]
		**
		**  Sort user_rels into descending weight order if more than one.
		*/

    if (USER_RELS_COUNT > 1) {
      sort_user_rels(user_rels, USER_RELS_COUNT, WGT_SORT);
    }
    
    /*
		**  Display all user rels
		*/
    end_doc = USER_RELS_COUNT;
    
    for (norec = 0; norec < USER_RELS_COUNT; norec++) {
      if ((user_rels + norec)->rj != R_NOT) {
				/*
				**  Only show in rels window if (user_rels)->rj
				**  in [R_FULL, R_PASSAGE]
				**
				**  First write history file entry
				*/

				if (debug) {
					fprintf(stderr, "hitlist.c :: build_hitlist():\n\n");
					fprintf(stderr, "(user_rels + %d)->abs_recno = [%d]\n",
									norec, (user_rels + norec)->abs_recno);
					
					fprintf(stderr, "(user_rels + %d)->set_recno = [%d]\n",
									norec, (user_rels + norec)->set_recno);
					fprintf(stderr, "    (user_rels + %d)->docid = [%s]\n",
									norec, (user_rels + norec)->docid);
					fprintf(stderr, "   (user_rels + %d)->weight = [%.3f]\n",
									norec, (user_rels + norec)->weight);
					fprintf(stderr, "   (user_rels + %d)->sflag = [%c]\n",
									norec, (user_rels + norec)->sflag);
					fprintf(stderr, "                  norec = [%d]\n", norec);
					fprintf(stderr, "                last_row = [%d]\n", norec);
				}					
				
				(row_index + norec) -> abs_recno = (user_rels + norec)->abs_recno;
				(row_index + norec) -> set_recno = (user_rels + norec)->set_recno;
				sprintf((row_index + norec) -> docid,
								"%s", (user_rels + norec)->docid);
				(row_index + norec) -> weight = (user_rels + norec)->weight;
        (row_index + norec) -> first_row = norec;
				(row_index + norec) -> last_row = norec;
				
				/*
				**  For rels made:
				**
				**  a) before rf cleared,       (start_tag, end_tag) = ('P', 'Q')
				**  b) during current search,   (start_tag, end_tag) = ('F', 'G')
				**
				**  For (a) all lines of the entry are tagged.
				**  For (b) only the first, header, line is tagged.
				*/
				
				switch ((user_rels + norec)->sflag) {
				  case 'N' :
						start_tag = 'P';
						end_tag = 'Q';
						break;
				  default :
						start_tag = 'F';
						end_tag = 'G';
						break;
				}

				switch ((user_rels + norec)->rj) {
					case R_PASSAGE :
						rjcode = 'P';
						break;
  				case R_FULL :
						rjcode = 'F';
						break;
				}

				fprintf(hlf, "<%c>[%c] ", start_tag, rjcode);

				switch ((user_rels + norec)->sflag) {
				  case 'N' :
						/*
						**  Don't show weight for rels made before rf cleared
						*/
						fprintf(hlf, "    ");
						break;
				  default :
						/*
						**  Show weight for rels made during current search
						*/
						fprintf(hlf, "%4d",
										MapWeight((user_rels + norec)->weight, max_weight));
						break;
				}

				fprintf(hlf, "     %s<%c>\n", (user_rels + norec)->docid,	end_tag);
				
				in = (user_rels + norec)->title;
				hitlist_line[0] = '\0';
				
				while (strlen(in) > 0) {
					sscanf(in, "%s%n", word, &chars_read);
					in += chars_read;
					
					if (strlen(hitlist_line) == 0) {
						/*
						**  The first word in the title_line
						*/

						switch ((user_rels + norec)->sflag) {
						  case 'N' :
								/*
								**  Tag all of entry for rel made before rf cleared
								*/
								sprintf(hitlist_line, "<%c>%s", start_tag, word);
								break;
						  default :
								sprintf(hitlist_line, "%s", word);
								break;
						}
					}
					else if ((strlen(hitlist_line) + strlen(word) + 1) <
									 MAX_DISPLAY_LENGTH) {
						sprintf(hitlist_line + strlen(hitlist_line),
										" %s", word);
					}
					else {
						/*
						**  Too long to fit on this line so write to hitlist_file
						**  and then set new hitlist-line to word.
						*/

						switch ((user_rels + norec)->sflag) {
						  case 'N' : 
								/*
								**  Tag all of entry for rel made before rf cleared
								*/
								fprintf(hlf, "%s<%c>\n", hitlist_line, end_tag);
								sprintf(hitlist_line, "<%c>%s", start_tag, word);
								break;
						  default : 
								fprintf(hlf, "%s\n", hitlist_line);
								sprintf(hitlist_line, "%s", word);
								break;
						}
					}
				}
				
				/*
				**  Write the last non-zero length hitlist_line to hitlist_file
				*/
				
				if (strlen(hitlist_line) > 0) {
					switch ((user_rels + norec)->sflag) {
					  case 'N' :
							/*
							**  Tag all of entry for rel made before rf cleared
							*/
							fprintf(hlf, "%s<%c>\n", hitlist_line, end_tag);
							break;

  				  default :
							fprintf(hlf, "%s\n", hitlist_line);
							break;
					}
				}
			}
		}
	}
	else {
    fprintf(stderr, "Something's gone wrong!!");
		fclose(hlf);
		fclose(hf);
		exit(0);
  }
	
  fclose(hlf);

  COMMAND_NO++;
  show_time("build_hitlist", FINISH);
}


int check_user_rels(char docid[], double weight) {
	int debug = atoi(getenv("CHECK_USER_RELS_DEBUG"));

  int np;
  int doc_set;
  int c;
  int found = FALSE;
	
  char *rbuf;

	int return_code;
	char bss_command[SMALL_BUFFER];
	char bss_result[SMALL_BUFFER];
	
	if (debug) {
		fprintf(stderr, "Entering:: check_user_rels\n");
	}

  /*
	**  Find set for document
	*/
	
	find_docset(docid, &doc_set, &np);
	
  sprintf(bss_command, "find s=%d s=%d op=and save=n", doc_set, SEEN_SET);
	
  if ((return_code = i0(bss_command, bss_result)) != 0) {
    terminate_run("check_user_rels()", bss_command, return_code);
  }
	
  rbuf = bss_result;
  sscanf(rbuf, "np=%d", &np);
	
  /*
	**  return FALSE for (np == 0), TRUE otherwise
	*/
	
  if (found = (np > 0)) {
    /*
		**  Doc already exists. Set weight of already recorded
		**  doc to the weight of the current one
		*/
		
    for (c = 0; c < USER_RELS_COUNT; c++) {
      if (strcasecmp(docid, (user_rels + c)->docid) == 0) {
				(user_rels + c)->weight = weight;
				break;
      }
    }
  }

	if (debug) {
		fprintf(stderr, " Leaving:: check_user_rels\n");
	}
	
  return found;
}


void construct_doclength_field(char doclength_field[],
															 int no_passage_recs, int passage_chars,
															 int fulldoc_chars, int page_size) {

	int debug = atoi(getenv("CONSTRUCT_DOCLENGTH_FIELD_DEBUG"));

	int display_doclen;
	int display_paslen;

	if (debug) {
		fprintf(stderr, "Entering:: construct_doclength_field\n");
	}

	sprintf(doclength_field, "");

	if (fulldoc_chars <= page_size) {
		display_doclen = 1;
	}
	else {
		display_doclen =
			(int) floor(((float) fulldoc_chars / (float) page_size) + 0.5);
	}
	
	if (no_passage_recs == 2) {
		if (passage_chars < fulldoc_chars) {
			if (passage_chars <= page_size) {
				display_paslen = 1;
			}
			else {
				display_paslen =
					(int) floor(((float) passage_chars /
											 (float) page_size) + 0.5);
			}
		}
		else {
			display_paslen = display_doclen;
		}

		sprintf(doclength_field + strlen(doclength_field),
						"%d/", display_paslen);
	}
	
	sprintf(doclength_field + strlen(doclength_field),
					"%d", display_doclen);
	
	sprintf(doclength_field + strlen(doclength_field), " page");
	
	if (display_doclen > 1) {
		sprintf(doclength_field + strlen(doclength_field), "s");
	}

	if (debug) {
		fprintf(stderr, " Leaving:: construct_doclength_field\n");
	}
}				


void construct_title(char text_line[]) {
  /*
  **  Make up the equivalent of a title field.
  **
  **  Remove all field titles (enclosed between angled brackets).
  */

	int debug = atoi(getenv("CONSTRUCT_TITLE_DEBUG"));

	int buffer_size;

  int old_char;
  int length;
  int new_char;

  char *in;

  char prev_ch = '\0';

  char *dummy;
  char title_line[512];
  
	if (debug) {
		fprintf(stderr, "Entering:: construct_title\n");
	}

  /*
	**  Now remove possible headers from the string.
	**  These are enclosed in angled brackets, e.g. <TEXT>
	*/
	
  while (strchr(text_line, '<') != NULL) {
    in = text_line;
    length = strchr(in, '<') - in;

		buffer_size = length + 1;

		if ((dummy = (char *) malloc(buffer_size)) == NULL) {
			malloc_error("hitlist.c", "construct_title",
									 "dummy", buffer_size);
		}
		else if (debug) {
			malloc_success("hitlist.c", "construct_title",
										 "dummy", buffer_size);
		}

    memcpy(dummy, in, length);
    dummy[length] = '\0';

    in += length + 1;
    if (strchr(in, '>') != NULL) {
      length = strchr(in, '>') - in;
      in += length + 1;
    }
    else {
      *in = '\0';
    }

    if (strlen(dummy) > 0) {
      sprintf(text_line, "%s %s", dummy, in);
    }
    else {
      sprintf(text_line, "%s", in);
    }

		free(dummy);
  }

  new_char = 0;
	
  for (old_char = 0; old_char < strlen(text_line); old_char++) {
    if (text_line[old_char] == ' ') {
      if (prev_ch != ' ') {
				title_line[new_char++] = text_line[old_char];
      }
    }
    else {
      title_line[new_char++] = text_line[old_char];
    }
    prev_ch = text_line[old_char];
  }
	
  title_line[new_char] = '\0';
	
  if (strlen(title_line) >= TITLE_CHARS - 7) {
    memcpy(text_line, title_line, TITLE_CHARS - 7);
    text_line[TITLE_CHARS - 7] = '\0';
  }
  else {
    sprintf(text_line, "%s", title_line);
  }
	
  sprintf(text_line + strlen(text_line), ".....");

	if (debug) {
		fprintf(stderr, " Leaving:: construct_title\n");
	}
}

