##  TREC-Okapi Interactive interface
##
##  TCL procedures Project
##
##  Created July 1995 for Trec IV
##
##  Alterations:
##
##    Mar, Apr, May 1996  --  Trec 5
##    Mar, Apr 1997       --  Trec 6
##

proc db_search { } {
	global hitlist_title
	global profile_name
	
	set message_text "\nSearching .....\n"
	set window_name ".waiting"
	
	set back "NavyBlue"
	set fore "Yellow"
	
	processing_message $window_name $message_text $back $fore
	
	set result [ busy2 c.db_search \"\" \"\" ]

	.hitlist_text configure -state normal
	load_file .hitlist_text $result
	
	##		highlight_hit_terms .hitlist_text Black LightCyan
	highlight_hit_terms .hitlist_text Black LightCyan
	highlight_hit_title .hitlist_text Black SpringGreen
	
	##  make hitlist read-only
	##
	
	.hitlist_text configure -state disabled
	.hitlist_label configure -text $hitlist_title
	
	window_title
	
	set result [ c.show_rels ]
	
	if { $result != "EMPTY"} {
		load_file .positive_rels_text $result
		
		highlight_current_rels .positive_rels_text Red Grey90
		highlight_previous_rels .positive_rels_text Grey50 Grey90
		
		##  make rels_pool read-only
		##
		.positive_rels_text config -state disabled
	}
	
	##  disable search button, ^S and pull-down menu choice
	
	.main_buttons.search configure -state disabled
	
	bind .term_entry_box <Control-s> "nothing"
	
	## .pulldown_menu_bar.search.menu disable 1
	
	##  Destroy pop-up processing_message window
	##
	destroy $window_name
	
	cancel_menu .options_menu
	window_title
}

proc show_rec { hit_rec} {
	global prev_rel
	upvar #0 $hit_rec recnum
	
	## first disable double click on hitlist entry
	## so user can't show record twice
	
	bind .hitlist_text <Double-Button-1> {
		## puts "DISABLED HITLIST SELECTION"
		nothing
	} 
	
	## show elapsed time in title bar
	window_title
	
	set dot [ string first . $hit_rec ]
	incr dot -1
	
	set row [ string range $hit_rec 0 $dot ] 
	set result [ c.show_rec $row ]

	##  Return the record header to here for display
	##  in the title bar of the full record display
	
	set hash [ string first "##" $result ]
	
	if { $hash != "-1" } {
		display_error .error_window { Error Message } $result
	} else {
		if { $result != "NONE_FOUND" } {
			##  prev_rel:docid:filename:best_pass:
			##  recvd_date:pathname:next_record:doc_part
			
			set colon [ string first ":" $result ]
			incr colon -1
			
			set prev_rel [ string range $result 0 $colon ]
			
			incr colon 2
			set result [ string range $result $colon end ]
			set colon [ string first ":" $result ]
			incr colon -1
			
			set docid [ string range $result 0 $colon ]
			
			incr colon 2
			set result [ string range $result $colon end ]
			set colon [ string first ":" $result ]
			incr colon -1
			
			set filename [ string range $result 0 $colon ]
			
			incr colon 2
			set result [ string range $result $colon end ]
			set colon [ string first ":" $result ]
			incr colon -1
			
			set best_pass [ string range $result 0 $colon ]
			
			incr colon 2
			set result [ string range $result $colon end ]
			set colon [ string first ":" $result ]
			incr colon -1
			
			set next_record [ string range $result 0 $colon ]
			
			incr colon 2
			
			set result [ string range $result $colon end ]

			set colon [ string first ":" $result ]
			incr colon -1

			set doc_part [ string range $result 0 $colon ]

			incr colon 2
			set passage_button [ string range $result $colon end ]
			
			show_document .fullrec-window \
					$docid $best_pass $filename $doc_part $passage_button
		} else {
			set msg "You did not Double-Click\n\nwithin the hitlist entry"
			display_error .error_window { Error message } $msg
		}
	}
	
	selection clear .
	
	##  Move next record to top of display
	
	set top_row [ expr $next_record ]
	.hitlist_text yview $top_row
	
	##  Renable double-click on hitlist entry

	bind .hitlist_text <Double-Button-1> {
		## puts "RE-ENABLED HITLIST SELECTION"
		set cursor_pos "[.hitlist_text index insert]"
		busy2 show_rec $cursor_pos \"\"
		selection clear
	}
	
	focus .term_entry_box
	window_title
}

##  Pop-up text window to display documents in
##

proc show_document { win docid best_pass filename doc_part passage_button } {
	##
	##  passage_button is P or N
	##
	##  P means need passage button

	global tk_priv
	global indbut
	global tfile
	global prev_rel
	
	set tfile $filename
	
	set relY Y
	set relN N
	
	catch {
		destroy $win
	}
	
	## 1. Create the top-level window and divide it into top and bottom parts.
	
	set title "$docid"
	
	toplevel $win -class Frame
	wm title $win $title
	wm iconname $win Dialog
	
	## gets rid of X window borders
	#
	## wm overrideredirect $win true
	
	frame $win.document_frame \
			-relief raised \
			-bd 1 \
			-bg White
	
	text $win.document_text \
			-height 40 \
			-width 80 \
			-relief raised \
			-bd 2 \
			-yscrollcommand "$win.tpscroll set" \
			-bg white \
			-fg black \
			-wrap word \
			-font -*-fixed-bold-r-normal--15-140-*-*-*-*-*-*
	
	scrollbar $win.tpscroll \
			-command "$win.document_text yview" \
			-bg LightGrey \
			-troughcolor SlateGrey
	
	##  Create frame to be packed at the bottom of the text window
	##  in which to place the relevance judgement buttons.
	
	frame $win.bot \
			-relief raised \
			-bd 1 \
			-bg Grey95
	
	##  Create Full document / Passage only / Not relevant buttons
	##  to be placed at the bottom of text frame.
	##
	##      all_rel   Black, SpringGreen
	##  passage_rel   Black, Orange
	##      not_rel   White, Red

	if { $passage_button == "P" } {
		set all_rel_text "Full Document Relevant"
	} else {
		set all_rel_text "  Relevant "
	}

	button $win.all_rel \
			-text $all_rel_text \
			-bg LightGrey \
			-fg Black \
			-activebackground LightGoldenRodYellow \
			-relief groove \
			-bd 2

	if { $passage_button == "P" } {
		button $win.passage_rel \
				-text {Passage Only Relevant} \
				-bg LightGrey \
				-fg Black \
				-activebackground LightGoldenRodYellow \
				-relief groove \
				-bd 2
	}
	
	button $win.not_rel \
			-text {Not Relevant} \
			-bg LightGrey \
			-fg Black \
			-activebackground LightGoldenRodYellow \
			-relief groove \
			-bd 2
	
	##  Now pack the widgets.
	##
	##  1. Text area and scroll bars
	
	pack $win.document_frame \
			-side top \
			-fill both
	
	pack $win.tpscroll \
			-in $win.document_frame \
			-side right \
			-fill y
	
	pack $win.document_text \
			-in $win.document_frame \
			-side left
	
	##  Bottom frame for the buttons
	
	pack $win.bot \
			-side bottom \
			-fill both
	
	##  The buttons
	
	pack $win.all_rel \
			-in $win.bot \
			-side left \
			-expand 1 \
			-padx 3m \
			-pady 3m \
			-ipadx 2m \
			-ipady 1m
	
	if { $passage_button == "P" } {
		pack $win.passage_rel \
				-in $win.bot \
				-side left \
				-expand 1 \
				-padx 3m \
				-pady 3m \
				-ipadx 2m \
				-ipady 1m
	}

	pack $win.not_rel \
			-in $win.bot \
			-side left \
			-expand 1 \
			-padx 3m \
			-pady 3m \
			-ipadx 2m \
			-ipady 1m
	
	##  End of the widget creation and packing procedure
	##
	##  2. Load file into top of window
	
	$win.document_text delete 1.0 end
	set f [ open $filename ]
	while { ![ eof $f ]} {
		$win.document_text insert end [ read $f 1000 ]
	}
	close $f
	
	##  3. highlight passage and terms

	##		highlight_passage $win.document_text Black LightCyan
	highlight_passage $win.document_text Black LightCyan
	highlight_terms $win.document_text Black SpringGreen
	
	## Make text buffer read only
	## $win.document_text config -state disabled
	
	##  4. move to the first highlighted term in the text
	##     set the global variable 'top_row' to the row the
	##     row the first highlighted term is in.
	
	if { $best_pass == "P"} {
		move_to_pass_start $win.document_text
	} else {
		move_to_first_hl $win.document_text
	}
	
	##  5. Highlight button for previous rel
	##     if prev_rel in [ 1,2,3 ] then a judgement has already been made
	
	if { $prev_rel == "1" } {
		$win.all_rel config -bg White -fg Red
	} elseif { $prev_rel == "2"} {
		if { $passage_button == "P" } {
			$win.passage_rel config -bg White -fg Red
		}
	} elseif { $prev_rel == "3"} {
		$win.not_rel config -bg White -fg Red
	}
	
	##  doc_part is IN [ 0, 1 ]
	##
	##  0: passage & fulldoc the same length
	##  1: only showing best passage
	##  2: fulldoc has the highest weight
	##  3: passage has the highest weight
	
	if { $doc_part == 0 } {
		if { $passage_button == "P" } {
			$win.passage_rel config -bg LightGrey -fg Gray95 -state disabled
		}
	} elseif { $doc_part == 1 } {
		$win.all_rel config -bg LightGrey -fg Gray95 -state disabled
	}
	
	##  6. Disable the close button on the popup text window
	##
	
	disable_close_button $win
	
	##  7. Set cursor key movement bindings. Must focus to the
	##     scrollbar directly for explicit key bindings like
	##     up, down, prior and next to have effect.
	
	focus $win.tpscroll
	$win.document_text configure -state disabled
	
	##  8. Bind relj buttons
	##
	##     If a previous rel on this document is the same as the
	##     current one do nothing, but need to call make_rel in
	##     order to write judgement to history file. In these cases
	##     pass n, f, or p to make_rel procedure instead of N, F or P
	##
	##    was, e.g. busy2 make_rel N \"\"		
	##

	if { $prev_rel != "3" } {
		bind $win.not_rel <Button-1> {
			busy2 make_rel N $prev_rel
			set tk_priv(button) not_rel
		}
	} else {
		bind $win.not_rel <Button-1> {
			busy2 make_rel n $prev_rel
			set tk_priv(button) not_rel
		}
	}
	
	if { $doc_part != 1 } {
		if { $prev_rel != "1" } {
			bind $win.all_rel <Button-1> {
				busy2 make_rel F $prev_rel
				set tk_priv(button) all_rel
			}
		} else {
			bind $win.all_rel <Button-1> {
				busy2 make_rel f $prev_rel
				set tk_priv(button) all_rel
			}
		}
	}
	
	if { $doc_part != 0 } {
		if { $prev_rel != "2" } {
			if { $passage_button == "P" } {
				bind $win.passage_rel <Button-1> {
					busy2 make_rel P  $prev_rel
					set tk_priv(button) passage_rel
				}
			}
		} else {
			if { $passage_button == "P" } {
				bind $win.passage_rel <Button-1> {
					busy2 make_rel p  $prev_rel
					set tk_priv(button) passage_rel
				}
			}
		}
	}
	
	##  9. Wait for the user to respond, then restore the focus
	##     and return the index of the selected button.
	
	tkwait variable tk_priv(button)
	
	set indbut $tk_priv(button)
	
	destroy $win
	focus .term_entry_box
}


##  Load file into a text widget

proc load_file { w file } {
	$w configure -state normal
	
	$w delete 1.0 end
	set f [ open $file ]
	
	while { ![ eof $f ]} {
		$w insert end [ read $f 1000 ]
	}
	
	close $f
}

proc reinstate_term { rm_rec} {
	##  ir_term returns the cursor position when the searcher
	##  doubled-clicked the lefthand mouse button in the form:
	##
	##    row.column
	##
	##  where row is numbered from 1 (not 0) and column is
	##  numbered from 0.
	
	upvar #0 $rm_rec recnum
	
	window_title
	
	set dot [ string first . $rm_rec ]
	incr dot -1
	set row [ string range $rm_rec 0 $dot ] 
	set filenames [ c.reinstate_term $row ]
	
	##  filenames will be returned from 'gui.cc:remove_term()'
	##
	##  It will contain a string of the form:
	##
	##    <query_fname>:<removed_fname>
	##
	##  where   <query_fname> = file of query terms
	##    and <removed_fname> = file of removed terms
	##
	##  or an error message '## TERM REMOVAL ERROR ##\n'
	
	set hash [ string first "##" $filenames ]
	
	if { $hash != "-1" } {
		display_error .error_window { Error Message } $filenames
	} else {
		set colon [ string first ":" $filenames ]
		incr colon -1
		set query_fname [ string range $filenames 0 $colon ]
		incr colon 2
		set removed_fname [ string range $filenames $colon end ]
		
		if { $query_fname != "EMPTY" } {
			load_file .current_query_text $query_fname
			highlight_hit_terms .current_query_text SeaGreen white
			highlight_hit_title .current_query_text red white
			
			##  turn search button on
			##
			.main_buttons.search configure -state normal
			bind .term_entry_box <Control-s> "db_search"
		} else {
			.current_query_text configure -state normal
			.current_query_text delete 1.0 end
		}
		
		if { $removed_fname != "EMPTY" } {
			load_file .removed_terms_text $removed_fname
		} else {
			.removed_terms_text configure -state normal
			.removed_terms_text delete 1.0 end
		}
	}
	
	.removed_terms_text mark set insert end
	selection clear .
	focus .term_entry_box
	
	##		.current_query_text configure -state disabled
	##		.removed_terms_text config -state disabled
	
	window_title
}


##  procedure to remove a term from the displayed termlist
##  by double-clicking select button on it.
##
##  the procedure takes TREC(IR_TERM) as a parameter.
##  If this is empty nothing is done
##

proc remove_term { rm_rec} {
	##  ir_term returns the cursor position when the searcher
	##  doubled-clicked the lefthand mouse button in the form:
	##
	##    row.column
	##
	##  where row is numbered from 1 (not 0) and column is
	##  numbered from 0.
	
	upvar #0 $rm_rec recnum

	window_title
	
	set dot [ string first . $rm_rec ]
	incr dot -1
	
	set row [ string range $rm_rec 0 $dot ] 

	##  puts "removed row = $row"

	set filenames [ c.remove_term $row ]
	selection clear .
	
	##  filenames will be returned from 'gui.cc:remove_term()'
	##
	##  It will contain a string of the form:
	##
	##    <query_fname>:<removed_fname>
	##
	##  where   <query_fname> = file of query terms
	##    and <removed_fname> = file of removed terms
	##
	##  or an error message '## TERM REMOVAL ERROR ##\n'
	
	set hash [ string first "##" $filenames ]
	
	if { $hash != "-1" } {
		display_error .error_window { Error Message } $filenames
	} else {
		set colon [ string first ":" $filenames ]
		incr colon -1
		set query_fname [ string range $filenames 0 $colon ]
		incr colon 2
		set removed_fname [ string range $filenames $colon end ]
		
		if { $query_fname != "EMPTY" } {
			selection clear .current_query_text
			load_file .current_query_text $query_fname
			highlight_hit_terms .current_query_text SeaGreen white
			highlight_hit_title .current_query_text red white
			
			##  turn search button on
			##
			.main_buttons.search configure -state normal
			bind .term_entry_box <Control-s> "db_search"
		} else {
			.current_query_text configure -state normal
			.current_query_text delete 1.0 end
		}
		
		##				.current_query_text configure -state disabled
		
		##  Need to pass the NO_REMOVED_TERMS value to the function
		##  so that we can move to the last row in the window
		
		if { $removed_fname != "EMPTY" } {
			selection clear .removed_terms_text
			load_file .removed_terms_text $removed_fname
			##  set bottom_row end
			##  .removed_terms_text yview $bottom_row
		} else {
			.removed_terms_text configure -state normal
			.removed_terms_text delete 1.0 end
		}
	}
	
	.current_query_text mark set insert end
	selection clear .
	focus .term_entry_box
	
	##		.current_query_text config -state disabled
	##		.removed_terms_text configure -state disabled
	
	selection clear .current_query_text
	selection clear .current_query_text
	
	window_title
}


proc clear_hl { w } {
	selection clear $w
}

##  Procedure to display elapsed time in
##  the main window title bar

proc window_title { } {
	set result [ c.window_title ]
	set full_title "$result"
	wm title . $full_title
}


proc disable_search { } {
	##  This procedure is called everytime a key is pressed in the
	##  ".term_entry_box". Its purpose is yo disable the the search
	##  button & Ctrl-key equivalent just in case a user tries to do
	##  a search after starting entry of a term but before s/he
	##  completes entry by pressing the <Return> key.
	##
	##  It is bound to ".term_entry_box <KeyPress>" in the
	##  procedure "global_bindings".

	global query

	.main_buttons.search configure -state disabled
	bind .term_entry_box <Control-s> "nothing"
}


proc term_entry { } {
	##
	##  The query terms are stored in the global variable 'query'.
	##  It is then passed to 'c.term_entry' (in gui.cc)  which calls
	##  TermEntry(query) -- in TermEntry.c
	##
	##  The string returned to 'gui.cc' is of the form:
	##
	##    S:[ <terms_string> ]
	##    N:[ <terms_string> ]
	##    E:"Probably entry is an empty string"
	##
	##  In the first case <terms_string> might be a NULL string
	
	global query
	global working_query_terms_title
	global TermEntryLabel_2
	
	window_title
	set result [ c.term_entry $query ]
	
	##  puts "tcl-term_entry{} \[$result\]"

	##  Returned from gui.cc[ term_entry() ] is a string of the form:
	##
	##  <display_file>:<terms_not_found>
	##
	##  where <display file> = "EMPTY" if there no new entry terms,
	##    and <terms_not_found> = "EMPTY" if all terms are found.
	##
	##  or
	##
	##  <EMPTY>:<error_message>
	##
	##  where <error_message> indicates that there nothing was entered.
	
	if { $result == "too long" } {
		set error1 "The term entry line is too long."
		set error2 "Re-enter terms in smaller batches."
		set error "$error1\n$error2"
		puts "TCL:: $error1"
		puts "      $error2"
		display_error .error_window { Term Entry Message(s) } $error
	} else {
		set pos [ string first : $result ]
		incr pos -1
		set display_fname [ string range $result 0 $pos ]
		
		incr pos 2
		set extra [ string range $result $pos end ]
		
		##  display_fname = file name,
		##  extra terms not found or an error message.
		
		if { $display_fname != "EMPTY" } {
			.current_query_label configure -text $working_query_terms_title
			
			## delete .term_entry_box from start (0) to end if term found
			## .term_entry_box delete 0 end
			
			load_file .current_query_text $display_fname
			
			highlight_hit_terms .current_query_text SeaGreen white
			highlight_hit_title .current_query_text red white
			
			##  turn search button on
			
			.main_buttons.search configure -state normal
			bind .term_entry_box <Control-s> "db_search"
			.term_entry_label configure -text $TermEntryLabel_2
		}
		
		if { $extra != "EMPTY" } {
			display_error .error_window { Term Entry Message(s) } $extra

			set pos1 [ string first "exists as" $extra ]
			set pos2 [ string first "already in the query" $extra ]
			set pos3 [ string first "too long" $extra ]

			if { $pos1 != "0" } {
				## .term_entry_box delete 0 end
				.term_entry_box select range 0 end
			} elseif { $pos2 != "0" } {
				.term_entry_box delete 0 end
			} elseif { $pos3 != "0" } {
				##  Do nothing
			} else {
				##  select the entire entry widget
				.term_entry_box select range 0 end
			}
			
			##
			##  delete .term_entry_box from start (0) to end if term found
			##
			##    .term_entry_box delete 0 end
			
		} else {
			.term_entry_box delete 0 end
		}
	}
	
	window_title
}


##  Relevance judgements
##


proc make_rel { relj } {
	global rels_title
	global prev_rel

	##  puts "make_rel.tcl :: PREV_REL = $prev_rel"

	window_title
	
	set message_text "\nUpdating query .....\n"
	set window_name ".extracting"

	set back "NavyBlue"
	set fore "Yellow"

	update idletasks

	if { $relj == "F" } {
		processing_message $window_name $message_text $back $fore
	} elseif { $relj == "P" } {
		processing_message $window_name $message_text $back $fore
	}

	set result [ c.make_rel $relj $prev_rel ]

	##  result = %d:%s:%d:%s
	##
	##  i.e. <no_terms>:<query_fname>:<enable/disable search>:
	##       <enable/disable optimise>:<relspool fname>
	##
	##  no_terms

	set pos [ string first : $result ]
	incr pos -1
	set no_terms [ string range $result 0 $pos ]
	incr pos 2

	##  make result now %s:%d:%d:%s

	set result [ string range $result $pos end ]
	set pos [ string first : $result ]
	incr pos -1
	set display_fname [ string range $result 0 $pos ]
	incr pos 2

	##  make result %d:%d:%s

	set result [ string range $result $pos end ]
	set pos [ string first : $result ]
	incr pos -1
	set enable_search [ string range $result 0 $pos ]
	incr pos 2

	##  make result %d:%s
	set result [ string range $result $pos end ]
	set pos [ string first : $result ]
	incr pos -1
	set enable_optimise [ string range $result 0 $pos ]
	incr pos 2

	set rels_fname [ string range $result $pos end ]

	if { $display_fname != "DO NOT REDISPLAY" } {
		if { $display_fname != "EMPTY" } {
			load_file .current_query_text $display_fname
		}
	}

	load_file .positive_rels_text $rels_fname

	highlight_hit_terms .current_query_text SeaGreen White
	highlight_hit_title .current_query_text Red White

	highlight_current_rels .positive_rels_text Red Grey90
	highlight_previous_rels .positive_rels_text Grey50 Grey90

	.positive_rels_text configure -state disabled

	##  .positive_rels_label configure -text $rels_title

	if { $enable_search == "1" } {
		.main_buttons.search configure -state normal
		bind .term_entry_box <Control-s> "db_search"
	}

	##  if { $enable_optimise == "1" } {
	##    if { $relj != "N" } {
	##      bind .main_buttons.expand <Button-1> {
	##        busy2 query_expand "" ""
	##      }
	##      .main_buttons.expand configure -state normal
	##    }
	##  }

	if { $relj == "F" } {
		destroy $window_name
	} elseif { $relj == "P" } {
		destroy $window_name
	}
	
	window_title
}

##  End of relevance judgements procedures

#-- Procedure to move down to row containing the first
#-- highlighted term in the document

proc move_to_first_hl { w } {
	set result [ $w tag ranges passage_tag ]
	set first_line "0"
	scan $result "%d.%d" first_line col
	set top_row [ expr $first_line - 2 ]
	$w yview $top_row
}

proc move_to_pass_start { w } {
	set result [ $w tag ranges passage_tag ]
	set first_line "0"
	scan $result "%d.%d" first_line col
	set top_row [ expr $first_line - 2 ]
	$w yview $top_row
}


proc move_cursor { w direction } {
	set ipos "[$w index insert]"
	
	if { $direction == "L" } {
		incr ipos -1
	} elseif { $direction == "R" } {
		incr ipos +1
	}
	
	$w icursor $ipos
}

proc cursor_home { w } {
	set ipos 0
	$w icursor $ipos
}

proc cursor_end { w } {
	set ipos end
	$w icursor $ipos
}

proc delete_to_end { w } {
	set ipos "[$w index insert]"
	$w delete $ipos end
}  

##  Carry out query expansion after some rels have been made
##

proc query_expand { } {
	##
	##  result rturned in the format <%d:%s>
	##  i.e.
	##
	##  <no_terms>:<query_display_filename>
	
	set result [ c.query_expand ]
	
	window_title
	
	##  no_terms
	
	set pos [ string first : $result ]
	incr pos -1
	set no_terms [ string range $result 0 $pos ]
	incr pos 2
	
	set display_fname [ string range $result $pos end ]
	
	if { $display_fname != "DO NOT REDISPLAY" } {
		load_file .current_query_text $display_fname
		
		highlight_hit_terms .current_query_text SeaGreen white
		highlight_hit_title .current_query_text red white
	}
	
	##
	## disable expand button
	##
	
	##  .main_buttons.expand configure -state disabled
	##  bind .main_buttons.expand <Button-1> { }
	
	##  enable search button
	
	.main_buttons.search configure -state normal
	bind .term_entry_box <Control-s> "db_search"  
	window_title
}



##  Clear complete query, including relevance feedback
##

proc clear_query { } {
	##
	##  procedure to call c.clear_query(), a function in
	##  gui.cc that clears the current query.
	
	set result [ c.clear_query ]
	
	set pos [ string first : $result ]
	incr pos -1
	set clear_message [ string range $result 0 $pos ]
	incr pos 2
	set relspool_filename [ string range $result $pos end ]
	
	if { $clear_message == "QUERY CLEARED OK" } {
		clear_widget .term_entry_box
		clear_widget .current_query_text
		clear_widget .hitlist_text
		clear_widget .removed_terms_text
		
		if { $relspool_filename != "NO RELS" } {
			load_file .positive_rels_text $relspool_filename
			highlight_previous_rels .positive_rels_text Grey50 Grey90
			.positive_rels_text configure -state disabled
		}
	} else {
		display_error .error_window { Error Message } $result
	}
}


##  Clear relevance feedback terms only
##

proc clear_rf { } {
	global working_query_terms_title
	global TermEntryLabel_2
	
	set result [ c.clear_rf ]
	
	set pos [ string first : $result ]
	incr pos -1
	set display_fname [ string range $result 0 $pos ]
	incr pos 2
	set relspool_filename [ string range $result $pos end ]
	
	clear_widget .term_entry_box
	clear_widget .removed_terms_text
	clear_widget .hitlist_text

	if { $display_fname != "EMPTY" } {
		.current_query_label configure -text $working_query_terms_title
		
		## delete .term_entry_box from start (0) to end if term found
		## .term_entry_box delete 0 end
		
		load_file .current_query_text $display_fname
		
		highlight_hit_terms .current_query_text SeaGreen white
		highlight_hit_title .current_query_text red white
		
		##  turn search button on
		
		.main_buttons.search configure -state normal
		bind .term_entry_box <Control-s> "db_search"
		.term_entry_label configure -text $TermEntryLabel_2

		clear_widget .hitlist_text
	} else {
		display_error .error_window { Error Message } $display_fname
	}

	if { $relspool_filename != "NO RELS" } {
		load_file .positive_rels_text $relspool_filename
		highlight_previous_rels .positive_rels_text Grey50 Grey90
		.positive_rels_text configure -state disabled
	}
}

proc query_size_entry { menu_name } {
	##  Adapted from '/system/lib/tk/demos/mkForm.tcl
	
	global query
	
	set w ".query_size_entry"
	
	if { $menu_name != "" } {
		cancel_menu $menu_name
	}
	
	catch {
		destroy $w
	}
	
	toplevel $w -class Frame
	dpos $w
	
	wm title $w "Alter Maximum Query Size"
	wm iconname $w "Form"
	
	##  Determine position of entry box relative
	##  to top lefthand corner of options button
	
	if { $menu_name == ".options_menu" } {
		set x [ winfo rootx .main_buttons.options ]
		set y [ winfo rooty .main_buttons.options ]
		
		incr x -280
		incr y -200
	} else {
		set x 200
		set y 200
	}
	
	set position "400x100+$x+$y"
	wm geometry $w $position
	
	$w configure \
			-bd 2 \
			-bg LightGrey
	
	frame $w.query_size_entry \
			-relief raised \
			-bd 2 \
			-bg lightgrey
	
	frame $w.button_bar \
			-bd 2 \
			-bg slategrey
	
	message $w.entry_message \
			-font -*-fixed-bold-r-normal--15-140-*-*-*-*-*-* \
			-width 4i \
			-bg LightGrey \
			-text "Enter maximum query size (20 to 50 terms)"
	
	entry $w.entry_line \
			-relief sunken \
			-width 40 \
			-bg White \
			-textvariable query_size
	
	pack $w.query_size_entry \
			-side top \
			-fill both
	
	pack $w.entry_message $w.entry_line \
			-in $w.query_size_entry \
			-side top \
			-fill x
	
	button $w.ok \
			-text "Accept"\
			-font -*-fixed-bold-r-normal--15-140-*-*-*-*-*-* \
			-bg LightGrey \
			-activebackground LightGoldenRodYellow \
			-command "set_query_size $w"
	
	button $w.cancel \
			-text "Cancel" \
			-font -*-fixed-bold-r-normal--15-140-*-*-*-*-*-* \
			-bg LightGrey \
			-activebackground LightGoldenRodYellow \
			-command "destroy $w"
	
	pack $w.button_bar \
			-side bottom \
			-fill both
	
	pack $w.ok \
			-in $w.button_bar \
			-side left \
			-expand 1 \
			-fill x \
			-ipadx 4 \
			-ipady 2
	
	pack $w.cancel \
			-in $w.button_bar \
			-side right \
			-expand 1 \
			-fill x \
			-ipadx 4 \
			-ipady 2
	
	bind $w.entry_line \
			<Return> \
			"set_query_size $w"
	
	bind $w.entry_line \
			<Control-c> \
			"destroy $w"
	
	disable_close_button $w
	
	focus $w.entry_line
}


proc set_query_size { query_size menu_name } {

	set message_text "\nSetting max query size to $query_size.\n"
	set window_name ".query_size"
	
	set back "NavyBlue"
	set fore "Yellow"
	
	processing_message $window_name $message_text $back $fore

	set result [ c.set_query_size $query_size ]

	##  c.set_query_size returns:
	##
	##  <query_size>:<filename>
	##
	##  If successfully changed <filename> contains the name of the
	##  working query filename to be loaded into .current_query_text.
	##
	##  Otherwise it will contain the string:
	##
	##  "UNCHANGED"    query_size < MIN_QUERY_SIZE ||
	##                 query_size > MAX_QUERY_SIZE
	##  "ERROR"        if there is an error in the call to the C++ function

	set colon [ string first ":" $result ]
	incr colon -1
	set no_terms [ string range $result 0 $colon ]
	incr colon 2
	set filename [ string range $result $colon end ]

	if { $filename == "UNCHANGED" } {
		destroy $window_name
		set message "Error - query_size is unchanged ($no_terms terms)"
		display_error .error_window { Error Message } $message
	} elseif { $filename == "ERROR" } {
		destroy $window_name
		set message "Error - query_size is unchanged ($no_terms terms)"
		display_error .error_window { Error Message } $message
	} elseif { $filename != "EMPTY QUERY" } {
		destroy $window_name
		load_file .current_query_text $filename
		highlight_hit_terms .current_query_text SeaGreen white
		highlight_hit_title .current_query_text red white

		##  turn search button on
		##
		.main_buttons.search configure -state normal
		bind .term_entry_box <Control-s> "db_search"
	}

	cancel_menu $menu_name
}

proc adjust_rsv_factor { direction } {

	if { $direction == "D" } {
		##  D means decreasing RSV_FACTOR => increase no RF terms
		set updown "INCREASE"
	} else {
		##  U means increasing RSV_FACTOR => decrease no RF terms
		set updown "DECREASE"
	}

	set message_text "\nWill $updown number of RF terms if possible.\n"
	set window_name ".adjust_rsv_factor"
	
	set back "NavyBlue"
	set fore "Yellow"
	
	processing_message $window_name $message_text $back $fore

	set result [ c.adjust_rsv_factor $direction ]

	##  c.adjust_rsv_factor returns:
	##
	##  <query_size>:<filename>
	##
	##  If successfully changed <filename> contains the name of the
	##  working query filename to be loaded into .current_query_text.
	##
	##  Otherwise it will contain the string:
	##
	##  "UNCHANGED"    query_size < MIN_QUERY_SIZE ||
	##                 query_size > MAX_QUERY_SIZE
	##  "ERROR"        if there is an error in the call to the C++ function

	set colon [ string first ":" $result ]
	incr colon -1
	set new_rsv_factor [ string range $result 0 $colon ]
	incr colon 2
	set filename [ string range $result $colon end ]

	destroy $window_name

	if { $filename == "UNCHANGED" } {
		set message "Error - cannot $updown the number of RF terms"
		display_error .error_window { Error Message } $message
	} elseif { $filename == "ERROR" } {
		set message "Error - cannot $updown the number of RF terms"
		display_error .error_window { Error Message } $message
	} elseif { $filename != "EMPTY QUERY" } {
		load_file .current_query_text $filename
		highlight_hit_terms .current_query_text SeaGreen white
		highlight_hit_title .current_query_text red white

		##  turn search button on
		##
		.main_buttons.search configure -state normal
		bind .term_entry_box <Control-s> "db_search"
	}
}


proc set_lr_threshold { lr_threshold menu_name } {

	set message_text "\nSetting LR threshold to $lr_threshold.\n"
	set window_name ".lr_threshold"
	
	set back "NavyBlue"
	set fore "Yellow"
	
	processing_message $window_name $message_text $back $fore

	set result [ c.set_lr_threshold $lr_threshold ]

	##  c.set_lr_threshold returns:
	##
	##  <query_size>:<filename>
	##
	##  If successfully changed <filename> contains the name of the
	##  working query filename to be loaded into .current_query_text.
	##
	##  Otherwise it will contain the string:
	##
	##  "ERROR"        if there is an error in the call to the C++ function
	##  "UNCHANGED"
	##  "EMPTY"        no defined query terms

	set colon [ string first ":" $result ]
	incr colon -1
	set no_terms [ string range $result 0 $colon ]
	incr colon 2
	set filename [ string range $result $colon end ]

	destroy $window_name

	if { $filename == "ERROR" } {

		set message "Error - little r threshold will be unchanged ($lr_threshold)"
		display_error .error_window { Error Message } $message

	} elseif { $filename == "UNCHANGED" } {

		set message "\nLittle r threshold unchanged.\n"
		set window_name ".set_lr_threshold"
		set back "NavyBlue"
		set fore "Yellow"
	
		display_error $window_name { Set Little r Threshold } $message

	} else {

		if { $filename != "EMPTY QUERY" } {
			load_file .current_query_text $filename
			highlight_hit_terms .current_query_text SeaGreen white
			highlight_hit_title .current_query_text red white
			
			##  turn search button on
			##
			.main_buttons.search configure -state normal
			bind .term_entry_box <Control-s> "db_search"
		}

	}

	cancel_menu $menu_name
}


proc clear_widget { w } {
	if { $w == ".term_entry_box" } {
		##  entry widget
		$w delete 0 end
	} else {
		##  text widget
		##  First enable widget for editing
		
		$w configure -state normal
		$w delete 1.0 end
	}
}


##  Places window

proc dpos w {
	wm geometry $w +300+300
}

##  Options menu

proc post_menu { menu_name } {
	##
	##  First find (x, y) coordinates of top lefthand
	##  corner of options button
	##
	
	if { $menu_name == ".startup_menu" } {
		set x 400
		set y 400
	} else  {
		set x [ winfo rootx .main_buttons.options ]
		set y [ winfo rooty .main_buttons.options ]
		incr x -120
		incr y -100
	}
	
	$menu_name post $x $y
}


#-- Cancels the menus when the 'Cancel' menu option is selected.

proc cancel_menu { menu_name } {
	$menu_name unpost
}

##  Renames logfiles
##

proc quit_search { } {
	set title_bar "Ending Search Program"
	set msg "Click OK to confirm\n"
	set quit_level "F"
	
	##  display_error .error_window $title_bar $msg
	c.quit_gui  $quit_level
	exit
}

proc quit { } {
	set result [ c.quit_gui ]
	exit
}

## Generates an error dialog box
##

proc display_error { w title text } {
	## 1. Create the top-level window and divide it
	##    into top and bottom parts.
	
	catch {
		destroy $w
	}
	
	toplevel $w -class Dialog
	wm title $w $title
	wm iconname $w Dialog
	
	frame $w.top \
			-relief raised \
			-bd 1 \
			-bg White
	
	pack $w.top \
			-side top \
			-fill both
	
	frame $w.bot \
			-relief raised \
			-bd 1 \
			-bg LightGrey
	
	pack $w.bot \
			-side bottom \
			-fill both
	
	## 2. Fill the top part with the message.
	##    Font was:
	##
	##      -font -*-helvetica-bold-r-normal-*-120-*
	##
	##  until NTD mucked up the font server!

	message $w.msg \
			-width 3i \
			-justify center \
			-text $text \
			-bg White \
			-font -*-fixed-bold-r-normal--15-140-*-*-*-*-*-*
	
	pack $w.msg \
			-in $w.top \
			-side right \
			-expand 1 \
			-fill both \
			-padx 5m \
			-pady 5m
	
	## 3. Create an Ok button at the bottom of the dialog.
	
	button $w.ok_button \
			-text "Ok" \
			-command "press_return $w" \
			-bg NavyBlue \
			-fg LightGoldenRodYellow \
			-activebackground LightGoldenRodYellow \
			-activeforeground NavyBlue
	
	pack $w.ok_button \
			-in $w.bot \
			-side left \
			-expand 1 \
			-padx 3m \
			-pady 3m \
			-ipadx 2m \
			-ipady 1m
	
	bind $w.ok_button \
			<Return> \
			"press_return $w"
	
	## 4. Withdraw the window, then update all the geometry
	##    information so we know how big it wants to be, then
	##    center the window in the display and de-iconify it.
	
	wm withdraw $w
	update idletasks

	set x [ expr [ winfo screenwidth $w ]/2 - [ winfo reqwidth $w ]/2 \
			- [ winfo vrootx [ winfo parent $w ] ] ]
	set y [ expr [ winfo screenheight $w ]/2 - [ winfo reqheight $w ]/2 \
			- [ winfo vrooty [ winfo parent $w ] ] ]
	wm geom $w +$x+$y
	wm deiconify $w
	
	## 5. Set a grab and claim the focus too.
	
	set oldFocus [ focus ]
	grab $w
	focus $w
}

proc press_return { w } {
	##  puts "THIS IS THE PRESS_RETURN PROCEDURE"
	destroy $w
	##  .term_entry_box delete 0 end
}


## Initial quit box - option to cancel
##

proc tk_quit1 { w title text bitmap default args} {
	##  Modify this to use as final quit popup window
	##
	global tk_priv
	global indbut
	
	## 1. Create the top-level window and divide it
	##    into top and bottom parts.
	
	catch { destroy $w}
	
	toplevel $w \
			-class Dialog
	
	wm title $w $title
	wm iconname $w Dialog
	
	frame $w.top \
			-relief raised \
			-bd 1 \
			-bg White
	
	pack $w.top \
			-side top \
			-fill both
	
	frame $w.bot \
			-relief raised \
			-bd 1 \
			-bg LightGrey
	
	pack $w.bot \
			-side bottom \
			-fill both
	
	## 2. Fill the top part with bitmap and message.
	
	message $w.msg \
			-width 3i \
			-text $text \
			-font -*-helvetica-bold-r-normal-*-120-* \
			-bg White
	
	pack $w.msg \
			-in $w.top \
			-side right \
			-expand 1 \
			-fill both \
			-padx 5m \
			-pady 5m
	
	if { $bitmap != ""} {
		label $w.bitmap -bitmap $bitmap
		pack $w.bitmap \
				-in $w.top \
				-side left \
				-padx 5m \
				-pady 5m
	}
	
	## 3. Create a row of buttons at the bottom of the dialog.
	
	set i 0
	
	foreach but $args {
		button $w.button$i \
				-text $but \
				-command "set tk_priv(button) $i" \
				-bg LightGrey \
				-fg Black \
				-activebackground LightGoldenRodYellow
		
		pack $w.button$i \
				-in $w.bot \
				-side left \
				-expand 1 \
				-padx 3m \
				-pady 3m \
				-ipadx 2m \
				-ipady 1m
		
		incr i
	}
				
	## 4. Withdraw the window, then update all the geometry
	##    information so we know how big it wants to be, then
	##    center the window in the display and de-iconify it.
	
	wm withdraw $w
	update idletasks
	set x [ expr [ winfo screenwidth $w ]/2 - [ winfo reqwidth $w ]/2 \
			- [ winfo vrootx [ winfo parent $w ] ] ]
	set y [ expr [ winfo screenheight $w ]/2 - [ winfo reqheight $w ]/2 \
			- [ winfo vrooty [ winfo parent $w ] ] ]
	wm geom $w +$x+$y
	wm deiconify $w
	
	## 5. Set a grab and claim the focus too.
	
	set oldFocus [ focus ]
	grab $w
	focus $w
	
	## 6. Wait for the user to respond, then restore the focus
	##    and return the index of the selected button.
	
	tkwait variable tk_priv(button)
	destroy $w
	focus $oldFocus
	set indbut $tk_priv(button)
}


#-- Procedure to tag query terms which appear in 
#-- the document display

proc highlight_terms { w fg_colour bg_colour } {
	forAllMatches $w {<S>[0-z][0-9 A-Za-z]*[\-]*[0-9 A-Za-z]*[^\<]*<E>} {
		$w tag add term_tag first last
	}
	
	forAllMatches $w {<S>} {
		$w delete first last
	}
	
	forAllMatches $w {<E>} {
		$w delete first last
	}
	
	$w tag configure term_tag -foreground $fg_colour -background $bg_colour
}

proc highlight_passage { w fg_colour bg_colour } {
	
	forAllMatches $w {<P>[ -~]*[^\<]*<Q>} {
		$w tag add passage_tag first last
	}
	
	forAllMatches $w {<P>} {
		$w delete first last
	}
	
	forAllMatches $w {<Q>} {
		$w delete first last
	}
	
	$w tag configure passage_tag -foreground $fg_colour -background $bg_colour
}


proc highlight_last_it_terms { w fg_colour bg_colour } {
	##  Highlight docid line
	
	forAllMatches $w {<L>[ -~]*[^\<]*<M>} {
		$w tag add docid_tag first last
	}
	
	forAllMatches $w {<L>} {
		$w delete first last
	}
	
	forAllMatches $w {<M>} {
		$w delete first last
	}
	
	$w tag configure docid_tag -foreground $fg_colour -background $bg_colour
}


proc highlight_hit_title { w fg_colour bg_colour } {
	##  Highlight docid line
	
	forAllMatches $w {<S>[ -~]*[^\<]*<E>} {
		$w tag add docid_tag first last
	}
	
	forAllMatches $w {<S>} {
		$w delete first last
	}
	
	forAllMatches $w {<E>} {
		$w delete first last
	}
	
	$w tag configure docid_tag -foreground $fg_colour -background $bg_colour
}

proc highlight_hit_terms { w fg_colour bg_colour } {
	##  Highlight term occurrence info
	
	forAllMatches $w {<T>[ -~]*[^\<]*<W>} {
		$w tag add hit_term_tag first last
	}
	
	forAllMatches $w {<T>} {
		$w delete first last
	}
	
	forAllMatches $w {<W>} {
		$w delete first last
	}
	
	$w tag configure hit_term_tag -foreground $fg_colour -background $bg_colour
}


proc highlight_current_rels { w fg_colour bg_colour } {
	##  Full record rels
	##
	
	forAllMatches $w {<F>[ -~]*[^\<]*<G>} {
		$w tag add full_tag first last
	}
	
	forAllMatches $w {<F>} {
		$w delete first last
	}
	
	forAllMatches $w {<G>} {
		$w delete first last
	}
	
	$w tag configure full_tag -foreground $fg_colour -background $bg_colour
}

proc highlight_previous_rels { w fg_colour bg_colour } {
	##  Passage rels
	##
	
	forAllMatches $w {<P>[ -~]*[^\<]*<Q>} {
		$w tag add passage_tag first last
	}
	
	forAllMatches $w {<P>} {
		$w delete first last
	}

	forAllMatches $w {<Q>} {
		$w delete first last
	}
	
	$w tag configure passage_tag -foreground $fg_colour -background $bg_colour
}



##  Procedure to insert <S> ... <E> around highlighted
##  terms which appear in the document display.
##
##  This was put in to re-mark terms in a passage
##  highlighted by the searcher (by dragging the
##  cursor with the lefthand mouse pressed.

#--MESSAGES ETC----------------------------------------

##  Puts a busy cursor in the interface when something that
##  takes a long time is happening.
##
##  'busy2' requires 'busy1'

proc busy1 {} {
	global errorInfo
	
	set busy1 {
		.app .root
	}
	
	set list [ winfo children . ]
	
	while { $list != "" } {
		set next {}
		foreach w $list {
			set class [ winfo class $w ]
			set cursor [ lindex [ $w config -cursor ] 4 ]
			
			if { [ winfo toplevel $w ] == $w || $cursor != "" } {
				lappend busy1 [ list $w $cursor ]
			}
			
			set next [ concat $next [ winfo children $w ] ]
		}
		
		set list $next
	}
	
	foreach w $busy1 {
		catch {
			[ lindex $w 0 ] config -cursor watch
		}
	}
	
	update idletasks
	
	set error [ catch {uplevel eval [ list term_entry ]} result ]
	set ei $errorInfo
	
	foreach w $busy1 {
		catch {[ lindex $w 0 ] config -cursor [ lindex $w 1 ]}
	}
	
	if $error {
		error $result $ei
	} else {
		return $result
	}
}

proc busy2 { name1 name2 name3 } {
	global errorInfo
	
	set busy1 {}
	set busy2 {}
	set list [ winfo children . ]
	
	while { $list != "" } {
		set next {}
		foreach w $list {
			set class [ winfo class $w ]
			set cursor [ lindex [ $w config -cursor ] 4 ]
			
			if { [ winfo toplevel $w ] == $w || $cursor != "" } {
				lappend busy1 [ list $w $cursor ]
			}
			if { $cursor == "" && $class != "Frame" } {
				lappend busy2 [ list $w xterm ]
			}
			
			set next [ concat $next [ winfo children $w ] ]
		}
		set list $next
	}

	set busy3 [ concat $busy1 $busy2 ]
	
	foreach w $busy3 {
		catch {[ lindex $w 0 ] config -cursor {watch red white}}
	}
	
	update idletasks
	
	set error [ catch {uplevel eval [ list $name1 $name2 ]} result ]
	set ei $errorInfo
	
	foreach w $busy1 {
		catch {[ lindex $w 0 ] config -cursor [ lindex $w 1 ]}
	}
	
	foreach w $busy2 {
		catch {[ lindex $w 0 ] config -cursor "" }
	}
	
	if $error {
		error $result $ei
	} else {
		return $result
	}
}


##  timer -- displays time in a widget

proc timer name {
	.toplevel .watch
	label .watch.label -text $name
	label .watch.value -text "Time"
	pack .watch.label .watch.value -side left
}

##  Procedure for tagging any text which matches
##  a certain pattern so that it may be redisplayed differently. 
##  Can be used for highlighting search terms

proc forAllMatches {w pattern script} {
	scan [ $w index end ] %d numLines
	
	for {set i 0} {$i < $numLines} {incr i} {
		$w mark set last $i.0
		while {[ regexp -indices $pattern \
				[ $w get last "last lineend" ] indices ]} {
			$w mark set first "last + [ lindex $indices 0 ] chars"
			$w mark set last "last + 1 chars + [ lindex $indices 1 ] chars"
			uplevel $script
		}
	}
}


##  procedure to do nothing

proc nothing { } {
	##  Do nothing
	##  puts "DO NOTHING"
}


proc processing_message { wname message_text back fore } {
	## 1. Create the top-level window.
	
	catch {
		destroy $wname
	}
	
	toplevel $wname -class Dialog
	wm transient $wname .
	
	frame $wname.message_window \
			-relief raised \
			-bd 1 \
			-bg $back
	
	pack $wname.message_window \
			-side top \
			-fill both
	
	## 2. Fill the message_window
	##
	##    font was:
	##
	##      -font -*-helvetica-bold-r-normal-*-120-*
	##
	##  until NTD mucked up the font server!
	
	message $wname.msg \
			-bg $back \
			-fg $fore \
			-width 3i \
			-justify center \
			-text $message_text \
			-font -*-fixed-bold-r-normal--15-140-*-*-*-*-*-*
	
	pack $wname.msg \
			-in $wname.message_window \
			-side right \
			-expand 1 \
			-fill both \
			-padx 5m \
			-pady 5m
	
	## 4. Withdraw the window, then update all the geometry
	##    information so we know how big it wants to be, then
	##    center the window in the display and de-iconify it.
	
	wm withdraw $wname
	update idletasks
	
	##  Determine position to display window
	
	set root_x [ winfo rootx . ]
	set root_y [ winfo rooty . ]
	set root_w [ winfo screenwidth . ]
	set root_h [ winfo screenheight . ]
	
	set popup_w "400"
	set popup_h "80"
	set popup_x [ expr $root_x + ($root_w - $popup_w)/2 - 100 ]
	set popup_y [ expr $root_y + ($root_h - $popup_h)/2 - 100 ]
	
	wm geometry $wname 400x${popup_h}+${popup_x}+${popup_y}
	wm deiconify $wname
	
	## 5. Set a grab and claim the focus too.
	
	set oldFocus [ focus ]
	grab $wname
	focus $wname
	
	## Wait till message is displayed before returning
	##

	tkwait visibility $wname.msg
}


proc disable_close_button { w } {
	wm protocol $w WM_DELETE_WINDOW {
		puts "The close button is disabled."
	}
}

##  Packing and binding procedures
##

proc pack_it_all { } {
	##  Pack everything.
	##
	##  1. Term entry area
	##  2. Button area at bottom of the screen
	##  3. Left & right frames between term entry and button areas.
	##  4. Current query
	##  5. Removed terms
	##  6. Hitlist
	##  7. Rels pool
	
	##  Moved from interface.tcl
	##

	global main_frame_geometry

	wm minsize . 20 20
	wm maxsize . 1000 700

##	wm geometry . 1000x675+0+0
	wm geometry . $main_frame_geometry
	
	disable_close_button .
	
	##  1. The term entry area
	
	pack .term_entry_header \
			-side top \
			-fill x
	
	pack .term_entry_label \
			-in .term_entry_header \
			-side left
	
	pack .term_entry_frame \
			-side top \
			-fill x
	
	pack .term_entry_box \
			-in .term_entry_frame \
			-side left
	
	##  2. The button area
	
	pack .main_buttons \
			-side bottom \
			-fill x  
	
	pack .main_buttons.search \
			-in .main_buttons \
			-side left \
			-expand 1 \
			-fill x \
			-ipadx 4 \
			-ipady 2 

	pack .main_buttons.options \
			-in .main_buttons \
			-side left \
			-expand 1 \
			-fill x \
			-ipadx 4 \
			-ipady 2 

	pack .main_buttons.quit \
			-in .main_buttons \
			-side left \
			-expand 1 \
			-fill x \
			-ipadx 4 \
			-ipady 2 
	
	##  3. left and right frames in which to put current query,
	##     hitlist and rels pool windows.
	
	pack .left \
			-side left \
			-expand 1 \
			-fill both
	
	pack .right \
			-side right \
			-expand 1 \
			-fill both
	
	##  4. The current query areas in .left
	
	pack .current_query_frame \
			-in .left \
			-side top \
			-expand 1 \
			-fill both
	
	pack .current_query_label \
			-in .current_query_frame \
			-side top \
			-fill x
	
	pack .current_query_vscroll \
			-in .current_query_frame \
			-side right \
			-fill y
	
	pack .current_query_text \
			-in .current_query_frame \
			-expand 1 \
			-fill both
	
	##  5. The removed terms area in .left
	
	pack .removed_terms_frame \
			-in .left \
			-side bottom \
			-expand 1 \
			-fill both
	
	pack .removed_terms_vscroll \
			-in .removed_terms_frame \
			-side right \
			-fill y
	
	pack .removed_terms_text \
			-in .removed_terms_frame \
			-expand 1 \
			-side top \
			-fill both
	
	##  6. The hitlist
	##
	##     The label & scrollbars first. The listbox will
	##     fill the remaining area.
	
	pack .hitlist_frame \
			-in .right \
			-side top \
			-expand 1 \
			-fill both
	
	pack .hitlist_label \
			-in .hitlist_frame \
			-side top \
			-fill x
	
	pack .hitlist_vscroll \
			-in .hitlist_frame \
			-side right \
			-fill y
	
	pack .hitlist_text \
			-in .hitlist_frame \
			-side left \
			-expand 1 \
			-fill both
	
	##  7. Packing the user rels pool area
	##
	##     labels & scrollbars first. Listbox will fill remaining area
	
	pack .positive_rels_frame \
			-in .right \
			-side bottom \
			-expand 1 \
			-fill both
	
	pack .positive_rels_vscroll \
			-in .positive_rels_frame \
			-side right \
			-fill y
	
	pack .positive_rels_text \
			-in .positive_rels_frame \
			-side left \
			-expand 1 \
			-fill both
	
	##  Set up some Control-<key> bindings
	##
	
	bind_main_window_buttons
	
	global_bindings .term_entry_box
	global_bindings .current_query_text
	global_bindings .removed_terms_text
	global_bindings .hitlist_text
	global_bindings .positive_rels_text
}


proc bind_main_window_buttons { } {
	bind .main_buttons.search <Control-s> "db_search"
	
	##  bind .main_buttons.expand <Control-o> "busy2 query_expand \"\" \"\""
	
	bind .main_buttons.options {
		<Control-m> "post_menu .options_menu"
	}
}


proc database_menu { w } {
	post_menu $w.database
	cancel_menu $w
}


proc global_bindings { w } {
	bind $w <Control-s> "db_search"
	bind $w <Control-x> "quit_search"
	
	##  bind $w <Control-o> "busy2 query_expand \"\" \"\""
	##  bind $w <Control-m> "post_menu .options_menu"
	##  bind $w <Control-d> "database_menu .options_menu"
	##  bind $w <Control-c> "cancel_menu .options_menu"
	
	##  Widget specific bindings
	##
	
	if { $w == ".term_entry_box" } {
		bind .term_entry_box <Return> term_entry

		##  call disable_search procedure if the key pressed
		##  is not a modifier key. This disables the search
		##  button if a new terms has been started.

		bind .term_entry_box <KeyPress> {
			##  puts stdout { %%K = %K  %%A = %A }
			##  See Brent B. Welch p. 138 for explanation.
			if { "%A" != "{}" } {
				disable_search
			}
		}
	}
	
	if { $w == ".current_query_text" } {
		bind .current_query_text <Double-Button-1> {
			set cursor_pos "[ .current_query_text index insert]"
			busy2 remove_term $cursor_pos \"\"
		}
	} elseif { $w == ".removed_terms_text" } {
		bind .removed_terms_text <Double-Button-1> {
			set cursor_pos "[ .removed_terms_text index insert]"
			busy2 reinstate_term $cursor_pos \"\"
		}
	} elseif { $w == ".hitlist_text" } {
		bind .hitlist_text <Double-Button-1> {
			set cursor_pos "[ .hitlist_text index insert]"
			busy2 show_rec $cursor_pos \"\"
			selection clear
		}
	} elseif { $w == ".positive_rels_text" } {
		bind .positive_rels_text <Any-KeyPress> {
			nothing 
		}
	}
}
